// $Date: 2002/02/12 14:06:03 $ $Author: stefanw $ $Revision: 1.2 $

// rules for 'simplify'
// 
// bij 2000/2001
// 
// Rule (internal parameters)
// 
//        PATTERN      - pattern (DOM_EXPR)
//        RETURN       - return expression (DOM_EXPR)
//        COND         - conditions (DOM_SET) or hold(Cond) = DOM_SET
//                       set containing procedures or expressions (sse match)
//        PRIORITY     - priority 
//        INDEX        - index for identifying
//        STRATEGY     - strategy (optional)

// examples
// Rule(sin(X)^2 + cos(X)^2, 1)
//
//

Rule:= newDomain("Rule"):
Rule::create_dom:=hold(Rule):

alias(ARGS = hold(["pattern", "result", "cond", "priority", "index", "strategy"])):

Rule::new:=
  proc(PATTERN = FAIL, RESULT = FAIL, COND = {}, PRIORITY = 0, INDEX = 0, STRATEGY = "")
  begin
    if testargs() then
      if args(0) < 2 then
        error("at least two expressions required")
      end_if
    end_if;
    if type(COND) = "_equal" and op(COND, 1) = hold(Cond) then
      COND:= op(COND, 2)
    end_if;
    if INDEX = 0 and type(PATTERN) <> DOM_IDENT then
      INDEX:= type(PATTERN)
    end_if;
    sysassign(Rule::COUNT, Rule::COUNT + 1);
    new(Rule, PATTERN, RESULT, COND, PRIORITY, INDEX, STRATEGY)
  end_proc:

Rule::COUNT:= 0:

proc()
  local k;
begin
// create 6 similar procedures
//   from alias 'ARGS' for each parameter of a rule 
for k from 1 to nops(ARGS) do
  slot(Rule, op(ARGS, k)):=
  subsop(proc(RULE, VALUE = null())
         begin
           if VALUE = null() then // get
             extop(RULE, FAIL)
           else // set
             extsubsop(RULE, FAIL = VALUE)
           end_if
         end_proc, [4, 2, 2] = k, [4, 3, 2, 1] = k, Unsimplified)
end_for:
end():

// apply rule
// ARGS: RULE, EXPRESSION
Rule::apply:=
  proc(RULE, EXPRESSION)
    local MATCH;
  begin
    if (MATCH:= match(EXPRESSION, Rule::pattern(RULE),
                      hold(Cond) = Rule::cond(RULE))) <> FAIL then
      eval(subs(Rule::result(RULE), MATCH))
    else
      FAIL
    end_if
  end_proc:

Rule::print:=
  proc(RULE)
  begin
    expr2text(Rule::pattern(RULE))."  ->  ".expr2text(Rule::result(RULE))."\n"
    ._if(Rule::cond(RULE) <> {}, "Cond: ".expr2text(Rule::cond(RULE))."\n", "")
    ._if(Rule::index(RULE) <> 0 and Rule::index(RULE) <> type(Rule::pattern(RULE)),
         "Index: ".expr2text(Rule::index(RULE)), "")
    ._if(Rule::strategy(RULE) <> "", "   Strat: ".expr2text(Rule::strategy(RULE)), "")
    ._if(Rule::priority(RULE) <> 0, "   Prio: ".expr2text(Rule::priority(RULE)), "")
  end_proc:

alias(path = pathname("SIMPLIFY"));

// collection of rules
Rule::Old := loadproc(Rule::Old , path, "Old"):

(slot(Rule, k) := loadproc(slot(Rule, k), path, k))
  $ k in [//"Ci", "D", "Ei", "Im", "Re", "RootOf", "Si",
          "_and",
          "_div", // "_divide",
            "_equal",
          //"_equiv", "_exists", "_exprseq", 
            "_fconcat",
          //"_forall",
          //"function",
          //"_implies",
            "_in", "_intersect",
          //"_invert",
          "_leequal", "_less",
          "_minus", "_mod",
          "_mult",
          //"_negate",
          "_not", "_or",
          "_plus", "_power",
          //"_subtract",
          "_unequal", "_union",
          
          "abs", // "arccos", "arccosh",
          "arccot", // "arccoth",
          //"arccsc", "arccsch", "arcsec", "arcsech", "arcsin", "arcsinh",
          "arctan", // "arctanh", "arg", "asympt",
          //"bernoulli", "besselI", "besselJ", "besselK", "besselY",
          //"beta",
          "binomial",
          //"conjugate",
          "cos", "cosh",
          // "cot", "coth", "csc", "csch",
          //"diff", "dilog", "dirac", "divide",
          //"erf", "erfc", "exp", "fact",
          //"gamma", "gcov", "heaviside",
          //"int", "lambertV", "lambertW",
          "ln", "log",
          "max", "min",
          "modp", "mods",
          //"norm", "pdivide", "poly", "polylog",
          //"product", "psi",
          //"sec", "sech", "series", "sign", "signIm",
          "sin",
          "sinh", // "solve", "sqrt", "sum",
          "tan", "tanh" // ,"taylor", "zeta"
          ]:

null():
