// $Date: 2002/03/04 15:46:46 $ $Author: stefanw $ $Revision: 1.5 $

//
// Reduzieren der Regelbasis
// Anwenden der Regeln
// Bewertung
// -- Verwerfen/Akzeptieren
// -- Rckgabe/neuer Versuch
// 

// options: Old         - use old 'simplify'
//          Expand      - expand expression
//          Normal      - build normal form
//          Compact     - short expression
//          Rewrite     - rewrite expression
//                        with given function (3rd argument)
//   hold(Rules) = RULE - Regelbasis - als Domain, je Typ ein slot
//          

alias(allOptions = [hold(New), hold(Expand), hold(Rewrite),
                    hold(Normal), hold(Compact), hold(Old),
                    hold(ApplyRules), hold(Discard),
                    hold(MaxSteps), hold(SelectRules),
                    hold(Valuation)]):

Simplify::simplify:=
  proc(EXPR,                // expression to simplify
       OPT = hold(New))     // standard option
    local RULES,
          SFUNC, 
          valuation,
          exclude,
          selectRules,
          ApplyRules,
          OPTS:                  DOM_TABLE,
          maxSteps:              DOM_PROC,
          simplifySubExpression: DOM_PROC;
  begin
    if args(0) = 0 then
      // this should not give NIL, as in the old simplifier
      return()
    end_if;
    
    // set default options
    OPTS:= table(
                 hold(ApplyRules)  = hold(All),
                 hold(Discard)     = Simplify::exclude,
                 hold(MaxSteps)    = (actdepth -> float(20*exp(-actdepth))),
                 hold(SelectRules) = Simplify::selectRules,
                 hold(Valuation)   = Simplify::complexity);

    // combine default options and user defined options
    OPTS:= table(op(OPTS),
                 op(prog::getOptions(args(2..args(0)),
                                     {hold(Valuation), hold(Discard),
                                      hold(MaxSteps), hold(SelectRules),
                                      hold(ApplyRules)}))); 
   
    valuation:=   OPTS[hold(Valuation)];
    ApplyRules:=  OPTS[hold(ApplyRules)];
    selectRules:= OPTS[hold(SelectRules)];
    maxSteps:=    OPTS[hold(MaxSteps)];
    exclude:=     OPTS[hold(Discard)];
    
    // follow strategy
    if contains(OPTS, hold(Strat)) then
      warning("Strategies not implemented")
    end_if;

    SFUNC:= (X, Y) -> op(X, 2) < op(Y, 2) or
            op(X, 2) = op(Y, 2) and sysorder(op(X, 1), op(Y, 1));


    // l o c a l  m e t h o d
    // simplifySubExpression
    // EXPR: expression to simplify
    // totalval: complexity of the entire expression to be simplified
    //          

    simplifySubExpression:=
      proc(EXPR, totalval, ACTDEPTH = 1 : DOM_INT)
        local QUEUE, QOP, /* PMIN, */  RET, QARRAY, i,
        MINVAL, // minimum aller gefundenen Bewertungen
        MAXSTEPS;
//        option remember;
      begin
        MINVAL:= valuation(EXPR);
        // !!!Bewertung - Strategie-abhngig!!!
        QUEUE:= [[EXPR, MINVAL]];
        QOP:= 1; // PMIN:= 1;
        // function to sort results with respect to quality

        MAXSTEPS:= maxSteps(ACTDEPTH);
        
        while QOP <= nops(QUEUE)
          // Strategie-abhngig
          and QOP <= MAXSTEPS do
          
          // return simple expressions
          if domtype(op(QUEUE, [QOP, 1])) <> DOM_EXPR then
            QOP:= QOP + 1;
            next
          end_if;
          
          // select rule base

          RULES:= selectRules(Rule, op(QUEUE, [QOP, 1]));
         
          
          // apply all matching rules to current expression in queue
          RET:= Simplify::apply(op(QUEUE, [QOP, 1]), RULES, ApplyRules);
          // Gleiche aussortieren
          RET:= [op({op(RET)})];
          // Aussortieren von schon gefundenen Ausdrcken (keine Schleifen)

          userinfo(10, "Applying all rules to ".expr2text(op(QUEUE, [QOP, 1])).
                   " gives ".expr2text(RET));
          
          RET:= select(RET, X -> (contains(map(QUEUE, op, 1), X) = 0));
          // Bewertung 
          RET:= map(RET, X -> [X, valuation(X)]);
          
          // MINVAL ggf. nach unten korrigieren
          if nops(RET) > 0 then
            min(op(map(RET, op, 2)));
            if % < MINVAL then MINVAL:= % end_if;
          end_if;
          
          // Sortieren nach Bewertung
          // !! Alle aktuelle !!
          // RET:= sort(RET, SFUNC);
          // Aussortieren von zu teuren Ergebnissen!
          RET:= select(RET, exclude, MINVAL, totalval, ACTDEPTH);


          userinfo(10, "Selected results :".expr2text(RET));
          
          
          // Zeiger auf Minimum
          //  if nops(RET) > 0 and op(RET, [1, 2]) < op(QUEUE, [PMIN, 2]) then
          //    PMIN:= nops(QUEUE) + 1
          //  end_if;
      
          // Anhngen an Queue, Queuv
          QUEUE:= QUEUE.RET;


// schon hier: Sortierung -> Operanden des einfachsten Ausdrucks analysieren
// eventuell: mehr globale Vereinfachungen vor Operanden

          // Vereinfachung der Operanden von QUEUE[QOP]
          
// 
          QARRAY := map([op(op(QUEUE[QOP], 1))], simplifySubExpression,
                        op(QUEUE[QOP], 2),
                        ACTDEPTH + 1);

          for i from 1 to nops(op(QUEUE[QOP], 1)) do
            userinfo(10, "Subexpression ".expr2text(op(op(QUEUE[QOP], 1), i)).
                     " can be simplied to ".expr2text(map(QARRAY[i], op, 1)))
          end_for;

          RET:= [subsop(op(QUEUE[QOP],1), i= op(QARRAY[i][j], 1))
                        $j=1..nops(QARRAY[i])  $i=1..nops(QARRAY)];

          userinfo(10, "Replacing subexpressions gives ".expr2text(RET));
          
          
          RET:= [op({op(RET)})];
          // Aussortieren von schon gefundenen Ausdrcken (keine Schleifen)
          RET:= select(RET, X -> (contains(map(QUEUE, op, 1), X) = 0));

          userinfo(10, "Queue of remaining expressions is: ".expr2text(RET));
          
          // Bewertung - Options-abhngig!
          RET:= map(RET, X -> [X, valuation(X)]);
          
// exclude again ??
          // Aussortieren von zu teuren Ergebnissen!
          //RET:= select(RET, exclude, MINVAL, totalval, ACTDEPTH);
          
          
          QUEUE:= [op(QUEUE, 1..QOP)].
                  sort([op(QUEUE, QOP+1..nops(QUEUE))].RET, SFUNC);
          
          // Zeiger auf nchstes Element in der Queue
          QOP:= QOP + 1;
          // weiter in der Schleife -- Abbruchbedingung
          
        end_while;
        
        
        //op(sort(QUEUE, SFUNC), [1, 1])
        sort(QUEUE, SFUNC)
        
      end_proc;

    // use remember mechanism
    simplifySubExpression:=
      prog::remember(simplifySubExpression, Rule::COUNT, hold(None), id);

    //  m a i n   p r o g r a m

    
    simplifySubExpression(EXPR, valuation(EXPR), 1)
    
  end_proc:





Simplify::selectRules:=
proc(Ruleset : DOM_DOMAIN, expression)
  local Rules: DOM_LIST;
begin
  Rules:= slot(Ruleset, type(expression));
  if Rules = FAIL then
    Rule::Old
  else
    Rules.(Rule::Old)
  end_if
end_proc:


// Ausdruck, Liste von Regeln
// Anwenden aller Regeln (bis eine passt)
Simplify::apply:=
  proc(EXPR, RULES, OPT = hold(All))
    local k, RET;
    //option remember;
  begin
    if domtype(RULES) = Rule then
      Rule::apply(RULES, EXPR)
    else
      case OPT
        of hold(One) do
          for k in RULES do
            RET:= Rule::apply(k, EXPR);
            if RET <> FAIL then
              return(RET)
            end_if
          end_for;
          break
        of hold(All) do
          RET:= map(RULES, Rule::apply, EXPR);
          RET:= select(RET, _unequal, FAIL);
      end_case;
      RET
    end_if
  end_proc:

Simplify::valuation:=
  proc(EXPR, OPT = hold(None))
    //option remember;
  begin
    Simplify::complexity(EXPR, OPT)
  end_proc:





Simplify::depth:=
proc(EXPR)
begin
  if domtype(EXPR) <> DOM_EXPR then
    1
  else  
    1 + max(Simplify::depth(op(EXPR, i)) $i=0..nops(EXPR))
  end_if
end_proc:


// Simplify::exclude(EXPR, MINVAL, totalval, ACTDEPTH)
// EXPR: [expression, valuation]
// MINVAL: minimal valuation found up to now
// totalval: total complexity of the expression to which
//           the subexpression EXPR belongs
// ACTDEPTH: recursion depth
Simplify::exclude:=
  proc(EXPR, MINVAL, totalval, ACTDEPTH)
    //option remember;
  begin
    op(EXPR, 2) < MINVAL*(1 + 10*MINVAL/totalval/float(exp(ACTDEPTH)))
  end_proc:
