// $Date: 2002/02/12 14:05:07 $ $Author: stefanw $ $Revision: 1.2 $

// returns the maximal depth of the expression tree
Simplify::depth:=
proc(EXPR)
begin
  if domtype(EXPR) <> DOM_EXPR then
    1
  else  
    1 + max(Simplify::depth(op(EXPR, i)) $i=0..nops(EXPR))
  end_if
end_proc:


Simplify::defaultOperatorTable:=
table(
"Default" = 500,  // unknown operators
"Atomic" = 1, // non -expressions     
"_plus" = 2,
"_mult" = 3,
"_power" = 7,
"exp" = 30,      
"sin" = 30,
"cos" = 30,
"tan" = 50,
"cot" = 70,
"arcsin" = 200,
"arccos" = 200,
"arctan" = 300      
      
      ):

Simplify::operatorComplexity:=
proc(T = Simplify::defaultOperatorTable: DOM_TABLE)
  option escape;
  local f: DOM_PROC;
begin

  // return value:
  f:=
  proc(EXPR)
  begin
    if domtype(EXPR) <> DOM_EXPR then
      T["Atomic"]
    elif contains(T, type(EXPR)) then
      T[type(EXPR)] + _plus(f(op(EXPR, i)) $i=1..nops(EXPR))
    else  
      T["Default"] + _plus(f(op(EXPR, i)) $i=1..nops(EXPR))
    end_if  
  end_proc
  
end_proc:

  

// returns an integer value correlated to the complexity of a
// given expression
Simplify::complexity:=  
proc(EXPR, OPT=hold(None))
  local compl, DEPTH, LENGTH, OPR, T;
    //option remember;
  begin
    compl:=
      proc(EXPR)
        local c;
      begin
        if domtype(EXPR) = DOM_EXPR then
          c:= map([op(EXPR)], compl);
          T:= table(type(EXPR) = 1);
          // number off all operands
          _plus(op(c, 4*k-3) $ k = 1..nops(c)/4),
          // max number operands of one subexpression
          stdlib::max(nops(EXPR), op(c, 4*k-2) $ k = 1..nops(c)/4),
          // maximal expr tree depth
          DEPTH*(1 + max(op(c, 4*k-1) $ k = 1..nops(c)/4)),
          // table of all operators and types
          (map(map([op(c, 4*k) $ k = 1..nops(c)/4], op),
               X -> (if contains(T, op(X, 1)) then
                       T[op(X, 1)] := T[op(X, 1)] + op(X, 2)
                     else
                       T[op(X, 1)]:= op(X, 2)
                     end_if)); T)
        else
          nops(EXPR)*LENGTH, nops(EXPR)*LENGTH, 0, table(type(EXPR) = 1)
        end_if
      end_proc;
    
    DEPTH:= 1;
    LENGTH:= 1;
    OPR:= table("All" = 1);

    case OPT
      of hold(Flat) do
        DEPTH:= 2;
        break
      of hold(Short) do
        LENGTH:= 2;
        break
      of hold(Simple) do
        // ??

    end_case;
    
    [compl, LENGTH, DEPTH, OPR]:= [compl(EXPR)];

    //print("sum nops" = compl, "max nops" = LENGTH, "max depth" = DEPTH, OPR);

    // standard arithmetic
    OPR:= split(OPR, X -> contains([DOM_INT, DOM_RAT, DOM_FLOAT, DOM_COMPLEX,
                                    "_plus", "_mult", "_power"], op(X, 1)) > 0);
    
    max(compl, LENGTH) + DEPTH*nops(op(OPR, 1)) + 3*DEPTH*nops(op(OPR, 2))
  end_proc:
