// $Date: 2001/08/13 07:40:28 $ $Author: tom $ $Revision: 1.2 $


// stefanw, 7. 2. 2001

// adt::Stream

// the domain of streams

// a Stream is an object that can be called many times,
// returning an infinite sequence of objects

// a stream has one operand:
// it equals a procedure to be evaluated
// in order to get the next element returned by the stream


adt::Stream:= newDomain("adt::Stream"):
adt::Stream::create_dom := hold(adt::Stream):


// Method new
// create a stream from three arguments:
// 1. a procedure called every time the stream is to return an element
// 2. a list of static variables
// 3. a statement sequence executed once during initialization

adt::Stream::new:=
proc(inner, static=[], initialization=NIL)        
  option hold;
  option escape;
  local prototype;
begin
  inner:= context(inner);
  repeat
    static := context(hold(level)(static, 1))
  until type(static)=DOM_LIST
  end_repeat;
  prototype:=
  text2expr("proc() option escape; local ".expr2text(op(static)).";".
   "begin ".
   expr2text(initialization).";".
   expr2text(inner)." end_proc");
  new(dom, prototype())
end_proc:

adt::Stream::staticVars:=
proc(S: adt::Stream)
begin
  op(S, [1, 12, 1, 2])
end_proc:

adt::Stream::staticValues:=
proc(S: adt::Stream)
begin
  op(S, [1, 12]);
  op(%, 6..nops(%))
end_proc:

adt::Stream::getStaticValue:=
proc(S: adt::Stream, x:DOM_IDENT)
  local i:DOM_INT;
begin
  i := contains([dom::staticVars(S)], x);
  if i=0 then
    error("No such static variable")
  else
    op(S, [1, 12, 5+i])
  end_if
end_proc:


    
adt::Stream::func_call:=
proc(S:adt::Stream)
begin
  extop(S,1)(args(2..args(0)))
end_proc:

adt::Stream::print:=
proc(S:adt::Stream)
begin
  "Stream(...)"
end_proc:

adt::Stream::expose:=
proc(S)
begin
  if args(0)=0 then
    hold(newDomain)(op(adt::Stream))
  elif args(0)=1 then
    if type(S) <> dom then
      error("Wrong type of argument")
    else
      stdlib::Exposed
      ("Stream\n".
       "Procedure used to generate next element:\n".
       op(expose(extop(S,1)))."\n".
       "Current state of static variables:\n".
       op(map(map(zip([adt::Stream::staticVars(S)],
                      [adt::Stream::staticValues(S)],
           _equal), expr2text), _concat, "\n"))
       )
    end_if
  else
    error("Wrong number of arguments")
  end_if
end_proc:
  
// end of file
