/*****************************************************************************
  $Id: cartesianProduct.mu,v 1.3 2001/11/21 21:23:47 nthiery Exp $
  File:		lib/COMBINAT/cartesianProduct.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	21/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::cartesianProduct

    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::cartesianProduct";

    info := "Library 'combinat::cartesianProduct': utilities for cartesianProducts";

    introduction := "
 Let set1, set2, ..., setN be sets.
 The cartesian product set1 x set2 x ... x setN of set1, set2, ..., setN
 is the set of all N-tuples [x1,x2,...,xN] with xi in seti for each 1<=i<=N. 
                                        
 Each set seti can be represented by a set (DOM_SET), a list
 (DOM_LIST), or an integer (which represents {1..n}). 
                
 Issue: does it mean anything to have twice the same element in one of sets ?

 See also combinat::structure.
";

    synopsis := "
 - type				any list (DOM_LIST)

 - count(set1,...,setN)		number of all elements of set1 x ... x setN
 - list(set1,...,setN)		list of all elements of set1 x ... x setN
 - generator(set1,...,setN)	generator for all elements of set1 x ... x setN

   seti: a set (DOM_SET, DOM_LIST) or an integer n>=0 representing {1..n}
";

    type := DOM_LIST;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc()
    begin
        _mult((if domtype(s)=DOM_INT then s else nops(s) end_if)
              $ s in [args()]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // list
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc()
        : combinat::generators::type
    /* 
       generator(set1,...,setN)	generator for all elements of set1 x ... x setN

       Each call of the generator yields the following element
       lexicographically or FAIL if there is no more. There is
       absolutely no waranty if the generator is called again after
       FAIL was returned.

       Is the result reasonable when a set contains twice an element ? */

        local word, word_int, A, i, first_time;
        option escape;
    begin
        if args(0)=0 then
            return(combinat::generators::repeater([],1));
        end_if;
	A:=[args()]:
        for i from 1 to nops(A) do
            if domtype(A[i])=DOM_INT then
                A[i]:={$1..A[i]};
            end_if;
            if nops(A[i])=0 then
                return(combinat::generators::empty());
            end_if;
        end_for;
        word_int := [1 $ args(0)];
        word     := [op(s,1) $ s in A];
        first_time:=TRUE;
        // Invariants:
        //  - word_int is a sequence of k integers with
        //		1 <= word_int[i] <= nops(A[i])
        //  - word is the corresponding element of A[1] x ... x A[k]
        //  - first_time is true if and only if it's the first run
        return(proc(): Type::Union(dom::type, DOM_FAIL)
                   local index;
               begin
                   if first_time then
                       first_time:=FALSE;
                       return(word);
                   end_if;
                   for index from nops(A) downto 1 do
                       if word_int[index] < nops(A[index]) then
                           word_int[index] := word_int[index]+1;
                           word[index] := op(A[index], word_int[index]);
                           return(word);
                       else 
                           word_int[index] := 1;
                           word[index] := op(A[index],1);
                       end_if;
                   end_for;
                   return(FAIL);
               end_proc)
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::cartesianProduct"] := TRUE:
