/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/compositions.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::compositions

    category combinat::integerListsLexBuilding;
    // category inlinedoc::cat;

    Name := "combinat::compositions";

    info := "Library 'combinat::compositions': utilities for compositions";

    introduction := "
 A composition of n is a list of positive integers which sum up to n.

 Compositions are naturally ordered lexicographically.
 Several functions of this library are directly inherited from
 combinat::integerListsLexBuilding, and accept generic
 constraints:

  - length of the composition:	MinLength = , MaxLength = , Length =
  - size of parts: 		MinPart = , MaxPart =
  - slope:			MinSlope = , MaxSlope =
  - containment:		Inner = , Outer =

 See the documentation of combinat::integerListsLexBuilding and the
 examples for details on those options.

 See also combinat::structure, combinat::integerListsLexBuilding.
";

    synopsis := "
 - type:		list of positive integers

 - count(n, <Constraints>)		number of compositions of n
 - list(n, <Constraints>)		list of all compositions of n
 - generator(n, <Constraints>)		generator for all compositions of n

 - first(n, <Constraints>)		first composition of n
 - last(n, <Constraints>)		last composition of n
 - Next(c, <Constraints>)		next composition
 - previous(c, <Constraints>)		previous composition
";

    interface := dom::interface_Structure union 
        {hold(first), hold(last), 
         hold(Next), hold(previous)
         };

    type := Type::ListOf(Type::NonNegInt);

    parseOptions :=
    proc()
    begin
        dom::parseOptions_IntegerListLexBuilding
        (table(hold(MinLength)	= 0,
               hold(MaxLength)	= infinity,
               hold(Length)	= NIL,
               hold(MinPart)	= 1,
               hold(MaxPart)	= infinity,
               hold(Inner)	= NIL,
               hold(Outer)	= NIL,
	       hold(MinSlope)	= -infinity,
	       hold(MaxSlope)	= infinity
              ),
         args()
         ):
    end_proc;

    /// count(n, <Options>)	number of compositions of n
    count :=
    proc(n: Type::Integer) : Type::NonNegInt
    begin
        if args(0) > 1 then
            // There is no efficient counter yet when there are
            // further constraints.
            return(dom::count_integerListsLexBuilding(args()));
        end_if;

        if n >= 1 then
            2^(n-1);
        elif n=0 then
            1
        else
            0;
        end_if;
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::compositions"] := TRUE:
