combinat::numberOfSetPartitionColorings :=		slot(funcenv(
proc(partition: Type::ListOf(Type::NonNegInt), 
     colors:    Type::ListOf(Type::NonNegInt))
    local rec_color, rec_size;
begin
    if testargs() and
	_plus(op(colors)) <> _plus(i*partition[i] $ i=1..nops(partition))
	then
	error("The size of the set partition and the number of colors do no match!");
    end_if;

    // Recursion on the number of colors
    rec_color:=
    proc(partition: Type::ListOf(DOM_INT), colors: Type::ListOf(DOM_INT))
    begin
	if nops(colors)=1 then
	    return(1);
	else
	    return(rec_size(partition, colors, nops(partition)));
	end_if;
    end_proc:

    // Recursion on the size of the subsets in the partition
    rec_size:=
    proc(partition: Type::ListOf(DOM_INT), colors: Type::ListOf(DOM_INT),
	 k:DOM_INT)
	local result, m;
    begin
	if colors[1]=0 then
	    return(rec_color(partition,[op(colors,2..nops(colors))]));
	end_if;
	if k=0 then
	    return(0);
	end_if;
	result:=0;
	
	// number of yet uncolored subsets of size k in the partition
	m:=partition[k];
	// Iterate through the number of subsets of size k to be
	// colored with color 1.
	while(partition[k]>=0 and colors[1]>=0) do
	    result	:= result+
		binomial(m, partition[k])*rec_size(partition, colors, k-1);
	    colors[1]	:= colors[1]-k;
	    partition[k]:= partition[k]-1;
	end_while;
	result;
    end_proc:
    
    // rec_color is more efficient if colors does not contain zeroes,
    // and is sorted decreasingly
    colors:=revert(sort(select(colors, not iszero)));
    
    rec_color(partition, colors);
end_proc                                                                                                                 ), "info","
numberOfSetPartitionColorings(partition: Type::ListOf(Type::NonNegInt),
                              colors:    Type::ListOf(Type::NonNegInt))
  -> Type::NonNegInt

Given a set partition of {1..n} such that partition[k] is the
number of subsets of size k in the partition, count in how many
ways the elements of {1..n} can be colored such that:
 - exactly colors[i] elements of {1..n} get color i;
 - all the elements in a given subset receive the same color.
                       
Complexity: exponential.

One cannot expect much better since already with two colors, this is
equivalent to the standard knapsack problem.
"):
