// $Id: rsa.mu,v 1.3 2002/06/28 12:28:31 tonner Exp $

// RSA key generation, signing, and signature verification.

crypto::rsa_keygen :=
proc(keylen, randseed)
  local p, q, phi, e, d, r1, r2, qp;
begin
  r1 := frandom(rtime());
  r2 := frandom(randseed);
  q := nextprime(2^(floor(keylen/2))+floor(2^(keylen/4)*(r1()-r2())));
  p := nextprime(2^(ceil(keylen/2))+floor(2^(keylen/4)*(2*r1()-r2())));

  if p<q then [p,q] := [q,p]; end_if;
  
  phi := (p-1)*(q-1);
  e := 2:
  while gcd(phi, e) <> 1 do
    e := abs(2*floor(50*(r1()-r2()))+1) mod phi;
  end_while;

  d := igcdex(e, phi)[2] mod phi;

  if (d<0) then d := d+phi; end_if;
  assert(d*e mod phi = 1);

  qp := igcdex(q, p)[2] mod p;
  assert(q*qp mod p = 1);
  
  return([[p*q, e], [p*q, p, q, qp, d]]);
end_proc:

crypto::rsa_sign :=
proc(message, privkey)
  local hash, N, p, q, qp, d, mp, mq, m;
begin
  [N, p, q, qp, d] := privkey;
  assert(q*qp mod p = 1);
  hash := crypto::cvhp_hash(message);

  hash := hash mod 2^200 + 2^200*floor(frandom(rtime())()*N);
  
  mp := powermod(hash, d mod (p-1), p);
  mq := powermod(hash, d mod (q-1), q);

  m := ((mp-mq)*qp mod p)*q + mq;
//  assert(m = powermod(hash, d, N));
  m;
end_proc:

crypto::rsa_check :=
proc(message, signature, pubkey)
  local hash, m, N, e;
begin
  [N, e] := pubkey;
  hash := crypto::cvhp_hash(message);

  m := powermod(signature, e, N);
  bool(m mod 2^200 = hash mod 2^200);
end_proc:
