#++
Cat::InvolutiveDivision -- the category of involutive divisions

Author: Marcus Hausdorf, Uni Mannheim
        (hausdorf@math.uni-mannheim.de)
Created: 	Mon Dec 10 16:18:31 2001
Last update: 	Thu Jan 31 12:53:39 2002
++#

#++
Arguments:

nV: length of the multiindices
	
Entries:

separationTableMult(mu_l: list of multiindices of length nV):
    sep_t: table(mu -> set of multiplicative variables for mu_l[i] $ mu in mu_l)
    
isInvDivisor(mu, nu: multiindices of length n):
    0, if mu does not involutively divide nu;
    (nu - mu) otherwise 
++#

category Cat::InvolutiveDivision(nV:Type::PosInt)

    category Cat::BaseCategory;

    //==================
    // undefined methods
    //==================
    
    numVars := nV;
    
    separationTableMult;
    
    separationTableNonMult :=
	(x -> map(dom::separationTableMult(x), l -> ({i $ i = 1..nV} minus l)));
    
    isInvDivisor := proc(mu:Type::ListOf(Type::NonNegInt), nu:Type::ListOf(Type::NonNegInt),
	mV_s:Type::SetOf(Type::PosInt))
	local tmp, k;
	begin
	    tmp := zip(nu, mu, _subtract);
	    for k from 1 to nV do
		if tmp[k] < 0 or (tmp[k] > 0 and not contains(mV_s, k)) then
		    return(0);
		end_if;
	    end_for;
	    tmp;
	end_proc;

    begin    
    if args(0) <> 1 then error("wrong no of args") end_if;

end_category;











