/*****************************************************************************
  $Id: SemiGroup.mu,v 1.1 2001/12/10 20:06:48 nthiery Exp $
  File:		lib/DOMAINS/CATEGORY/PermutationGroup.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
		Original version by the MuPAD group
  Taken from:	PerMuVAR, 02/12/2001
  Taken from:	MuPAD 1.4/2.0, 1998-2000
  Created:	08/08/1999
  License:	LGPL
*****************************************************************************/

/*++
Cat::SemiGroup -- the category of semi-groups

A Cat::SemiGroup is a set with an associative operation '*'.

Methods:
_mult(x1,x2,...) - returns the product of x1*x2*...
_power(x,i)	 - returns x*x*...*x (i times) for a positive integer i
multiplicativeClosure(generators) - list of elements generated by generators

++*/

category Cat::SemiGroup
    category Cat::BaseCategory;

    _mult;

    _power := domains::repeatedSquaring(dom, "_mult");
    
/*--
    multiplicativeClosure(generators)

    generators: list of elements of the group.

    Compute the multiplicative closure of `generators', that is the
    list of all the elements of the group that can be obtained by
    product of the generators in `generators'. Or the sub-semigroup
    generated by `generators'.

    Complexity: O(#closure#generators), where #closure is the size of
    the generated closure, and #generators is the number of generators.

    Warning: there is no way of knowing how big the closure is going to
    be or even if its finite or not. Therefore there are no warranty
    of termination.

--*/

    multiplicativeClosure :=
    proc(generators: Type::ListOf(dom))
	local closure, todo, g, g2, r;
    begin
	closure:=table(r=TRUE $ r in generators);
	todo:=generators;
	while nops(todo)>0 do
	    g:=todo[1];
	    delete(todo[1]);
	    for r in generators do
		g2:=g*r;
		if contains(closure,g2) then
		    userinfo(hold(NoNL), 10, "*");
		else
		    userinfo(hold(NoNL), 10, "+");
		    if nops(closure) mod 100=0 then
			userinfo(hold(NoNL), 0, "(".nops(closure).")")
		    end_if;
		    closure[g2]:=TRUE;
		    todo:=append(todo, g2);
		end_if
	    end_for
	end_while;
	userinfo(10," Done.");
	map([op(closure)], op,1);
    end_proc;

begin

end_category:
