#++
Dom::WeylAlgebra

Author: Marcus Hausdorf, Uni Mannheim
        (hausdorf@math.uni-mannheim.de)
Created: 	Tue Oct  1 11:03:25 2001
Last update: 	Tue Feb 19 17:19:30 2002
++#

alias(REP(x) = extop(x, 1)):

domain Dom::WeylAlgebra(Vars, R = Dom::Rational, Order = LexOrder)
local Dummy_l:DOM_LIST, Var_l:DOM_LIST, Dummy2Var:DOM_TABLE, Var2Dummy:DOM_TABLE,
    Rcoerce:DOM_PROC, Rinvert:DOM_PROC, nV:DOM_INT, Pdiff:DOM_PROC, oM, MO:DOM_DOMAIN,
    Homogenised:DOM_BOOL, HomVar:DOM_IDENT;
inherits Dom::MultivariatePolynomial(Dummy_l, R, Order);
category Cat::SolvableAlgebra(R);
axiom (if R::hasProp(Ax::normalRep) then Ax::normalRep end_if),
      (if R::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if);
	  
//==============
// In- / Output 
//==============

new := proc(x)
    begin
 	dom::convert(x);
 	if % = FAIL then
 	    error("illegal arguments");
 	else
	    return(%);
 	end_if;
    end_proc;

convert := proc(x)
    local s_l:DOM_LIST;
    begin
	case domtype(x)    
	of dom do
	    return(x);
	of DOM_LIST do
	    if ({op(map(x, domtype))} <> {DOM_LIST}
		or {op(map(x, nops))} <> {2})
	    then
		return(FAIL);
	    end_if;
	    if ({op(map(map(x, l -> l[2]), domtype))} <> {DOM_LIST}
		or {op(map(map(x, l -> l[2]), nops))} <> {nV}
		or {map(op(x[i][2]) $ i = 1..nops(x), testtype, 
		Type::NonNegInt)} <> {TRUE})
	    then
		return(FAIL);
	    end_if;
	    if has((x := map(x, l -> [Rcoerce(l[1]), l[2]])), FAIL) then
		return(FAIL);
	    end_if;
	    return(new(dom, poly(x, Dummy_l, R)));
	of DOM_INT do
	of DOM_RAT do
	of DOM_FLOAT do
	of DOM_COMPLEX do
	of DOM_IDENT do
	of DOM_EXPR do
	    if iszero(x) then
		return(dom::zero);
	    end_if;
	    x := expand(x);
	    case type(x)	    
	    of "_plus" do
		s_l := [map(op(x), l -> (if type(l) = "_mult" then [op(l)] else [l] end_if))];
		break;
	    of "_mult" do
		s_l := [[op(x)]];
		break;
	    otherwise
		s_l := [[x]];
	    end_case;
	    s_l := map(s_l, l -> map(l, l2 ->
		(if type(l2) = "_power" then
		    if (testtype(op(l2, 2), Type::PosInt) 
			and contains(Var2Dummy, op(l2, 1)))
		    then
			return(hold(_power)(Var2Dummy[op(l2, 1)], op(l2, 2)));
		    end_if;
		elif contains(Var2Dummy, l2) then
		    return(Var2Dummy[l2]);
		else
		    return(Rcoerce(l2));
		end_if;)));
	    if not iszero(contains(map(s_l, op), FAIL)) then
		return(FAIL);
	    end_if;
	    if nops(s_l) = 1 then 
		if nops(s_l[1]) = 1 then
		    return(new(dom, poly(s_l[1][1], Dummy_l, R)));
		else
		    return(new(dom, poly(hold(_mult)(op(s_l[1])), Dummy_l, R)));
		end_if;
	    else
		return(new(dom, poly(hold(_plus)(op(map(s_l, l -> 
		    (if nops(l) = 1 then op(l) else hold(_mult)(op(l)) end_if)))), 
		    Dummy_l, R)));
	    end_if;
	of DOM_POLY do
	    return(dom::convert(expr(x)));
	of R do
	    return(new(dom, poly(x, Dummy_l, R)));
	otherwise
	    return(FAIL);
	end_case;
    end_proc;
    
convert_to := proc(x:dom, T)
    local DOM:DOM_DOMAIN;
    begin
	case T
	of dom do
	    return(x);
	of DOM_EXPR do
	    return(dom::expr(x));
	of DOM_POLY do
	    return(subs(REP(x), op(Dummy2Var)));
	of R do
	    if iszero(dom::degree(x)) then
		return(dom::coeff(x));
	    end_if;
	otherwise
	    return(FAIL);
	end_case;
    end_proc;
	    
expr := proc(x:dom)
    begin
	if dom::iszero(x) then
	    return(0);
	end_if;
 	subs(expr(REP(x)), op(Dummy2Var));
    end_proc;

print := proc(x:dom)
    local res_l:DOM_LIST;
    begin
	if dom::iszero(x) then
 	    return(0);
	end_if;
	res_l := map(map(polylib::sortMonomials(REP(x), Order), expr), subs, op(Dummy2Var));
	if nops(res_l) = 1 then
	    return(op(res_l));
	else
	    return(hold(_plus)(op(res_l)));
	end_if;
    end_proc;
	    	
coeffRing := R;
    
variables := Vars;

exponentOrder := Order;

orderMatrix := oM;

numVars := nV;

dummyList := Dummy_l;

//====================
// Polynomial Methods 
//====================

poly2list := (p -> (if dom::iszero(p) then return([[Rzero, [0 $ nV]]])
    elif nV > 1 then return(poly2list(REP(p)))
    else return(map(poly2list(REP(p)), (l -> [l[1], [l[2]]]))) end_if));
        
exponent2monomial := (mu -> new(dom, poly([[R::one, mu]], Dummy_l, R)));

monomial2exponent := (p -> dom::poly2list(p)[1][2]); 

coeff := proc(x:dom)
    local tmp;
    begin
	if args(0) = 1 then
	    return(coeff(REP(x)));
	else
	    tmp := coeff(REP(x), args(2..args(0)));
	    if domtype(tmp) = DOM_POLY then
		return(new(dom, tmp));
	    else
		return(tmp);
	    end_if;
	end_if;
    end_proc;
    
lmonomial := proc(x:dom)
    local DOM; 
    begin
	DOM := dom;
	if args(0) = 1 then
	    return(new(dom, lmonomial(REP(x), Order)));
	elif args(2) = hold(Rem) then
	    return(map(lmonomial(REP(x), Order, hold(Rem)), 
		l -> (new(DOM, l))));
	elif args(0) = 2 then
	    return(new(dom, lmonomial(REP(x), args(2))));
	else 
	    return(map(lmonomial(REP(x), args(2), args(3)), 
		l -> (new(DOM, l))));
	end_if;
    end_proc;

terms := proc(x:dom)
    local DOM;
    begin
	map(dom::poly2list(x), l -> (DOM::multcoeffs(DOM::exponent2term(l[2]), l[1])));
    end_proc;

//====================================
// Arithmetics / Differential Methods 
//====================================

zero := new(dom, poly(0, Dummy_l, R));

one := new(dom, poly(1, Dummy_l, R));

invertibleCoefficients := R::hasProp(Cat::Field);

characteristic := R::characteristic;

_mult := proc():dom
    local s:DOM_LIST, res1:dom, res2:R, DOM:DOM_DOMAIN, tmp;
    begin
	DOM := dom;
	s := [args(1..args(0))];
	tmp := split(s, l -> domtype(l) = DOM);
	res1 := dom::_mult2(op(tmp[1]));
	if tmp[2] <> [] then
	    res2 := Rmult(op(tmp[2]));
	    return(dom::multcoeffs(res1, res2));
	end_if;
	res1; 
    end_proc; 

_mult2 := proc():dom
    local i:DOM_INT, m2r:DOM_PROC, factMu:DOM_PROC, tmp, e1, e2;
    begin			
	case args(0)
	of 0 do
	    return(dom::one);
	of 1 do 
	    return(args(1));
	of 2 do
	    m2r := (mu -> [(i $ mu[i]) $ i = 1..nX]);
	
	    factMu := proc(mu)
		local k:DOM_INT, m:DOM_INT, res:DOM_INT;
		begin
		    res := 1;
		    m := 1;
			for k from 2 to nops(mu) do
			    if mu[k] = mu[k - 1] then
			    m := m + 1;
			    res := res * m;
			else
			    m := 1;
			end_if;
		    end_for;
		    1 / res;
		end_proc;
	
	    f1 := args(1);
	    f2 := args(2);
	    
// 	    e1 := map(map(dom::poly2list(f1), op, 2), l1 -> ([op(l1, (nX + 1)..nV)]));
// 	    e2 := map(map(dom::poly2list(f2), op, 2), l1 -> ([op(l1, 1..nX)]));
// 	    e1 := map([map(e1, op, i) $ i = 1..nX], l -> max(op(l)));
// 	    e2 := map([map(e2, op, i) $ i = 1..nX], l -> max(op(l)));
// 	    tmp := zip(e1, e2, min);	    	    
// 	    tmp := map(combinat::powerset(m2r(tmp)), 
// 		l2 -> multcoeffs(Pdiff(map(l2, _plus, nX), REP(f1)) *  
// 			Pdiff(l2, REP(f2)), factMu(l2)));

// and now...
// THE LONGEST, MOST COMPLICATED LINE OF CODE IN THE WORLD
// (it contains 10 maps!)  

	    if not Homogenised then
		return(new(dom, _plus(op(map(combinat::subwords(
		    m2r(zip(map([map(map(map(dom::poly2list(f1), op, 2), 
			l1 -> ([op(l1, (nX + 1)..nV)])), op, i) $ i = 1..nX], 
			    l2 -> max(op(l2))), map([map(map(map(dom::poly2list(f2), op, 2), 
				l3 -> ([op(l3, 1..nX)])), op, i) 
				    $ i = 1..nX], l4 -> max(op(l4))), min))), 
					l5 -> multcoeffs(Pdiff(map(l5, _plus, nX), REP(f1)) *  
					    Pdiff(l5, REP(f2)), factMu(l5)))))));
	    else // this one is EVEN LONGER
		return(new(dom, _plus(op(map(combinat::subwords(
		    m2r(zip(map([map(map(map(dom::poly2list(f1), op, 2), 
			l1 -> ([op(l1, (nX + 2)..nV)])), op, i) $ i = 1..nX], 
			    l2 -> max(op(l2))), map([map(map(map(dom::poly2list(f2), op, 2), 
				l3 -> ([op(l3, 2..(nX + 1))])), op, i) 
				    $ i = 1..nX], l4 -> max(op(l4))), min))), 
					l5 -> multcoeffs(Pdiff(map(l5, _plus, (nX + 1)), 
					    REP(f1)) *  Pdiff(map(l5, _plus, 1), REP(f2))
						* poly([[1, [2 * nops(l5)] . [0 $ (2 * nX)]]], 
						    Dummy_l, R), factMu(l5)))))));	
	    end_if;
	otherwise
	    i := args(0) div 2;
	    return(dom::_mult2(dom::_mult2(args(1..i)),
			       dom::_mult2(args((i + 1)..args(0)))));
	end_case;
    end_proc;

mapcoeffs := (x -> new(dom, mapcoeffs(REP(x), args(2..args(0)))));

multcoeffs := (x -> new(dom, multcoeffs(REP(x), args(2))));

Rdivide := if not dom::invertibleCoefficients then FAIL else
    ((x, r) -> dom::multcoeffs(x, Rinvert(r))) end_if;

commutator := ((x, y) -> (dom::_subtract(dom::_mult(x, y),
					 dom::_mult(y, x))));

reduce := proc(f:dom, g:dom, mu:DOM_LIST, t:DOM_LIST)
    local d, tmp;
    begin
	if args(0) = 5 then
	    d := args(5);
	else
	    d := zip(t[2], mu[2], _subtract);
	    if select(d, l -> bool(l < 0)) <> [] then
		error("cannot reduce");
	    end_if;
	end_if;
	tmp := dom::exponent2monomial(d) * g;
	dom::multcoeffs(f, dom::lcoeff(tmp)) - dom::multcoeffs(tmp, t[1]);
    end_proc;
    	
homogenisedComplete := proc(F:Type::ListOf(dom), IDiv:DOM_DOMAIN)
    local HW:DOM_DOMAIN;
    begin
	if (IDiv::hasProp = FAIL
	    or not IDiv::hasProp(Cat::InvolutiveDivision)
	    or IDiv::numVars <> nV)
	then
	    error("incorrect involutive division");
	end_if;
	HW := Dom::HomogenisedWeylAlgebra(dom, genident("_chvar_"));
	map(HW::involutiveComplete(map(F, HW::homogenise), 
	    Dom::LiftedDivision(IDiv)), HW::dehomogenise);
    end_proc;

//================
// body of domain 
//================
		    
begin
    if args(0) < 1 or args(0) > 4 then
	error("wrong number of arguments");
    end_if;
    
    if type(Vars) = "_index" then
	Vars := [op(vars, 1)[i] $ i = 1..(op(Vars, 2))];
    elif not testtype(Vars, Type::ListOf(DOM_IDENT)) then
	error("wrong type of arguments");
    end_if;
    
    if (domtype(R) <> DOM_DOMAIN
	or R::hasProp = FAIL
	or not R::hasProp(Cat::CommutativeRing))
    then
	error("wrong type of arguments");
    end_if;
    
    if (not contains({LexOrder, DegreeOrder, DegInvLexOrder}, Order)
	and domtype(Order) <> Dom::MonomOrdering)
    then
	error("unknown monomial ordering");
    end_if;
    
    if args(0) = 4 then
	HomVar := args(4);
    	if (domtype(HomVar) = DOM_IDENT 
	    and iszero(contains(Vars, HomVar))) 
	then
	    Homogenised := TRUE;
	else	
	    error("wrong type of arguments");
	end_if;
    else
	Homogenised := FALSE;
    end_if;
    
    nX := nops(Vars);
    if not Homogenised then
	nV := 2 * nX;
	Dummy_l := [genident("_wa_") $ i = 1..nV];
	if {op(zip(map(Vars, D), map(Vars, hold(D)), (l1, l2) -> (bool(l1 = l2))))} <> {TRUE} 	
	    then error("invalid variable");
	end_if;
	Var_l := Vars . map(Vars, D);
	zip(Dummy_l, Var_l, (l1, l2) -> (Dummy2Var[l1] := l2));
	zip(Var_l, Dummy_l, (l1, l2) -> (Var2Dummy[l1] := l2));
    else
	nV := 2 * nX + 1;
	Dummy_l := [genident("_wa_") $ i = 1..nV];
	if {op(zip(map(Vars, D), map(Vars, hold(D)), (l1, l2) -> (bool(l1 = l2))))} <> {TRUE} 	
	    then error("invalid variable");
	end_if;
	Var_l := Vars . map(Vars, D);
	Dummy2Var[Dummy_l[1]] := HomVar;
	Var2Dummy[HomVar] := Dummy_l[1];
	zip([op(Dummy_l, 2..nV)], Var_l, (l1, l2) -> (Dummy2Var[l1] := l2));
	zip(Var_l, [op(Dummy_l, 2..nV)], (l1, l2) -> (Var2Dummy[l1] := l2));
	Var_l := [HomVar] . Var_l;
    end_if;
    MO := Dom::MonomOrdering;
    case Order 
    of LexOrder do
	Order := MO(Lex(nV));
	break;
    of DegreeOrder do
	Order := MO(DegLex(nV));
	break;
    of DegInvLexOrder do
	Order := MO(DegInvLexOrder(nV));
	break;
    end_case;
    oM := MO::ordermatrix(Order);
    Pdiff := polylib::Dpoly;
    Rcoerce := R::coerce;
    Rinvert := R::invert;
end_domain;

unalias(REP(x)):




















































