// $Id: ceil.mu,v 1.5 2002/01/04 12:53:44 ccr Exp $

DOM_INTERVAL::ceil := proc(a)
begin
  a := hull(a); // handle unions etc.

  if rhs(a) = RD_INF then infinity;
  // ceil(RD_NINF) cannot be -infinity, since there
  // are finite numbers represented as RD_NINF.
  // We return the smallest integer pari can handle on 32 bit.
  // This takes some while, but it's the best correct answer
  // we can give.
  elif rhs(a) = RD_NINF then /* too slow - _plus(2^i$i=0..2097023) */
        -2^2097023;
  else ceil(rhs(Re(a))+I*rhs(Im(a))); end_if;
end_proc:

DOM_INTERVAL::floor := proc(a)
begin
  a := hull(a); // handle unions etc.

  if lhs(a) = RD_NINF then -infinity;
  // floor(RD_INF) cannot be infinity, since there
  // are finite numbers represented as RD_INF.
  // We return the largest integer pari can handle on 32 bit.
  // This takes some while, but it's the best correct answer
  // we can give.
  elif lhs(a) = RD_INF then  /* too slow _plus(2^i$i=0..2097023) */
        2^2097023;
  else floor(lhs(Re(a))+I*lhs(Im(a))); end_if;
end_proc:

DOM_INTERVAL::trunc := proc(a)
begin
  a := hull(a);

  if 0 in a then 0
  elif op(a, 0) = hold(hull) then
    if a > 0 then DOM_INTERVAL::floor(a)
    else DOM_INTERVAL::ceil(a);
    end_if:
  else DOM_INTERVAL::trunc(Re(a)) + I*DOM_INTERVAL::trunc(Im(a));
  end_if;
end_proc:

DOM_INTERVAL::round := proc(a)
begin
  a := hull(a);

  if lhs(a) = RD_NINF or rhs(a) = RD_INF then
     if lhs(a) = RD_NINF then
        if rhs(a) = RD_INF then 0
        else -infinity
        end_if;
     else
        infinity
     end_if;
  else
     round(lhs(Re(a))/2 + rhs(Re(a))/2
        + I*lhs(Im(a))/2 + I*lhs(Im(a))/2);
  end_if;
end_proc:
