// $Id: iv_sin.mu,v 1.19 2001/10/24 13:55:57 ccr Exp $

// trigonometric functions on DOM_INTERVALs.  These all go
// down to DOM_INTERVAL::internal_sin.

DOM_INTERVAL::sin := proc(a)
//  local sign_fix, move;
begin
  if a::dom <> DOM_INTERVAL then
     if a = {} then return(a); end_if;
     return(FAIL);
  end_if;
  if op(a, 0) = hold(_union) then
     _union(map(op(a), DOM_INTERVAL::sin));
  elif op(a, 0) = hold(hull) then // a is real
     DOM_INTERVAL::sin_internal(a);
  else // a is a complex simple interval
     DOM_INTERVAL::sin_internal(Re(a)) * DOM_INTERVAL::cosh(Im(a))
        + I*(DOM_INTERVAL::cos(Re(a)) * DOM_INTERVAL::sinh(Im(a)));
  end_if;
end:

DOM_INTERVAL::cos := a -> DOM_INTERVAL::sin(a+hull(PI/2)):

DOM_INTERVAL::cot := proc(a)
  local sina, cosa;
begin
  if a::dom <> DOM_INTERVAL then
     if a = {} then return(a); end_if;
     return(FAIL);
  end_if;
  if op(a, 0) = hold(_union) then
     _union(map(op(a), DOM_INTERVAL::cot));
  else
     sina := DOM_INTERVAL::sin(a);
     cosa := DOM_INTERVAL::cos(a);
     cosa/sina;
  end_if;
end_proc:

DOM_INTERVAL::tan := proc(a)
  local sina, cosa;
begin
  if a::dom <> DOM_INTERVAL then
     if a = {} then return(a); end_if;
     return(FAIL);
  end_if;
  if op(a, 0) = hold(_union) then
     _union(map(op(a), DOM_INTERVAL::tan));
  else // simple interval
     sina := DOM_INTERVAL::sin(a);
     cosa := DOM_INTERVAL::cos(a);
     sina/cosa;
  end_if;
end_proc:

// The inverse functions all use the kernel routine DOM_INTERVAL::arctan

DOM_INTERVAL::arctan := proc(a)
begin
  if a::dom <> DOM_INTERVAL then
     if a = {} then return(a); end_if;
     return(FAIL);
  end_if;

  if op(a, 0) = hold(_union) then
     _union(map(op(a), DOM_INTERVAL::arctan));
  elif op(a,0) = FAIL then // complex interval
     // TODO: Improve the overestimation
     I/2 * DOM_INTERVAL::ln(1-I*a)
      - I/2 * DOM_INTERVAL::ln(1+I*a);
  else // real interval
     DOM_INTERVAL::arctan_internal(a);
  end_if;
end_proc:

DOM_INTERVAL::arccot := proc(a)
begin
  if a::dom <> DOM_INTERVAL then
     if a = {} then return(a); end_if;
     return(FAIL);
  end_if;
  DOM_INTERVAL::arctan(1/a);
end_proc:


DOM_INTERVAL::arcsin := proc(a)
begin
  if a::dom <> DOM_INTERVAL then
     if a = {} then return(a); end_if;
     return(FAIL);
  end_if;

  if op(a, 0) = hold(_union) then
     _union(map(op(a), DOM_INTERVAL::arcsin));
  else
     DOM_INTERVAL::arctan(a/sqrt(1-a^2));
  end_if;
end_proc:

DOM_INTERVAL::arccos := x -> (hull(PI/2) - DOM_INTERVAL::arcsin(x)):


// To avoid errors for intervals containing zero:
DOM_INTERVAL::arccsc := x -> DOM_INTERVAL::arcsin(1/x):
DOM_INTERVAL::arcsec := x -> DOM_INTERVAL::arccos(1/x):

DOM_INTERVAL::csc := x -> 1/DOM_INTERVAL::sin(x):
DOM_INTERVAL::sec := x -> 1/DOM_INTERVAL::cos(x):
