/*****************************************************************************
  $Id: combinat.mu,v 1.12 2002/01/15 01:37:17 nthiery Exp $
  File:		lib/LIBFILES/combinat.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
		Authors of the combinat package of MuPAD 2.0.0
  Taken from:	MuPAD 2.0.0, 16/02/2001
  License:	LGPL
*****************************************************************************/

traperror(
    // The traperror avoids an error if the combinat package had
    // already been defined (say from MuPAD's library).

//          if not testtype(combinat, DOM_DOMAIN) then
          domain combinat
          // category inlinedoc::cat;
          end_domain
//          end_if
          ):

combinat::info := "Library 'combinat': combinatorial structures and functions
 This library is a subset of the open-source algebraic
 combinatorics contribution package MuPAD-Combinat 0.2:
 http://mupad-combinat.sourceforge.net.
":

combinat::introduction := "
 This library provides combinatorial functions, as well as tools for
 counting, generating, and manipulating combinatorial structures.

 Online help for subpackages:
 ?combinat::subpackage or combinat::subpackage::doc()
":

combinat::synopsis := "
 Functions:
 - bell(n)			Bell numbers
 - catalan(n)			Catalan numbers
 - stirling1(n)			Stirling numbers of the first kind
 - stirling2(n)			Stirling numbers of the second kind
 - modStirling(n)		Modified Stirling numbers

 Subpackages for combinatorial structures:
 - words			Words over an alphabet
 - subwords			Subwords of a word
 - permutations			Permutations of a list

 - integerListsLexBuilder	Lexicographic generation of lists of integers
				under constraints
 - partitions			Partitions of an integer
 - compositions			Compositions of an integer
 - integerVectors		Integer vectors
 - integerVectorsWeighted 	Integer vectors, with weighted parts

 - dyckWords			Dyck words (words of well balanced parenthesis)
 - nonCrossingPartitions	Non crossing set partitions

 - tableaux			Tableaux

 - subsets			Subsets of a set
 - cartesianProduct		Cartesian product of sets or lists

 Other subpackages:
 - generators			Tools for building and manipulating generators

 Categories:
 - structure			Generic tools for combinatorial structures
 - integerListsLexBuilding	Wrapper category around integerListsLexBuilder
 - inlinedoc			Tools for inline/online package documentation

 Deprecated functions:
 - warnDeprecated(bool)		switch on/off warnings for deprecated functions
 - powerset(s)			powerset of a set
 - cartesian(s1,s2)		cartesian product of s1 and s2
 - composition(n,k)		compositions of n in k parts
 - partitions(n)		partitions of n
":

combinat::interface :=
   { 
    hold(bell), hold(catalan),
    hold(stirling1), hold(stirling2), hold(modStirling),
//    hold(structure), hold(integerListsLexBuilding),
    hold(generators), 
    hold(words), hold(subwords), hold(dyckWords), hold(permutations),
//    hold(integerListsLexBuilder), 
    hold(partitions), hold(compositions), hold(integerVectors),
    hold(integerVectorsWeighted),
    hold(tableaux),
    hold(subsets), 
    hold(cartesianProduct),
    hold(warnDeprecated)
    // old stuff
    //hold(powerset), hold(cartesian),
    //hold(numberOfSetPartitionColorings)
   }:

alias(path = pathname("COMBINAT")):

/*****************************************************************************
   Utilities that should move into the MuPAD library
*****************************************************************************/

combinat::getOptions :=
    loadproc(combinat::getOptions, path, "getOptions"):

/*****************************************************************************
   Classical integer sequences
*****************************************************************************/

combinat::bell :=
    loadproc(combinat::bell, path, "bell"):
combinat::stirling1 :=
    loadproc(combinat::stirling1, path, "stirling1"):
combinat::stirling2 :=
    loadproc(combinat::stirling2, path, "stirling2"):
combinat::modStirling :=
    loadproc(combinat::modStirling, path, "modStirling"):
combinat::catalan :=
    loadproc(combinat::catalan, path, "catalan"):

/*****************************************************************************
   Categories for combinatorial objects
*****************************************************************************/

combinat::structure :=
    loadproc(combinat::structure, path, "structure"):
combinat::integerListsLexBuilding :=
    loadproc(combinat::integerListsLexBuilding, path, "integerListsLexBuilding"):

/*****************************************************************************
   Sub packages for classical combinatorial objects
*****************************************************************************/

combinat::words :=
    loadproc(combinat::words, path, "words"):
combinat::subwords :=
    loadproc(combinat::subwords, path, "subwords"):
combinat::permutations :=
    loadproc(combinat::permutations, path, "permutations"):

combinat::dyckWords :=
    loadproc(combinat::dyckWords, path, "dyckWords"):
combinat::nonCrossingPartitions :=
    loadproc(combinat::nonCrossingPartitions, path, "nonCrossingPartitions"):

combinat::integerListsLexBuilder :=
    loadproc(combinat::integerListsLexBuilder, path, "integerListsLexBuilder"):
combinat::partitions :=
    loadproc(combinat::partitions, path, "partitions"):
combinat::compositions :=
    loadproc(combinat::compositions, path, "compositions"):
combinat::integerVectors :=
    loadproc(combinat::integerVectors, path, "integerVectors"):
combinat::integerVectorsWeighted :=
    loadproc(combinat::integerVectorsWeighted, path, "integerVectorsWeighted"):

combinat::tableaux :=
    loadproc(combinat::tableaux, path, "tableaux"):

combinat::subsets :=
    loadproc(combinat::subsets, path, "subsets"):
combinat::cartesianProduct :=
    loadproc(combinat::cartesianProduct, path, "cartesianProduct"):

/*****************************************************************************
   Sub package for generators
*****************************************************************************/

combinat::generators :=
    loadproc(combinat::generators, path, "generators"):

/*****************************************************************************
   Aliases for backward compatibility
*****************************************************************************/

combinat::warnDeprecated :=
    proc()
        local oldvalue;
    begin
        if args(0) = 0 then
            return(FALSE)
        else
            if testargs() and not testtype(args(1), DOM_BOOL) then
                error("expecting TRUE or FALSE");
            end_if;
            oldvalue := combinat::warnDeprecated();
            sysassign(combinat::warnDeprecated(), args(1));
        end_if;
        return(oldvalue);
    end_proc:

combinat::permute := 
    proc()
    begin
        if combinat::warnDeprecated() then
            warning("combinat::permute(l) is obsolete.
Please use 'combinat::permutations::list(l)' instead.");
        end_if;
        combinat::permutations::list(args());
    end_proc:

combinat::choose := 
    proc()
    begin
        if combinat::warnDeprecated() then
            warning("combinat::choose(s,k) is obsolete.
Please use 'combinat::subsets::list(s,k)' instead.");
        end_if;
        op(combinat::subsets::list(args()))
    end_proc:

combinat::powerset	:= 
    proc(set: Type::Union(Type::NonNegInt, DOM_SET, DOM_LIST))
        /*
	   If set is a list, it's interpreted as a multiset
           The algorithm is suboptimal in this case !!!
        */
    begin
        if testargs() then
            if args(0) <> 1 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;
        if combinat::warnDeprecated() then
            warning("combinat::powerset is obsolete.
Please use 'combinat::subsets' and 'combinat::subwords' instead.");
        end_if;
        if domtype(set)=DOM_LIST then
            return({op(combinat::subwords::list(sort(set)))})
        end_if;
        return({op(combinat::subsets::list(set))});
    end_proc:

combinat::composition	:=
    proc(n,k)
    begin
        if combinat::warnDeprecated() then
            warning("combinat::composition(n,k) is obsolete.
Please use 'combinat::integerVector::list(n,k, MinPart=1)' instead.");
        end_if;
        revert(combinat::integerVectors::list(n,k,hold(MinPart)=1));
    end_proc:

combinat::cartesian := 
     proc()
     begin
        if combinat::warnDeprecated() then
        warning("combinat::cartesian(s1,s2,...) is obsolete.
Please use 'combinat::cartesianProduct::list(s1,s2,...)' instead.");
        end_if;
         {op(combinat::cartesianProduct::list(args()))}
     end_proc:

combinat::partitions:	// without this MuPAD get confused with autoloading
combinat::partitions::new :=
    proc(n: Type::NonNegInt)
    begin
        if combinat::warnDeprecated() then
            warning("combinat::partitions(n) is obsolete.
Please use 'combinat::partitions::count(n)' instead.");
        end_if;
        combinat::partitions::count(n);
    end_proc:

/*****************************************************************************
   Stuff not yet reorganized!!!
*****************************************************************************/

combinat::numberOfSetPartitionColorings :=
    loadproc(combinat::numberOfSetPartitionColorings, path, "nbspcols"):

stdlib::LoadlibState["combinat"] := TRUE:

null():
