// $Date: 2002/02/18 16:20:01 $ $Author: tonner $ $Revision: 1.127 $ 
// tonner, 23/06/98 

/*++
output::Content -- Domain of Content markup
++*/

output:

/** Creating Content trees
 *
 *   This domain implements the infrastructure to create Content
 *   trees
 */
domain Content
  category output::CatContent;
  inherits Dom::BaseDomain;
  
  Name := "Content";

  create :=
  proc(typString : DOM_STRING, typInfo : DOM_STRING, attributes : DOM_STRING)
    local typ;
  begin
    if testargs() then
      if args(0) < 3 then
        error("internal output error")
      end_if;
      if map({args(4..args(0))}, domtype) minus
        {DOM_STRING, DOM_INT, DOM_IDENT, DOM_FLOAT, Content} <> {} then
        error("children may only be idents, numbers, strings or elements of Content");
      end_if;
    end_if;

    typ := dom::ContentTable[typString];
    if type(typ) = "_index" then
      error("No template \"".typString."\" defined");
    end_if;
    
    new(dom, args())
  end_proc;

  // access element entries
  typString := proc(c)
  begin
    if contains({DOM_INT,DOM_STRING,DOM_FLOAT,DOM_IDENT},domtype(c)) then
        expr2text(domtype(c))
    else
        extop(c, 1)
    end
  end;

  attributes := proc(c)
  begin
    if contains({DOM_INT,DOM_STRING,DOM_FLOAT,DOM_IDENT},domtype(c)) then 
        ""
    else
        extop(c, 2)
    end
  end;

  infos := proc(c)
  begin
    if contains({DOM_INT,DOM_STRING,DOM_FLOAT,DOM_IDENT},domtype(c)) then 
      ""
    else
      extop(c, 3)
    end
  end;
    
  op := proc(c,i=-1)
               begin
                 if i = -1 then
                   extop(c, 4..extnops(c))
                 elif testtype(i, "_range") then
                   extop(c, 3+op(i,1)..3+op(i,2))
                 else
                   extop(c, 3+i)
                 end_if
               end_proc;

  subsop := proc(c,eq=-1)
            begin
              if type(eq) = "_equal" then
                if domtype(op(eq, 1)) = DOM_INT then
                  extsubsop(c, 3+op(eq, 1) = op(eq, 2))
                elif domtype(op(eq, 1)) = DOM_LIST then
                  if nops(op(eq, 1)) > 1 then
                    extsubsop(c, 3+op(eq, [1, 1]) =
                              dom::subsop(dom::op(c, op(eq, [1, 1])), subsop(eq, [1, 1] = null())))
                  elif nops(op(eq, 1)) = 1 then
                    extsubsop(c, 3+op(eq, [1, 1]) = op(eq, 2))
                  else // empty list - replace whole expression
                    op(eq, 2)
                  end_if
                end_if
              else
                c
              end_if
            end_proc;

  nops := c -> extnops(c) - 3;

  map := (x, y) -> dom::create(extop(x, 1..3),
                               map(extop(x, 4..extnops(x)), y));
  
  Content2tree :=
  proc(x)
    local res;
  begin
    if domtype(x) <> dom then [x]
    else
      res := [dom::typString(x).
       ", info: ".expr2text(dom::infos(x)).", attrib : ".
       expr2text(dom::attributes(x))];
      if nops(x)=0 then res
      else append(res, op(map([op(x)], dom::Content2tree)));
      end_if;
    end_if
  end_proc;

  print := 
  proc(x)
  begin
    output::tree(dom::Content2tree(x), {hold(Small)});
  end_proc;
    
  /* output in XML notation */
  xml := proc(arg1) begin dom::do_print_xml(arg1) end_proc;
  
  do_print_xml :=
  proc(arg1)
    name output::print;
    local DOM, process, Print,
          outString:DOM_STRING, indentLevel:DOM_INT;
  begin
    indentLevel := 0;
    outString := "";
    DOM := dom;
    
// extract args
  Print :=
  proc(d)
    local attributes, infos, Indent, Empty;
  begin
    Indent := {};
    Empty := { "int" };
    if contains({DOM_STRING, DOM_INT, DOM_IDENT, DOM_FLOAT, output::Indent} , domtype(d)) then
       case domtype(d) 
         of DOM_STRING do
	return(d, DOM::newline());
            return("<cstring>".d."</cstring>", DOM::newline())
         of DOM_INT do
            return("<cn type=\"integer\"> ".expr2text(d)." </cn>", DOM::newline())
         of DOM_FLOAT do
            return("<cn> ".expr2text(d)." </cn>", DOM::newline())
         of DOM_IDENT do
            return("<ci> ".expr2text(d)." </ci>", DOM::newline())
	 of output::Indent do
            return(d)
       end_case;
    end_if;

    attributes := DOM::attributes(d);
    infos := DOM::infos(d);
    "<", DOM::typString(d),
    (if contains(Empty, DOM::typString(d)) then "/>"
     else
       ">",
     DOM::newline(), 
     DOM::indent(
       (if attributes <> "" then
          "<annotation encoding=\"MuPAD-attributes\">", 
          DOM::indent(attributes), DOM::newline(),
          "</annotation>"
        else
          null()
        end_if),
       (if infos <> "" then
          "<annotation encoding=\"MuPAD-infos\">", 
          DOM::indent(infos),  DOM::newline(),
          "</annotation>"
        else
          null()
        end_if),
       DOM::op(d)), DOM::newline(),
     "</", DOM::typString(d),">"
     end_if),
    DOM::newline()
  end_proc;

  process :=
    proc(x)
      name  output::print::process;
      local Oseq, i, entries;
    begin
      Oseq := [Print(x)];
// print("process", x, domtype(x), Print, Oseq);
      for i in Oseq do
	if domtype(i) = DOM_STRING then
	  outString := outString.DOM::translate(i);
	elif domtype(i) = output::Indent then
	  if output::Indent::isNewline(i) then
	    outString := _concat(outString, "\n", ("  "$indentLevel));
	  else
	    indentLevel := indentLevel+1;
	    outString := outString."  ";
	    if (entries := output::Indent::entries(i)) <> null() then
	      map(entries,process);
	    end_if;
	    indentLevel := indentLevel-1;
	  end_if;
	elif domtype(i) = dom then
	  process(i);  
	else
// print("process", i, domtype(i));
	  error("output element expected");
	end_if;
      end_for;
    end_proc;

    process(dom::enclose(arg1));
    outString;
  end_proc;

  translate := id;

  // formatting of source output
  newline := proc() option remember; begin output::Indent("newline") end_proc;
  indent :=  () -> output::Indent("indent", args());

  expr :=
  proc(c)
    local i, cType, attributes;
  begin
    cType := domtype(c);

    // here we need a special conversion from backquoted identifiers
    // representing MML-constants to MuPAD constants
    if cType = DOM_IDENT and contains(dom::MML2ident, c) then
       return(dom::MML2ident[c]);
    end_if;

    if not contains({DOM_STRING, DOM_INT, DOM_IDENT, DOM_FLOAT}, cType) and domtype(c) <> dom then
      error("wrong type")
    end_if;

    if nops(dom::ContentTable[dom::typString(c)]) > 1 then
      attributes := table(text2expr(dom::attributes(c)));
      return(dom::ContentTable[dom::typString(c)][2](c, attributes, dom::infos(c)));
    // --- frankp: otherwise we would enter into an infinite loop, e.g., if c 
    // ---         is of domain type DOM_INT, and the constructor is returns
    // ---         an object of a library domain.

    // inserted  'and domtype(c) <> dom' because Franks patch broke
    // expr(Content(x!))         tonner 
    elif nops(c) = 1 and domtype(c) <> dom then
      subs(hold(XxX)(c),  hold(XxX)=op(dom::ContentTable[dom::typString(c)], 1), Unsimplified)
    // ----
    else
      subs(hold(XxX)(dom::expr(op(c,i)) $ i = 1 .. nops(c)),  hold(XxX)=op(dom::ContentTable[dom::typString(c)], 1), Unsimplified)
    end_if;
  end_proc;


/** Tables for translations from MML-constants to MuPAD constants
    and vice versa.
  */ 
  MML2ident := table();

/** Table for translations from MML-constants to MuPAD constants
  */ 
  ident2MML := table();


/** fills the tables MML2ident and ident2MML in the init section
  */ 
  translateIdents :=
  proc(equation) //  : Type::Equation(DOM_IDENT) falsch!!
  begin
    dom::MML2ident[op(equation, 1)] := op(equation, 2);
    dom::ident2MML[op(equation, 2)] := op(equation, 1);
  end_proc;



/**  Table of all known Content types
 *
 *   Format of the table entries:
 *   
 *   index = [mupadType, backTranslate (optional)]
 *
 *   index              : string which adresses the template in MuPAD
 *
 *   mupadType          : type of the MuPAD object which is represented
 *                        by this Content
 *
 *   backTranslate (opt): MuPAD function which translates a Content 
 *                        element of the type index into a valid MuPAD
 *                        object
 */
  ContentTable := table();

/** Table of all known Content Layouts
 *
 */
  LayoutTable := table();

    // everthing which was written with the last call of writeLayoutSetup
  LastLayoutTable := {};
 
    // with the only option Complete a complete Setup file is written.
    // otherwise only a patch 
  writeLayoutSetup :=
  proc(opt=null())
    local oldFid, fid, fName, CL, templates, update, cl,
          openFile, closeFile, DOM;
  begin
    if testargs() then
      if args(0) > 1 then
        error("wrong number of arguments.");
      elif args(0) = 1 and opt <> hold(Complete) then
        error("unknown option, 'Complete' expected")
      end_if;
    end_if;

    DOM := dom;
    CL := ContentLayout;  // shortcut
    fName := "";  // will be replaced by openFile, iff called
    
    // opens a new temporary setup file for writing
    openFile :=
    proc()
    begin
      oldFid := CL::OutputFile;
      fid := fopen(Text, TempFile, Write);
      fName := fname(fid);
      // get corresponding filename and insert it into a toBeDeleted
      // set for Pref::callOnExit
      sysassign(stdlib::TempFiles, stdlib::TempFiles union {fName});
      sysassign(CL::OutputFile, fid);
    end;

    // closes the output file, iff one was opened
    // and informs the clients about the new setup File
    closeFile :=
    proc()
    begin
      if fName <> "" then
        // openFile was called, close the file and inform clients in cl
        fclose(fid);
        sysassign(CL::OutputFile, oldFid);
        map(cl, x -> traperror(clientCall(x, "templateFile", fName)));
      end_if;
    end_proc;
    
    templates := {map(op(dom::LayoutTable), op, 2)};
    if opt = hold(Complete) then
      // write complete setup file and inform the current client
      openFile();
      CL::xml(CL::mathsetup(op(templates)));
      cl := [clientObject()];
    else
      update := templates minus dom::LastLayoutTable;
      if update <> {} then
       // write update setup file and inform all clients
       openFile();
        CL::xml(CL::mathsetup(op(update)));
        cl := clientObject(All);
      else
        // update setup is empty, nothing to do
        cl := [];
      end_if;
    end_if;
    dom::LastLayoutTable := templates;
    closeFile();
    delete cl;
    null()
  end_proc;
    
/* Define new entries in Content::ContentTable */
  set :=
  proc(NAME : DOM_STRING,
       layout,
       mupadType,
       backTranslate = null() : DOM_PROC)
    option hold;
    local ltype;
  begin
    if testargs() then
      ltype := domtype(context(layout));
      if not contains({DOM_FAIL, ContentLayout}, ltype) then
        print(args());
        error("wrong type of 2. argument")
      end_if;
    end_if;
    dom::ContentTable[NAME] := [mupadType, backTranslate];
    if layout <> FAIL then
      dom::LayoutTable[NAME] := context(layout);
    end_if;
  end_proc;
  
  Ccn:=       // Content Number
  proc() 
    local num; 
  begin 
    num := args(args(0));
    if testtype(num, Type::Real) and num < 0 then 
      dom(hold(_negate)(-num))
    else
      num
    end_if
  end_proc;

  Cci:= () -> if contains(dom::ident2MML, args()) then
		 dom::Ccn({type="constant"}, dom::ident2MML[args()]);
	      else
                 args();      // Content Identifier
              end_if;

  /**  Capply - the MathML apply construct
   *  The implementation of the MathML apply construct for MuPAD Content
   *  output is quite complicated, but makes it easy to define the Content
   *  slot for MuPADs expression types.  E.g. 
   */
  Capply:=
  proc()
  begin
    if testargs() and args(1) = FAIL then
       error("trying to apply undefined entity");
    end_if;
    args(1)(args(2..args(0)))
  end_proc;   

  Cfn := x -> dom(x);      // user-defined function
  Cinterval := // interval constructor
     proc(x, left, right)
     begin
       if domtype(x) = DOM_INTERVAL then
         // kernel interval
         return(dom::create("interval", "", "", left, right));
       end_if;

       // form here we handle library intervals
       if op(x,[1,2]) = "open" then
         dom::create("intervaloo", "", "", left, right);
       elif op(x,[1,2]) = "open-closed" then
         dom::create("intervaloc", "", "", left, right);
       elif op(x,[1,2]) = "closed-open" then
         dom::create("intervalco", "", "", left, right);
       else
         dom::create("intervalcc", "", "", left, right);
       end_if;
     end_proc;

  Cinverse :=    // functional inverse
          () -> dom::create("inverse", "", "", args());
  Ccondition // properties of an identifier
	 := id;
//  Cdeclare;   // declaration
  
//  Csep;       // separator
  Ccompose :=   // composition operator
          () -> dom::create("compose", "", "", args());
  Cident  :=    () -> dom::create("ident", "", "", args());// identity function

  // Arithmetic, Algebra and Logic
  Cquotient :=  () -> dom::create("quotient", "", "", args());
  Cexp :=       data -> if data = 1 then dom::Cexponentiale() else 
                         dom::create("exp", "", "", args()) end_if;   
  Cfactorial := () -> dom::create("factorial", "", "", args());
  Cdivide :=    () -> dom::create("divide", "", "", args());
  Cmax :=  "min";//     () -> dom::create("max", "", "", args());
  Cmin :=  "max";//     () -> dom::create("min", "", "", args());
  Cminus :=     () -> dom::create("plus", "Signs=\"-\"", "", args());
  //kd: Cminus :=     () -> dom::create("negate", "", "", args());
  Cplus :=
  proc()
    local DOM, list, Signs, s, i;
  begin
    DOM := dom;
    /* here we have to translate all operands of type
       "minus" into out special sign encoding for _plus and
       additionally we flat out nested "plus" templates */
    list := map([args()],
                contents ->
                if domtype(contents) = DOM then
                  if DOM::typString(contents) = "plus" and
                    DOM::attributes(contents) = "Signs=\"-\"" or
                    DOM::typString(contents) = "negate" then
                    [op(contents), "-"]
                  elif DOM::typString(contents) = "plus" then
                    s := DOM::attributes(contents);
                    [op(contents, i), substring(s, 6+i)] $
                        i = 1..nops(contents)
                  else
                    [contents, "+"]
                  end_if;
                elif contains({DOM_INT, DOM_RAT, DOM_FLOAT},
                            domtype(contents)) and contents < 0 then
                  [-contents, "-"]
                else
                  [contents, "+"]
                end_if);
    Signs := _concat(op(map(list, op, 2)));
    list := map(list, op, 1);
    if Signs = "+" or nops(list) = 0 then
      dom::create("apply", "", "", hold(_plus), op(list))
    else
      dom::create("plus", "Signs=".expr2text(Signs), "",
                  op(list));
    end_if;
  end_proc;
    // Csubtract will always create a "plus"-expression and if one of
    // the operands is already "plus", automatical flattening will be done
  Csubtract :=  proc(op1, op2)
                  local arg1, arg2, signs1, signs2, s, toggleSigns;
                begin
                  toggleSigns := proc(str)
                                   local i;
                                 begin
                                   _concat(map(str[i] $ i=0..length(str)-1,
                                               x -> if x = "+" then "-"
                                                    else "+" end))
                                 end_proc;
                  
                  if dom::typString(op1) = "plus" then
                    s := dom::attributes(op1);
                    signs1 := substring(s, 7.. length(s)-2);
                    arg1 := op(op1);
                  else
                    signs1 := "+"; arg1 := op1;
                  end;
                  if dom::typString(op2) = "plus" then
                    s := dom::attributes(op2);
                    signs2 := toggleSigns(substring(s, 7.. length(s)-2));
                    arg2 := op(op2);
                  else
                    signs2 := "-"; arg2 := op2;
                  end;
                  dom::create("plus", "Signs=".expr2text(signs1.signs2),
                              "", arg1, arg2);
                end_proc;
  Cpower :=     () -> dom::create("power", "", "", args());
  Crem   :=     () -> dom::create("rem", "", "", args());
  Ctimes :=     () -> dom::create("times", "", "", args());
  Croot :=      () -> dom::create("root", "", "", args());
  Csqrt :=      () -> dom::create("sqrt", "", "", args());
//  Cgcd;
  Cand :=       () -> dom::create("and", "", "", args());
  Cor :=        () -> dom::create("or", "", "", args());
  Cxor :=       () -> dom::create("xor", "", "", args());
  Cnot :=       () -> dom::create("not", "", "", args());
  Cimplies :=       () -> dom::create("implies", "", "", args());
  Cequiv :=       () -> dom::create("equiv", "", "", args());
    // Cforall;  Cexists;
  Cabs :=       () -> dom::create("abs", "", "", args());
  Cconjugate :=       () -> dom::create("conjugate", "", "", args());

  // Relations
  Ceq  :=     () -> dom::create("eq", "", "", args());
  Cneq :=    () -> dom::create("neq", "", "", args());
  Cgt  :=    () -> dom::create("gt", "", "", args());
  Clt  := () -> dom::create("lt", "", "", args());
  Cgeq := () -> dom::create("geq", "", "", args());
  Cleq := () -> dom::create("leq", "", "", args());

  // Calculus
  Cln := () -> "ln";
  Clog := () -> dom::create("log", "", "", args(2), args(1));
  Cint := proc()
          begin
            if args(0) = 4 then
              // definite integral
              dom::create("definiteint", "", "", args(4), args(1), args(2..3))
            else
              // indefinite integral
              dom::create("int", "", "", args(2), args(1))
            end_if
          end_proc;
//  Cdiff;  
  Cpartialdiff := () ->
      if args(0) = 2 then
        dom::create("partialdiff", "", "", args());
      else
        dom::create("partialdiff1", "", "", args());
      end_if;
  Clowlimit := () -> args();
  Cuplimit  := () -> args();
  Cbvar := () -> args();
  Cdegree := () -> args();
  Cgrad :=  () -> dom::create("grad", "", "", args());
  Cdivergence := () -> dom::create("divergence", "", "", args());
  Ccurl :=  () -> dom::create("curl", "", "", args());
  Claplacian :=  () -> dom::create("laplacian", "", "", args());

  // Constants
  Cintegers := () -> dom::create("integers", "", "", args());
  Creals    := () -> dom::create("reals", "", "", args());
  Crationals := () -> dom::create("rationals", "", "", args());
  Cnaturalnumbers := () -> dom::create("naturalnumbers", "", "", args());
  Ccomplexes := () -> dom::create("complexes", "", "", args());
  Cprimes := () -> dom::create("primes", "", "", args());
  Cexponentiale := () -> dom::create("exponentiale", "", "", args());
  Cimaginaryi := () -> dom::create("imaginaryi", "", "", args());
  Cnotanumber := () -> dom::create("notanumber", "", "", args());
  Ctrue := () -> dom::create("true", "", "", args());
  Cfalse := () -> dom::create("false", "", "", args());
  Cunknown := () -> dom::create("unknown", "", "", args());
  Cemptyset := () -> dom::create("emptyset", "", "", args());
  Ceulergamma := () -> dom::create("eulergamma", "", "", args());
  Cinfinity := () -> dom::create("infinity", "", "", args());

    
  // Theory of sets
  Cset := () -> if args(0) = 0 then dom::Cemptyset() else 
              dom::create("set", "", "", args()) end_if;
  Clist := () -> dom::create("list", "", "", args());
  Cunion := () -> dom::create("union", "", "", args());
  CUnion := () -> dom::create("Union", "", "", args());
  Cintersect := () -> dom::create("intersect", "", "", args());
  Cin := () -> dom::create("in", "", "", args()); 
  Cnotin := () -> dom::create("notin", "", "", args());
  Csubset := () -> dom::create("subset", "", "", args());
  Cprsubset := () -> dom::create("prsubset", "", "", args());
  Cnotsubset := () -> dom::create("notsubset", "", "", args());
  Cnotprsubset := () -> dom::create("notprsubset", "", "", args());
  Csetdiff := () -> dom::create("setdiff", "", "", args());
  Ccard := () -> dom::create("card", "", "", args());

  // Sequences and Series
  Csum := proc()
          begin
            if args(0) = 4 then
              dom::create("definitesum", "", "", args(4), args(1..3))
            elif args(0) = 3 then
              dom::create("sumrootof", "", "", args(3), args(1..2))
            else
              dom::create("sum", "", "", args(2), args(1))
            end_if
          end_proc;
                  
  Cproduct := proc()
          begin
            if args(0) = 4 then
              dom::create("definiteproduct", "", "", args(4), args(1..3))
            else
              dom::create("product", "", "", args(2), args(1))
            end_if
          end_proc;

  Climit := () -> if domtype(args(2)) = DOM_STRING then
                    if args(2) = "Direction = Right" then
                      dom::create("limitRight", args(2), "", args(4), args(1), args(3) );
                    else  // Left
                      dom::create("limitLeft", args(2), "", args(4), args(1), args(3) );
                    end_if;
                  else 
                    dom::create("limit", "", "", args(3), args(1..2));
                  end_if;

  Ctendsto := proc(x: "_equal")
		option escape;
                local direction;
              begin
		if contains({"above", "below", "two-sided"}, op(x,2)) then
		  if op(x,2) = "above" then 
		    direction := "Direction = Right"
                  elif op(x,2) = "below" then 
		    direction := "Direction = Left"
 	 	  else
		    direction := null()
		  end_if;
		  proc(bvar, tendsto)
		  begin
		    direction, tendsto
		  end_proc
		else
		  id
		end_if;
              end_proc;

  // Trigonometry
  Csin    := () -> "sin";
  Ccos    := () -> "cos";
  Ctan    := () -> "tan";
  Csec    := () -> "sec";
  Ccsc    := () -> "csc";
  Ccot    := () -> "cot";
  Csinh   := () -> "sinh";
  Ccosh   := () -> "cosh";
  Ctanh   := () -> "tanh";
  Csech   := () -> "sech";
  Ccsch   := () -> "csch";
  Ccoth   := () -> "coth";
  Carcsin := () -> "arcsin";
  Carccos := () -> "arccos";  
  Carctan := () -> "arctan";

  // Statistics
  Cmean := () -> dom::create("mean", "", "", args());
  CquadraticMean := () -> dom::create("quadraticMean", "", "", args());
  Csdev := () -> dom::create("sdev", "", "", args());
    Cvariance := () -> dom::create("variance", "", "", args());
    Cmedian := () -> dom::create("median", "", "", args());
    Cmode := () -> dom::create("mode", "", "", args());
    Cmoment := () -> dom::create("moment", "", "", args(3), args(1), args(2));
    Cmomentabout := () -> args();

  // Linear Algebra
//  Cvector;
  Cmatrix :=
  proc()
    local cols, rows; 
  begin
    cols := nops(args(1)); rows := args(0);
    dom::create("matrix", expr2text(hold(Rows)=expr2text(rows), hold(Cols)=expr2text(cols)), "Dom::Matrix()", op(_concat(args())));
  end_proc;
  Cmatrixrow := () -> [args()];
  Cdeterminant := () -> dom::create("determinant", "", "", args());  
  Ctranspose := () -> dom::create("transpose", "", "", args());
//  Cselect;
  Cscalarproduct := () -> dom::create("scalarproduct", "", "", args()); 
  Cvectorproduct := () -> dom::create("vectorproduct", "", "", args()); 
// Couterproduct;

  // Semantic Mapping Elements
//  Cannotation;  Csemantics;  CannotationXml;

  Cexprseq := () -> dom::create("seq", "", "", args());
  Clambda := () -> dom::create("lambda", "", "", args());
  Cfnest := () -> dom::create("fnest", "", "", args());
  Cceil := () -> dom::create("ceil", "", "", args());
  CbesselI := () -> dom::create("besselI", "", "", args(2), args(1));
  CbesselJ := () -> dom::create("besselJ", "", "", args(2), args(1));
  CbesselK := () -> dom::create("besselK", "", "", args(2), args(1));
  CbesselY := () -> dom::create("besselY", "", "", args(2), args(1));
  CD := () -> if args(0) = 1 then
                dom::create("D", "", "", args())
              else
                dom::create("Dpara", "", "", args())
              end_if;
  Cfloor := () -> dom::create("floor", "", "", args());
  Cslot :=  () -> dom::create("slot", "", "", args());
  Cdilog := () -> dom::create("dilog", "", "", args());
  Cpolylog := () -> dom::create("polylog", "", "", args());
  Cnorm := () -> if args(0) = 1 then
                   dom::create("infinitynorm", "", "", args());
                 elif args(2) = hold(Infinity) then
                   dom::create("infinitynorm", "", "", args(1));
                 elif args(2) = 1 then
                   dom::create("onenorm", "", "", args(1));
                 elif args(2) = hold(Frobenius) then
                   dom::create("norm", "", "", args(1));
                 else
                   dom::create("othernorm", "", "", args());
                 end_if;
  Cindex := () -> dom::create("index", "", "", args());
  Cbinomial := () -> dom::create("binomial", "", "", args()); 
  Crange := () -> dom::create("range", "", "", args()); 
  Cimageset := () -> dom::create("imageset", "", "", args());
  Cpiecewise :=  () -> dom::create("piecewise",  expr2text(hold(Rows)=expr2text(args(0) div 2), hold(Cols)="2"), "", args());
  Ctable := () -> dom::create("table",  expr2text(hold(Rows)=expr2text(args(0) div 2), hold(Cols)="2"), "", args());

  // equation system;  will be translated into a set of equations
  CEQS := () -> dom::create("EQS",
                            expr2text(hold(Rows)=expr2text(args(0) div 2),
                                      hold(Cols)="2"),
                            "", args());
  Ctableau :=
  proc()
    local options; 
  begin
    if args(0)>0 and domtype(args(1)) = DOM_SET then
      options := table(op(args(1)));
    else
      error("Attributes missing");
    end;
    
    dom::create("tableau", expr2text(op(options)), "", args(2..args(0)));
  end_proc;
  Cassign :=     () -> dom::create("assign", "", "", args());
  Crow :=        () -> dom::create("row", "", "", args());
  Cconcat :=     () -> dom::create("concat", "", "", args());
  Cfenced :=     () -> dom::create("fenced", "", "", args());
  Cgamma :=      () -> dom::create("gamma", "", "", args());
  Cigamma :=      () -> dom::create("igamma", "", "", args());
  end_domain:


alias(CL = ContentLayout):
alias(OP = output::Priority):

Content::set("plus", FAIL,  // no layout for _plus, hard-coded
             _plus, 
             proc(plusExpr, attributes : DOM_TABLE)
               name convert_plus;
               local i, signs;
               save XxX;
             begin
               signs := attributes[Signs];
               if signs = "-" then
                 _negate(Content::expr(op(plusExpr, 1)))
               else
                 subsop(XxX((Content::expr(op(plusExpr, i));
                             if signs[i-1] = "-" then
                               generate::negate(%)
                             else
                               %
                             end_if) $ i = 1 .. nops(plusExpr)),
                        0=hold(_plus), Unsimplified);
               end_if;
             end_proc):
Content::set("subtract",
             CL::content("id"="subtract", "keywd"="Subtraction",
                        "accel"="-", "arity"="2", "prio"=OP::Plus,
                        CL::layout(CL::mrow(CL::oper("idx"="0",
                                                     "prio"=OP::Plus),
                                            CL::mo("form"="infix", "-"),
                                            CL::oper("idx"="1",
                                                     "prio"=OP::Plus)
                                            ))),
             _subtract):  /* ! keine Simplifizierung */
Content::set("negate", FAIL, _power,
             obj -> _negate(Content::expr(op(obj, 1)))):
Content::set("times",
             CL::content("id"="times", "keywd"="Multiplication",
                         "accel"="*", "arity"="2*", "prio"=OP::Mult,
                        CL::layout(
                                   CL::oprow(CL::oper("prio"=OP::Mult),
                                             CL::mo("form"="infix",
                                                 "&CenterDot;")))),
             _mult):  /* ! keine Simplifizierung */
Content::set("divide",
             CL::content("id"="divide", "keywd"="Division", "accel"="/",
                        "arity"="2", "prio"=OP::Mult+1,
                        CL::layout(CL::mfrac(CL::oper("idx"="0"),
                                             CL::oper("idx"="1")))),
             _divide):  /* ! keine Simplifizierung */
Content::set("power",
             CL::content("id"="power", "keywd"="Exponentiation",
                         "accel"="^", "arity"="2", "prio"=OP::Power,
                         CL::layout(CL::msup(CL::oper("idx"="0","prio"=OP::Power),
                                             CL::oper("idx"="1")))),
             _power, proc(x)
                       local expo, base;
                     begin
                       base := Content::expr(op(x, 1));
                       expo := Content::expr(op(x, 2));
                       if base = exp(1) then
                         exp(expo)
                       else
                         hold(_power)(base, expo)
                       end;
                     end):
Content::set("DOM_INT", FAIL, DOM_INT, x -> x):
Content::set("DOM_IDENT", FAIL, DOM_IDENT, x -> x):
Content::set("DOM_FLOAT", FAIL, DOM_FLOAT, x -> x):
Content::set("DOM_STRING", FAIL, DOM_STRING, x -> x):
Content::set("list",
             CL::content("id"="list", "keywd"="List", "accel"="[",
                        "arity"="0*", "dummies"="1", "flatten-seq"="0,-1",
                    CL::layout(
                          CL::mrow(
                              CL::mo("form"="prefix", "["),
                              CL::oprow(CL::oper("parens"="true"),
                                        CL::mo("form"="infix", ",")),
                              CL::mo("form"="postfix", "]")))),
             DOM_LIST,
             proc(x)
               local i;
             begin
               [Content::expr(op(x, i)) $ i = 1 ..nops(x)]
             end_proc):
Content::set("set",
             CL::content("id"="set", "keywd"="Set", "accel"="{",
                        "arity"="0*", "dummies"="1", "flatten-seq"="0,-1",
                    CL::layout(
                          CL::mrow(
                              CL::mo("form"="prefix", "{"),
                              CL::oprow(CL::oper("parens"="true"),
                                        CL::mo("form"="infix", ",")),
                              CL::mo("form"="postfix", "}")))),
             DOM_SET,
             proc(x)
               local i;
             begin
               {Content::expr(op(x,i)) $ i = 1 .. nops(x)}
             end_proc):
Content::set("EQS",
             CL::content("id"="EQS", "keywd"="Equation System", "arity"="2*",
                          "assoc"="false",
                  CL::layout(
                         CL::optable("cols"="2",
                                     "columnalign"="rl",
                                     CL::oper(),
                                     CL::mrow(CL::mo("="),
                                              CL::oper())))),
             DOM_LIST,
             proc(x)
               local i;
             begin
               [Content::expr(op(x,2*i-1)) =
                    Content::expr(op(x,2*i))$ i = 1 .. nops(x)/2]
             end_proc):
Content::set("diff",
             CL::content("id"="diff", "keywd"="Derivative",
                        "arity"="2", "prio"=OP::Mult,
                  CL::layout(
                      CL::mrow(
                          CL::mfrac(CL::mo("form"="prefix",
                                           "&DifferentialD;"),
                                    CL::mrow(CL::mo("form"="prefix",
                                                    "&DifferentialD;"),
                                             CL::oper("idx"="1"))),
                          CL::mo("form"="infix", "&blank;"),
                          CL::oper("idx"="0",
                                   "prio"=OP::Mult)))),
             diff):
Content::set("partialdiff",
             CL::content("id"="partialdiff", "keywd"="Partial Derivative",
                        "arity"="2", "prio"=OP::Mult + 10,
                  CL::layout(
                      CL::mrow(
                          CL::mfrac(CL::mo("form"="prefix",
                                           "&PartialD;"),
                                    CL::mrow(CL::mo("form"="prefix",
                                                    "&PartialD;"),
                                             CL::oper("idx"="1"))),
                          CL::mo("form"="infix", "&blank;"),
                          CL::oper("idx"="0",
                                   "prio"=OP::Mult)))),
             diff):
Content::set("partialdiff1",
             CL::content("id"="partialdiff1", "keywd"="Partial Derivative",
                        "arity"="3", "prio"=OP::Mult + 10,
                  CL::layout(
                      CL::mrow(
                          CL::mfrac(
                              CL::msup(CL::mo("form"="prefix", "&PartialD;"),
                                       CL::oper("idx"="2")),
                              CL::mrow(CL::mo("form"="prefix", "&PartialD;"),
                                       CL::msup(CL::oper("idx"="1"),
                                                CL::oper("idx"="2")))),
                          CL::mo("form"="infix", "&blank;"),
                          CL::oper("idx"="0", "prio"=OP::Mult)))),
             diff):
Content::set("eq",
             CL::content("id"="eq", "keywd"="Equal", "accel"="=",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix", "&Equal;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _equal,
             x -> Content::expr(op(x,1)) = Content::expr(op(x,2))):
Content::set("lt",
             CL::content("id"="lt", "keywd"="Less", "accel"="&lt;",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix", "&lt;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _less):
Content::set("leq",
             CL::content("id"="leq", "keywd"="LessEqual", 
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix", "&leq;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _leequal):
Content::set("neq", CL::content("id"="neq", "keywd"="NotEqual", 
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&NotEqual;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _unequal):
Content::set("gt", CL::content("id"="gt", "keywd"="Greater", "accel"="&gt;",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix", "&gt;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _greater):
Content::set("geq", CL::content("id"="geq", "keywd"="GreaterEqual", 
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix", "&geq;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _geequal):
Content::set("and",
             CL::content("id"="and", "keywd"="And",
                        "arity"="2*", "prio"=OP::And,
                        CL::layout(CL::oprow(CL::oper("prio"=OP::And),
                                             CL::mo("&wedge;")))),
             _and):
Content::set("or",
             CL::content("id"="or", "keywd"="Or",
                        "arity"="2*", "prio"=OP::Or,
                        CL::layout(CL::oprow(CL::oper("prio"=OP::And),
                                             CL::mo("&vee;")))),
             _or):
Content::set("xor",
             CL::content("id"="xor", "keywd"="Xor",
                        "arity"="2*", "prio"=OP::Xor,
                        CL::layout(CL::oprow(CL::oper("prio"=OP::And),
                                             CL::mo("xor")))),
             _xor):
Content::set("not",
             CL::content("id"="not", "keywd"="Not",
                         "arity"="1", "prio"=OP::Not,
                         CL::layout(CL::mrow(CL::mo("form"="prefix",
                                                    "&not;"),
                                          CL::oper("idx"="0", "prio"=OP::Not)))),
             _not):
Content::set("implies",
             CL::content("id"="implies", "keywd"="Implies",
                        "arity"="2", "prio"=OP::Implies,
                         CL::layout(CL::oprow(CL::oper("prio"=OP::Implies),
                                              CL::mo("form"="infix",
                                                     "&DoubleRightArrow;")))),
             _implies):
Content::set("equiv",
             CL::content("id"="equiv", "keywd"="Equivalence",
                        "arity"="2", "prio"=OP::Equiv,
                         CL::layout(CL::oprow(CL::oper("prio"=OP::Equiv),
                                              CL::mo("form"="infix",
                                                     "&DoubleLeftRightArrow;")))),
             _equiv):  
Content::set("union",
              CL::content("id"="union", "keywd"="Union",
                          "arity"="2*", "prio"=OP::Union,
                          CL::layout(CL::oprow(CL::oper("prio"=OP::Intersect),
                                               CL::mo("&cup;")))),
             _union):
Content::set("Union",
              CL::content("id"="Union", "keywd"="UnionSet",
                          "arity"="3", "prio"=OP::Union,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "&Union;"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Relation),
                                       CL::mo("form"="infix", "&Element;"),
                                       CL::oper("idx"="2", "prio"=OP::Relation))),
                              CL::oper("idx"="0", "prio"=OP::Intersect)))),
             solvelib::Union):
Content::set("intersect",
              CL::content("id"="intersect", "keywd"="Intersection",
                          "arity"="2*", "prio"=OP::Intersect,
                          CL::layout(CL::oprow(CL::oper("prio"=OP::Intersect),
                                               CL::mo("&cap;")))),
             _intersect):
Content::set("setdiff",
              CL::content("id"="setdiff", "keywd"="SetMinus",
                          "arity"="2*", "prio"=OP::Minus,
                          CL::layout(CL::oprow(CL::oper("prio"=OP::Intersect),
                                               CL::mo("&setminus;")))),
             _minus):
Content::set("card",
             CL::content("id"="card", "keywd"="Cardinality", "arity"="1", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix",
                                                    "&verbar;"),
                                             CL::oper("idx"="0"),
                                             CL::mo("form"="postfix",
                                                    "&verbar;")))),
             card):
Content::set("subset",
             CL::content("id"="subset", "keywd"="Subset",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&Subset;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _subset):
Content::set("prsubset",
             CL::content("id"="prsubset", "keywd"="Proper Subset",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&Subset;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _prsubset):
Content::set("notsubset",
             CL::content("id"="notsubset", "keywd"="NotSubset",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&NotSubset;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _notsubset):
Content::set("notprsubset",
             CL::content("id"="notprsubset", "keywd"="Not Proper Subset",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&NotSubset;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _notprsubset):
Content::set("exp",
             CL::content("id"="exp", "keywd"="Exponential",
                         "arity"="1", "prio"=OP::Power,
                         CL::layout(CL::msup(CL::mo("&ee;"),
                                             CL::oper("idx"="0")))),
             exp):
Content::set("lg",
             CL::content("id"="lg", "keywd"="Logarithm10", "arity"="1", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("log"),
                                                   CL::mo("10")),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             lg):  // defined in package calc
Content::set("log",
             CL::content("id"="log", "keywd"="Logarithm", "arity"="2", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("log"),
                                                   CL::oper("idx"="1")),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             log,
             X->hold(log)(Content::expr(op(X,2)), Content::expr(op(X,1)))):
Content::set("integers",
             CL::content("id"="integers", "keywd"="Integers", "arity"="0", 
                         CL::layout(CL::mo("&Zopf;"))),
             Z_, ()->Z_):
Content::set("reals",
             CL::content("id"="reals", "keywd"="Reals", "arity"="0", 
                         CL::layout(CL::mo("&Ropf;"))),
             R_, ()->R_):
Content::set("rationals",
             CL::content("id"="rationals", "keywd"="Rationals", "arity"="0", 
                         CL::layout(CL::mo("&Qopf;"))),
             Q_, ()->Q_):
Content::set("naturalnumbers",
             CL::content("id"="naturalnumbers", "keywd"="Natural Numbers",
                         "arity"="0", 
                         CL::layout(CL::mo("&Nopf;"))),
             N_, ()->N_):  // dummy, no BasicSet
Content::set("complexes",
             CL::content("id"="complexes", "keywd"="Complexes", "arity"="0", 
                         CL::layout(CL::mo("&Copf;"))),
             C_, ()->C_):
Content::set("primes",
             CL::content("id"="primes", "keywd"="Primes", "arity"="0", 
                         CL::layout(CL::mo("&Popf;"))),
             P_, ()->P_):        // dummy, no BasicSet
Content::set("exponentiale",
             CL::content("id"="exponentiale", "keywd"="Exponential e",
                         "arity"="0", 
                         CL::layout(CL::mo("&ee;"))),
             exp(1), ()->exp(1)):
Content::set("imaginaryi",
             CL::content("id"="imaginaryi", "keywd"="Imaginary I",
                         "arity"="0", 
                         CL::layout(CL::mo("&ii;"))),
             I, ()->I):
Content::set("notanumber",
             CL::content("id"="notanumber", "keywd"="Not a Number",
                         "arity"="0", 
                         CL::layout(CL::mo("&NotANumber;"))),
             RD_NAN, ()->RD_NAN):
Content::set("true",
             CL::content("id"="true", "keywd"="True", "arity"="0", 
                         CL::layout(CL::mo("&true;"))),
             TRUE, ()->TRUE):
Content::set("false",
             CL::content("id"="false", "keywd"="False", "arity"="0", 
                         CL::layout(CL::mo("&false;"))),
             FALSE, ()->FALSE):
Content::set("unknown",
             CL::content("id"="unknown", "keywd"="Unknown", "arity"="0", 
                         CL::layout(CL::mo("&UNKNOWN;"))),
             UNKNOWN, ()->UNKNOWN):
Content::set("emptyset",
             CL::content("id"="emptyset", "keywd"="Empty Set", "arity"="0", 
                         CL::layout(CL::mo("&emptyv;"))),
             {}, ()->{}):
Content::set("pi",
             CL::content("id"="pi", "keywd"="Pi", "arity"="0", 
                         CL::layout(CL::mo("&pi;"))),
             PI, ()->PI):
Content::set("eulergamma",
             CL::content("id"="eulergamma", "keywd"="Euler Gamma",
                         "arity"="0", 
                         CL::layout(CL::mo("&gamma;"))),
             EULER, ()->EULER):
Content::set("infinity",
             CL::content("id"="infinity", "keywd"="Infinity", "arity"="0", 
                         CL::layout(CL::mo("&infin;"))),
             infinity, ()->infinity):
Content::set("conjugate",
             CL::content("id"="conjugate", "keywd"="Conjugate", "arity"="1", 
                      CL::layout(CL::mover(CL::oper("idx"="0"),
                                           CL::mo("form"="postfix",
                                                  "&OverBar;")))),
             conjugate):
Content::set("definiteint",
             CL::content("id"="definiteint", "keywd"="Definite Integration",
                         "accel"=":", "arity"="4",
               CL::layout(
                   CL::mrow(
                       CL::munderover(CL::mo("form"="prefix", "&Integral;"),
                                      CL::oper("idx"="2"),
                                      CL::oper("idx"="3")),
                       CL::oper("idx"="0", "prio"=OP::Plus),
                       CL::mo("form"="infix", "&DifferentialD;"),
                       CL::oper("idx"="1")))),
             int,
             x -> hold(int)(Content::expr(op(x,1)),
                            Content::expr(op(x,2))=Content::expr(op(x,3)) ..
                                                   Content::expr(op(x,4)))):
Content::set("int",
             CL::content("id"="int", "keywd"="Indefinite Integration",
                         "arity"="2",
               CL::layout(
                   CL::mrow(CL::mo("form"="prefix", "&Integral;"),
                            CL::oper("idx"="0", "prio"=OP::Plus),
                            CL::mo("form"="infix", "&DifferentialD;"),
                            CL::oper("idx"="1")))),
             int):
Content::set("definitesum",
             CL::content("id"="definitesum", "keywd"="Definite Sum",
                        "accel"=";", "arity"="4", "prio"=OP::Plus,
                  CL::layout(
                      CL::mrow(
                          CL::munderover(
                              CL::mo("form"="prefix", "&Sum;"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Relation),
                                       CL::mo("form"="infix", "="),
                                       CL::oper("idx"="2", "prio"=OP::Relation)),
                              CL::oper("idx"="3")),
                          CL::oper("idx"="0", "prio"=OP::Plus)))),
             sum,
             x -> hold(sum)(Content::expr(op(x,1)),
                            Content::expr(op(x,2))=Content::expr(op(x,3)) ..
                                                    Content::expr(op(x,4)))):
Content::set("sum",
             CL::content("id"="sum", "keywd"="Indefinite Sum",
                         "arity"="2", "prio"=OP::Plus,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "&Sum;"),
                              CL::oper("idx"="1")),
                               CL::oper("idx"="0", "prio"=OP::Plus)))),
             sum,
             x -> hold(sum)(Content::expr(op(x,1)),
                            Content::expr(op(x,2)))):
Content::set("sumrootof",
             CL::content("id"="sumrootof", "keywd"="Definite Sum over RootOf",
                         "arity"="3", "prio"=OP::Plus,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "&Sum;"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Relation),
                                       CL::mo("form"="infix", "="),
                                       CL::oper("idx"="2", "prio"=OP::Relation))),
                              CL::oper("idx"="0", "prio"=OP::Plus)))),
             sum,
             x -> hold(sum)(Content::expr(op(x,1)),
                            Content::expr(op(x,2))=Content::expr(op(x,3)))):
Content::set("definiteproduct",
             CL::content("id"="definiteproduct", "keywd"="Definite Product",
                         "arity"="4", "prio"=OP::Mult,
                  CL::layout(
                      CL::mrow(
                          CL::munderover(
                              CL::mo("form"="prefix", "&Pi;"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Relation),
                                       CL::mo("form"="infix", "="),
                                       CL::oper("idx"="2", "prio"=OP::Relation)),
                              CL::oper("idx"="3")),
                          CL::oper("idx"="0", "prio"=OP::Mult)))),
             product,
             x -> hold(product)(Content::expr(op(x,1)),
                                Content::expr(op(x,2))=
                                  Content::expr(op(x,3)) ..
                                    Content::expr(op(x,4)))):
Content::set("product",
             CL::content("id"="product", "keywd"="Indefinite Product",
                         "arity"="2", "prio"=OP::Mult,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "&Pi;"),
                              CL::oper("idx"="1")),
                              CL::mo("form"="infix", "&blank;"),
                          CL::oper("idx"="0", "prio"=OP::Mult)))),
             product,
             x -> hold(product)(Content::expr(op(x,1)),
                                Content::expr(op(x,2)))):
Content::set("root",
             CL::content("id"="root", "keywd"="Root", "arity"="2",
                         CL::layout(CL::mroot(CL::oper("idx"="0"),
                                              CL::oper("idx"="1")))),
             _power, 
             x -> hold(_power)(Content::expr(op(x,1)),
                               1/Content::expr(op(x,2)))):
Content::set("sqrt",
             CL::content("id"="sqrt", "keywd"="Square Root", "arity"="1",
                         CL::layout(CL::msqrt(CL::oper("idx"="0")))),
             _power, 
             x -> hold(_power)(Content::expr(op(x)), 1/2)):
Content::set("limit",
             CL::content("id"="limit", "keywd"="Limit",
                         "arity"="3", "prio"=OP::Plus - 50,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "lim"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Stmtseq),
                                       CL::mo("form"="infix", "&RightArrow;"),
                                       CL::oper("idx"="2", "prio"=OP::Stmtseq))),
                                       CL::mo("form"="infix", "&blank;"),
                              CL::oper("idx"="0", "prio"=OP::Plus - 50)))),
             limit, 
             (x, attributes) ->
                hold(limit)(Content::expr(op(x,1)),
                            Content::expr(op(x,2))=Content::expr(op(x,3)))):
Content::set("limitRight",
             CL::content("id"="limitRight", "keywd"="Right Limit",
                         "arity"="3", "prio"=OP::Plus - 50,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "lim"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Stmtseq),
                                       CL::mo("form"="infix", "&RightArrow;"),
                                       CL::msup(CL::oper("idx"="2",
                                                         "prio"=OP::Stmtseq),
                                                CL::mo("+")))),
                          CL::mo("form"="infix", "&blank;"),
                          CL::oper("idx"="0", "prio"=OP::Plus - 50)))),
             limit, 
             (x, attributes) ->
               hold(limit)(Content::expr(op(x,1)),
                           Content::expr(op(x,2))=Content::expr(op(x,3)),
                           hold(Right))):
Content::set("limitLeft",
             CL::content("id"="limitLeft", "keywd"="Left Limit",
                         "arity"="3", "prio"=OP::Plus - 50,
                  CL::layout(
                      CL::mrow(
                          CL::munder(
                              CL::mo("form"="prefix", "lim"),
                              CL::mrow(CL::oper("idx"="1", "prio"=OP::Stmtseq),
                                       CL::mo("form"="infix", "&RightArrow;"),
                                       CL::msup(CL::oper("idx"="2",
                                                         "prio"=OP::Stmtseq),
                                                CL::mo("-")))),
                          CL::mo("form"="infix", "&blank;"),
                          CL::oper("idx"="0", "prio"=OP::Plus - 50)))),
             limit, 
             (x, attributes) ->
                hold(limit)(Content::expr(op(x,1)),
                            Content::expr(op(x,2))=Content::expr(op(x,3)),
                            hold(Left))):
Content::set("fenced",
             CL::content("id"="fenced", "keywd"="Fence",
                         "accel"=")", "arity"="1",
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "("),
                                             CL::oper("idx"="0"),
                                             CL::mo("form"="postfix", ")")))),
             id, x -> Content::expr(op(x,1))):
Content::set("factorial",
             CL::content("id"="factorial", "keywd"="Factorial",
                         "accel"="!", "arity"="1", "prio"=OP::Fact,
                      CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Fact),
                                          CL::mo("form"="postfix", "!")))),
             fact):
Content::set("in",
             CL::content("id"="in", "keywd"="In",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&Element;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _in):
Content::set("notin",
             CL::content("id"="notin", "keywd"="Not In",
                        "arity"="2", "prio"=OP::Relation,
                        CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Relation),
                                            CL::mo("form"="infix",
                                                   "&NotElement;"),
                                            CL::oper("idx"="1", "prio"=OP::Relation)
                                            ))),
             _notin):
Content::set("abs",
             CL::content("id"="abs", "keywd"="Absolute Value",
                         "accel"="|", "arity"="1", 
                        CL::layout(CL::mrow(CL::mo("form"="prefix",
                                                   "&verbar;"),
                                            CL::oper("idx"="0"),
                                            CL::mo("form"="postfix",
                                                   "&verbar;")))),
             abs):
Content::set("norm",
             CL::content("id"="norm", "keywd"="Frobenius Norm", "arity"="1",
                         CL::layout(CL::mrow(CL::mo("form"="prefix",
                                                   "&Verbar;"),
                                             CL::oper("idx"="0"),
                                             CL::mo("form"="postfix",
                                                   "&Verbar;")))),
             norm,
             proc(x)
               local ex, evex;
             begin
               ex := Content::expr(op(x));
               evex := eval(level(ex));
               if evex::dom::hasProp(Cat::Matrix) = TRUE then
                  hold(norm)(ex, Frobenius);
               else
                  hold(norm)(ex, 2);
               end_if;
             end_proc):
Content::set("onenorm",
             CL::content("id"="onenorm", "keywd"="1-Norm", "arity"="1",
                         CL::layout(CL::msub(CL::mrow(CL::mo("form"="prefix",
                                                             "&Verbar;"),
                                                      CL::oper("idx"="0"),
                                                      CL::mo("form"="postfix",
                                                             "&Verbar;")),
                                             CL::mo("1")))),
             norm,
             x -> hold(norm)(Content::expr(op(x)), 1)):
Content::set("infinitynorm",
             CL::content("id"="infinitynorm", "keywd"="Infinity Norm",
                         "arity"="1",
                         CL::layout(CL::msub(CL::mrow(CL::mo("form"="prefix",
                                                             "&Verbar;"),
                                                      CL::oper("idx"="0"),
                                                      CL::mo("form"="postfix",
                                                             "&Verbar;")),
                                             CL::mo("&infin;")))),
             norm):
Content::set("othernorm",
             CL::content("id"="othernorm", "keywd"="Other Norm",
                         "arity"="2",
                         CL::layout(CL::msub(CL::mrow(CL::mo("form"="prefix",
                                                             "&Verbar;"),
                                                      CL::oper("idx"="0"),
                                                      CL::mo("form"="postfix",
                                                             "&Verbar;")),
                                             CL::oper("idx"="1")))),
                         norm):
Content::set("determinant",
             CL::content("id"="determinant", "keywd"="Determinant",
                         "arity"="1", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix",
                                                    "det"),
                                             CL::oper("prio"=OP::Stmt,
                                                      "idx"="0")))),
             linalg::det):
Content::set("ceil",
             CL::content("id"="ceil", "keywd"="Ceil", "arity"="1",
                         CL::layout(CL::mrow(CL::mo("form"="prefix",
                                                    "&LeftCeiling;"),
                                             CL::oper("idx"="0"),
                                             CL::mo("form"="postfix",
                                                    "&RightCeiling;")))),
             ceil):
Content::set("floor",
             CL::content("id"="floor", "keywd"="Floor", "arity"="1",
                         CL::layout(CL::mrow(CL::mo("&LeftFloor;"),
                                             CL::oper("idx"="0"),
                                             CL::mo("&RightFloor;")))),
             floor):
Content::set("rem",
             CL::content("id"="rem", "keywd"="Remainder",
                         "arity"="2", "prio"=OP::Mod,
                         CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Mod),
                                             CL::mo("mod"),
                                             CL::oper("idx"="1", "prio"=OP::Mod)
                                             ))),
             _mod):
Content::set("quotient",
             CL::content("id"="quotient", "keywd"="Integer Quotient",
                         "arity"="2", "prio"=OP::Mod,
                         CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Mod),
                                             CL::mo("div"),
                                             CL::oper("idx"="1", "prio"=OP::Mod)
                                             ))),
             _div):
Content::set("fnest",
             CL::content("id"="fnest", "keywd"="Nested Functions",
                         "arity"="2", "prio"=OP::Fnest,
                         CL::layout(CL::oprow(CL::oper("prio"=OP::Fnest),
                                              CL::mo("form"="infix",
                                                     "&commat;&commat;")))),
             _fnest):
Content::set("lambda",
             CL::content("id"="lambda", "keywd"="Function Definition",
                        "arity"="2", "prio"=OP::Exprseq,
                        CL::layout(
                                  CL::oprow(CL::oper("prio"=OP::Exprseq),
                                           CL::mo("form"="infix",
                                                 "&rightarrow;")))),
             procdef,
             proc(x)
               local tmp;
             begin
               tmp := "(".expr2text(Content::expr(op(x,1))).") -> (".
                      expr2text(Content::expr(op(x,2))).")";
               text2expr(tmp)
             end_proc):
Content::set("slot",
             CL::content("id"="slot", "keywd"="Slot", "arity"="2", 
                         CL::layout(CL::mrow(CL::oper("idx"="0"),
                                             CL::mo("&Colon;"),
                                             CL::oper("idx"="1")))),
             slot,
             x -> hold(slot)(Content::expr(op(x,1)), expr2text(op(x, 2)))):
Content::set("dilog",
              CL::content("id"="dilog", "keywd"="Dilog",
                        "arity"="1*", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("Li"),
                                                    CL::mo("2")),
                                           CL::mo("form"="prefix","("),
                                           CL::oprow("idx"="0,-1",
                                                    CL::oper("prio"=OP::Stmtseq),
                                                    CL::mo(",")),
                                           CL::mo("form"="postfix",
                                                  ")")))),
             dilog):
Content::set("polylog",
             CL::content("id"="polylog", "keywd"="Polylog", "arity"="2",
                         CL::layout(CL::mrow(CL::mo("polylog"),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo(","),
                                          CL::oper("idx"="1"),
                                          CL::mo("form"="postfix", ")")))),
             polylog):
Content::set("D",
              CL::content("id"="D", "keywd"="Differentiation Operator",
                        "accel"="'", "arity"="1", "prio"=OP::Prime,
                        CL::layout(CL::msup(CL::oper("prio"=OP::Prime-10, "idx"="0"),
                                           CL::mo("form"="postfix",
                                                  "&apos;")))),
             D):
Content::set("Dpara",
              CL::content("id"="Dpara", "keywd"="Differentiation Operator",
                        "arity"="2*",  "assoc"="false",
                   CL::layout(CL::mrow(CL::msub(CL::mo("form"="prefix","D"),
                                               CL::oprow("idx"="0,-2",
                                                        CL::oper("prio"=OP::Stmtseq),
                                                        CL::mo(","))),
                                      CL::mo("form"="prefix","("),
                                      CL::oper("idx"="-1"),
                                      CL::mo("form"="postfix",
                                             ")")))),
             error,
             () -> error("\"Dpara\" -> MuPAD not yet implemented")):
Content::set("besselI",
             CL::content("id"="besselI", "keywd"="BesselI", "arity"="2", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("I"),
                                                     CL::oper("idx"="1")),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             besselI,
             x -> hold(besselI)(Content::expr(op(x, 2)),
                                Content::expr(op(x, 1)))):
Content::set("besselJ",
             CL::content("id"="besselJ", "keywd"="BesselJ", "arity"="2", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("J"),
                                                     CL::oper("idx"="1")),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             besselJ,
             x -> hold(besselJ)(Content::expr(op(x, 2)),
                                Content::expr(op(x, 1)))):
Content::set("besselK",
             CL::content("id"="besselK", "keywd"="BesselK", "arity"="2", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("K"),
                                                     CL::oper("idx"="1")),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             besselK,
             x -> hold(besselK)(Content::expr(op(x, 2)),
                                Content::expr(op(x, 1)))):
Content::set("besselY",
             CL::content("id"="besselY", "keywd"="BesselY", "arity"="2", 
                        CL::layout(CL::mrow(CL::msub(CL::mo("Y"),
                                                     CL::oper("idx"="1")),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             besselY,
             x -> hold(besselY)(Content::expr(op(x, 2)),
                                Content::expr(op(x, 1)))):
Content::set("binomial",
             CL::content("id"="binomial", "keywd"="Binomial", "arity"="2", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "("),
                                             CL::mfrac("linethickness"="0",
                                                       CL::oper("idx"="0"),
                                                       CL::oper("idx"="1")),
                                             CL::mo("form"="postfix", ")")))),
             binomial):
Content::set("range",
             CL::content("id"="range", "keywd"="Range",
                         "arity"="2", "prio"=OP::Range,
                         CL::layout(CL::mrow(CL::oper("idx"="0"),
                                    CL::mo("&period;&period;"),
                                    CL::oper("idx"="1")))),
             _range):
Content::set("compose",
             CL::content("id"="compose", "keywd"="Compose Functions",
                         "accel"="@", "arity"="2*", "prio"=OP::Fconcat,
                         CL::layout(CL::oprow(CL::oper("prio"=OP::Fconcat),
                                              CL::mo("form"="infix",
                                                     "&SmallCircle;")))),
             _fconcat):
Content::set("inverse",
             CL::content("id"="inverse", "keywd"="Functional Inverse",
                         "arity"="1", "prio"=OP::Power,
                         CL::layout(CL::msup(CL::oper("idx"="0","prio"=OP::Power),
                                             CL::mo("(-1)")))),
             _inverse):
Content::set("imageset",
             CL::content("id"="imageset", "keywd"="ImageSet", "arity"="2", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "{"),
                                             CL::oper("idx"="0"),
                                             CL::mo("&verbar;"),
                                             CL::oper("idx"="1"),
                                             CL::mo("form"="postfix", "}")))),
             Dom::ImageSet,
             x -> Dom::ImageSet(Content::expr(op(x,1)),
                                [Content::expr(op(x, [2, 1]))],
                                [Content::expr(op(x, [2, 2]))])):
Content::set("piecewise",
             CL::content("id"="piecewise", "keywd"="Piecewise", "arity"="2*",
                          "assoc"="false",
                 CL::layout(
                     CL::mrow(CL::mo("form"="prefix", "{"),
                         CL::optable("cols"="2",
                                     "columnalign"="cl",
                                     CL::oper(),
                                     CL::mrow(CL::mtext("if&Space;&Space;"),
                                              CL::oper()))))),
             piecewise,
             x -> piecewise([Content::expr(op(x, 2*i)),
                             Content::expr(op(x, 2*i-1))] $ i=1..nops(x)/2)):
Content::set("table",
             CL::content("id"="table", "keywd"="Table", "arity"="0*", 
                  CL::layout(CL::mrow(CL::mo("form"="prefix", "["),
                         CL::optable("cols"="2",
                                     "columnalign"="rl",
                                     CL::oper(),
                                     CL::mrow(CL::mo("="),
                                              CL::oper()))))),
             table,
             x -> table(Content::expr(op(x, 2*i-1)) =
                        Content::expr(op(x, 2*i)) $ i=1..nops(x)/2)):
Content::set("intervaloo",
             CL::content("id"="intervaloo", "keywd"="IntervalOpenOpen",
                         "arity"="2", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "("),
                                             CL::oper("idx"="0"),
                                             CL::mo(","),
                                             CL::oper("idx"="1"),
                                             CL::mo("form"="postfix", ")")))),
             Dom::Interval,
             x -> Dom::Interval(Content::expr(op(x, 1)),
                                Content::expr(op(x, 2)))):
Content::set("intervaloc",
             CL::content("id"="intervaloc", "keywd"="IntervalOpenClosed",
                         "arity"="2", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "("),
                                             CL::oper("idx"="0"),
                                             CL::mo(","),
                                             CL::oper("idx"="1"),
                                             CL::mo("form"="postfix", "]")))),
             Dom::Interval,
             x -> Dom::Interval(Content::expr(op(x, 1)),
                                [Content::expr(op(x, 2))])):
Content::set("intervalco",
             CL::content("id"="intervalco", "keywd"="IntervalClosedOpen",
                         "arity"="2", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "["),
                                             CL::oper("idx"="0"),
                                             CL::mo(","),
                                             CL::oper("idx"="1"),
                                             CL::mo("form"="postfix", ")")))),
             Dom::Interval,
             x -> Dom::Interval([Content::expr(op(x, 1))],
                                Content::expr(op(x, 2)))):
Content::set("intervalcc",
             CL::content("id"="intervalcc", "keywd"="IntervalClosedClosed",
                         "arity"="2", 
                         CL::layout(CL::mrow(CL::mo("form"="prefix", "["),
                                             CL::oper("idx"="0"),
                                             CL::mo(","),
                                             CL::oper("idx"="1"),
                                             CL::mo("form"="postfix", "]")))),
             Dom::Interval,
             x -> Dom::Interval([Content::expr(op(x, 1)),
                                Content::expr(op(x, 2))])):
Content::set("interval",
             CL::content("id"="interval", "keywd"="Interval",
                         "arity"="2", 
                         CL::layout(CL::mrow(CL::oper("idx"="0"),
                                             CL::mo("..."),
                                             CL::oper("idx"="1")))),
             DOM_INTERVAL,
             x -> hull(Content::expr(op(x,1)), Content::expr(op(x,2)))):

Content::set("apply",
             CL::content("id"="apply", "keywd"="Apply Function", "accel"="(",
                        "arity"="2*", "assoc"="false", "flatten-seq"="1,-1",
                         "prio"=OP::Index,
                    CL::layout(
                          CL::mrow(
                              CL::oper("idx"="0", "prio"=OP::Index-1),
                              CL::mo("form"="prefix", "("),
                              CL::oprow("idx"="1,-1",
                                        CL::oper(),
                                        CL::mo("form"="infix", ",")),
                              CL::mo("form"="postfix", ")")))),
             fp::apply, 
             proc(x)
               save XxX;
             begin
               delete XxX;
               
               subs(XxX(Content::expr(op(x, i)) $ i = 2 ..nops(x)),
                    XxX=Content::expr(op(x,1)), Unsimplified)
             end_proc):
Content::set("applyvoid",
             CL::content("id"="applyvoid", "keywd"="Apply Void Function",
                        "arity"="1", "prio"=OP::Index,
                    CL::layout(
                          CL::mrow(
                              CL::oper("idx"="0", "prio"=OP::Index-1),
                              CL::mo("form"="prefix", "("),
                              CL::mo("form"="postfix", ")")))),
             fp::apply, 
             proc(x)
               save XxX;
             begin
               delete XxX;
               
               subs(XxX(), XxX=Content::expr(op(x,1)), Unsimplified)
             end_proc):
Content::set("scalarproduct",
             CL::content("id"="scalarproduct", "keywd"="Scalar Product",
                         "arity"="2", "prio"=OP::Mult,
                         CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Mult),
                                             CL::mo("."),
                                             CL::oper("idx"="1", "prio"=OP::Mult)))),
             linalg::scalarProduct):
Content::set("vectorproduct",
             CL::content("id"="vectorproduct", "keywd"="Vector Product",
                         "arity"="2", "prio"=OP::Mult,
                         CL::layout(CL::mrow(CL::oper("idx"="0", "prio"=OP::Mult),
                                             CL::mo("&times;"),
                                             CL::oper("idx"="1", "prio"=OP::Mult)))),
             linalg::crossProduct):
Content::set("transpose",
             CL::content("id"="transpose", "keywd"="Transpose",
                         "arity"="1", "prio"=OP::Power,
                         CL::layout(CL::msup(CL::oper("idx"="0",
                                                      "prio"=OP::Power),
                                             CL::mo("t")))),
             linalg::transpose):
Content::set("grad",
              CL::content("id"="grad", "keywd"="Gradient",
                        "arity"="2", "prio"=OP::Mult,
                   CL::layout(CL::mrow(CL::msub(CL::mo("&Del;"),
                                                CL::oper("idx"="1",
                                                         "prio"=OP::Noop)),
                                       CL::mo("&blank;"),
                                                 CL::oper("idx"="0",
                                                         "prio"=OP::Mult)
                                                 ))),
             linalg::grad,
             x -> hold(linalg::grad)(Content::expr(op(x, 1)),
                                     [Content::expr(op(x, 2))])):
Content::set("divergence",
              CL::content("id"="divergence", "keywd"="Divergence",
                        "arity"="2", "prio"=OP::Mult,
                   CL::layout(CL::mrow(CL::msub(CL::mo("&Del;"),
                                                CL::oper("idx"="1",
                                                         "prio"=OP::Noop)),
                                       CL::mo("&cdot;"),
                                       CL::oper("idx"="0",
                                                "prio"=OP::Mult)
                                                 ))),
             linalg::divergence,
             x -> hold(linalg::divergence)([Content::expr(op(x, 1))],
                                           [Content::expr(op(x, 2))])):
Content::set("curl",
              CL::content("id"="curl", "keywd"="Rotation",
                        "arity"="6", "prio"=OP::Mult,
                   CL::layout(CL::mrow(CL::msub(CL::mo("&Del;"),
                                                CL::mrow(CL::oper("idx"="3"),
                                                         CL::mo(","),
                                                         CL::oper("idx"="4"),
                                                         CL::mo(","),
                                                         CL::oper("idx"="5"))),
                                       CL::mo("&times;"),
                                       CL::mo("("),
                                       CL::oper("idx"="0"),
                                       CL::mo(","),
                                       CL::oper("idx"="1"),
                                       CL::mo(","),
                                       CL::oper("idx"="2"),
                                       CL::mo(")")))),
             linalg::curl,
             x -> hold(linalg::curl)([map(op(x, 1..3), Content::expr)],
                                     [map(op(x, 4..6), Content::expr)])):
Content::set("laplacian",
              CL::content("id"="laplacian", "keywd"="Laplacian",
                        "arity"="2", "prio"=OP::Mult,
                   CL::layout(CL::mrow(CL::msub(CL::mo("&Delta;"),
                                                CL::oper("idx"="1",
                                                         "prio"=OP::Noop)),
                                       CL::mo("&blank;"),
                                                 CL::oper("idx"="0",
                                                         "prio"=OP::Mult)
                                                 ))),
             linalg::laplacian,
             x -> hold(linalg::laplacian)(Content::expr(op(x, 1)),
                                          [Content::expr(op(x, 2))])):
Content::set("mean",
             CL::content("id"="mean", "keywd"="Mean Value", "arity"="1", 
                      CL::layout(CL::mover(CL::oper("idx"="0"),
                                           CL::mo("form"="postfix",
                                                  "&OverBar;")))),
             stats::mean):
Content::set("quadraticMean",
             CL::content("id"="quadraticMean", "keywd"="Quadratic Mean Value", "arity"="1", 
                      CL::layout(CL::mover(
                                    CL::msup(CL::oper("idx"="0"),
                                             CL::mo("2")),
                                           CL::mo("form"="postfix",
                                                  "&OverBar;")))),
             stats::quadraticMean):
Content::set("sdev",
             CL::content("id"="sdev", "keywd"="Standard Derivation",
                         "arity"="1*", "prio"=OP::Index,
                      CL::layout(CL::mrow(CL::mo("&sigma;"),
                                          CL::mo("form"="prefix", "("),
                                          CL::oprow(CL::oper(),
                                                    CL::mo(",")),
                                          CL::mo("form"="postfix", ")")))),
             stats::stdev):
Content::set("variance",
             CL::content("id"="variance", "keywd"="Variance",
                         "arity"="1*", "prio"=OP::Index,
                      CL::layout(CL::msup(CL::mrow(CL::mo("&sigma;"),
                                                   CL::mo("form"="prefix", "("),
                                                   CL::oprow(CL::oper(),
                                                             CL::mo(",")),
                                                   CL::mo("form"="postfix",
                                                          ")")),
                                          CL::mo("2")))),
             stats::variance):
Content::set("median",
             CL::content("id"="median", "keywd"="Median", "arity"="1*", 
                      CL::layout(CL::mrow(CL::mo("median"),
                                          CL::mo("form"="prefix", "("),
                                          CL::oprow(CL::oper(),
                                                    CL::mo(",")),
                                          CL::mo("form"="postfix", ")")))),
             stats::median):
Content::set("mode",
             CL::content("id"="mode", "keywd"="Modus",
                         "arity"="1*", "prio"=OP::Index,
                      CL::layout(CL::mrow(CL::mo("mode"),
                                          CL::mo("form"="prefix", "("),
                                          CL::oprow(CL::oper(),
                                                    CL::mo(",")),
                                          CL::mo("form"="postfix", ")")))),
             stats::modal):
Content::set("moment",
             CL::content("id"="moment", "keywd"="moment", "arity"="3",
               CL::layout(CL::mrow(CL::mo("form"="prefix", "&lang;"),
                                   CL::msup(CL::oper("idx"="0", "prio"=OP::Exprseq),
                                            CL::oper("idx"="1")),
                                   CL::mo(","),
                                   CL::oper("idx"="2"),
                                   CL::mo("form"="postfix", "&rang;")))),
             stats::moment,
             x -> hold(stats::moment)(Content::expr(op(x, 2)),
                                      Content::expr(op(x, 3)),
                                      Content::expr(op(x, 1)))):

  
Content::set("matrix",
             CL::content("id"="matrix", "keywd"="Matrix",
                         "accel"="#", "arity"="1*",
                    CL::layout(
                          CL::mrow(
                              CL::mo("form"="prefix", "( "),
                              CL::optable(CL::oper()),
                              CL::mo("form"="postfix", ")")))),
             DOM_ARRAY, 
             proc(m, attributes, infoString)
               local rows, cols, i, j;
             begin
               rows := text2expr(attributes[Rows]);
               cols := text2expr(attributes[Cols]);
               array(1..rows, 1..cols,
                     (i,j) = Content::expr(op(m,(i-1)*cols+j))
                     $ j=1..cols $ i=1..rows);
               if infoString <> "" then
                 eval(text2expr(infoString))(%)
               else
                 Dom::Matrix()(%);
               end_if;
             end_proc):
Content::set("invmatrix",
             CL::content("id"="invmatrix", "keywd"="Inverse", "arity"="1",
                         CL::layout(CL::msup(CL::oper("idx"="0"),
                                             CL::mo("-1")))),
             _power,
             m -> hold(_power)(Content::expr(op(m, 1)), -1)):
Content::set("tableau",
             CL::content("id"="tableau", "keywd"="Tableau", "arity"="1*",
                    CL::layout(
                          CL::mrow(
                              CL::optable(CL::oper())))),
             Tableau,
             () -> error("\"tableau\" -> MuPAD not yet implemented")):
Content::set("assign",
             CL::content("id"="assign", "keywd"="Assign",
                         "arity"="2", "prio"=OP::Noop,
                         CL::layout(CL::mrow(CL::oper("idx"="0"),
                                             CL::mo(":="),
                                             CL::oper("idx"="1")))),
             _assign):
Content::set("seq", FAIL, _exprseq):
Content::set("index",
             CL::content("id"="index", "keywd"="Index", "arity"="2",
                         CL::layout(
                                    CL::msub(CL::oper("idx"="0"),
                                             CL::oper("idx"="1")))),
             _index):
Content::set("row",
             CL::content("id"="row", "keywd"="Layout Row",
                         "arity"="1*", "prio"=OP::Index,
                         CL::layout(CL::oprow(CL::oper(),
                                              CL::mo(" ")))),
             error,
             () -> error("\"row\" -> MuPAD not yet implemented")):


Content::set("concat",
             CL::content("id"="concat", "keywd"="Concatenation",
                         "arity"="2*", "prio"=OP::Concat,
                         CL::layout(CL::oprow(CL::oper("prio"=OP::Concat),
                                              CL::mo(".")))),
             _concat):

Content::set("gamma",
             CL::content("id"="gamma", "keywd"="Gamma Function",
                         "arity"="1", "prio"=OP::Index,
                      CL::layout(CL::mrow(CL::mo("&Gamma;"),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo("form"="postfix", ")")))),
             gamma):

Content::set("igamma",
             CL::content("id"="igamma", "keywd"="Integer Gamma Function",
                         "arity"="2", "prio"=OP::Index,
                      CL::layout(CL::mrow(CL::mo("&Gamma;"),
                                          CL::mo("form"="prefix", "("),
                                          CL::oper("idx"="0"),
                                          CL::mo(","),
                                          CL::oper("idx"="1"),
                                          CL::mo("form"="postfix", ")")))),
             igamma):


unalias(CL):


  
// initialize identifier translation tables
map([  hold(`&NIL;`) = NIL, hold(`&FAIL;`) = FAIL, 
       hold(``) = null(),  hold(`&pi;`) = PI,
       Content::Cinfinity() = hold(infinity),
       hold(`&NULL;`) = null(), Content::Ceulergamma() = EULER,  
       hold(`&image;`) = hold(Im), hold(`&real;`) = hold(Re),
       Content::Cintegers() = hold(Z_), Content::Crationals() = hold(Q_),  
       Content::Creals() = hold(R_), Content::Ccomplexes() = hold(C_), 
       Content::Cnaturalnumbers() = hold(N_),  Content::Cprimes() = hold(P_)], 
     Content::translateIdents):

// end of file 
