/*  Diese Routine entstand durch Aufteilung der frueheren Routine "plot::data"
    ($Date: 2002/03/01 10:03:10 $  $Author: walter $   $Revision: 1.8 $)
    Die Subroutinen "Utilities", welche sich vorher direkt in "plot::data" befanden,
    sind nun in "plot::datatools" abgelegt.

    Aufruf: bars(data,<options>)

    data: List of Lists, Matrix or Array(one- or two-dimensional)
    neue Options: DrawMode = Vertical/Horizontal (Default=Vertical)
                  Style = Box, Lines, Points, LinesPoints (Default=Box)
    Alessandro Dell Aere 2001/01/30
*/

/*xx done: 
     DrawMode=Vert./Hor.
     dtyp eingespart
	 Unterprozeduren umgangen (beams,columns)
 	 Fehler beim Schatten fr m=1 beseitigt (in data_rectangle)

	 to do:
	 plot::group zurueckliefern
	
xx*/
  
/*--
    Rechtecke muessen durch zwei Polygone, die jeweils ein Dreieck bilden,
    dargestellt werden, da derzeit nur Dreiecke mit einer Farbe gefuellt
    werden koennen. Ein weiteres Polygon bildet den Rahmen des Rechtecks.
    Ein Rechteck besteht daher aus drei Polygonen.

    Das gesamte Saeulen- oder Balkendiagramm koennte durch ein Objekt dar-
    gestellt werden. Man koennte dann aber nur zwei Beschriftungen eingeben,
    eine fuer die Szene und eine fuer das Objekt. Aus diesem Grund wird
    jedes Polygon als ein eigenes Objekt dargestellt. Demnach koennen
    3*m*n+1 Beschriftungen eingefuegt werden.
--*/
/*++
    >> export(plot): export(RGB):


    >> plot(bars([[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue] ))
    >> plot(bars([[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Style=Lines ))

    >> plot(bars([[5,10,24,-3,6,5,2,18]], DrawMode = Vertical))

    >> plot(bars([[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue]))
	
    >> plot(bars([[binomial(15,i)$i=0..15]],Colors=[[1-j/15,j/15,0.9]$j=0..15]))

    >> plot(bars([[i^2$i=-15..15],[10*j$j=-15..15]]))

    >> L := [[2,1,2,4,5],[1,2,3,2,1],[2,-1,-3,1,2],[5,4,3,2,1],[2,1,2,1,2]]:
       M := matrix(L):
       A :=array(1..5, 1..5, 
                  (1,1) = 2, (1,2) = 1, (1,3) = 2, (1,4) = 4, (1,5) = 5,
                  (2,1) = 1, (2,2) = 2, (2,3) = 3, (2,4) = 2, (2,5) = 1,
                  (3,1) = 2, (3,2) = 1, (3,3) = -3, (3,4) = 1, (3,5) = 2, 
                  (4,1) = 5, (4,2) = 4, (4,3) = 3, (4,4) = 2, (4,5) = 1, 
                  (5,1) = 2, (5,2) = 1, (5,3) = 2, (5,4) = 1, (5,5) = 2):
    >> plot(bars(L))
    >> plot(bars(M))
    >> plot(bars(A))

    >> plot(bars([[cos(j)*sin(i)$i=1..10]$j=1..10]))
    >> plot(bars([[(sin(1/i + j)/i, -cos(1/2/i + j)/i)$i = 1..10]$j = 1..5],
          DrawMode = Horizontal))
    >> plot(bars(array(1..60,
         (i) = sqrt(i)*(-1)^i*sin(1/i)*cos(i/50*PI) $i=1..60)))


++*/
plot::bars:= proc(l)
    local s, st, c, i, j, m, n, nc,d, t, p, v, a, b, nli, nlj, dim, pw,
          sspec, tspec, cspec, vspec, options, offset, scene_default;
begin

if args(0) < 1 then
        error("1st argument: expecting data")
else

      // get options:

      c := RGB::ColorList; // default color list
      t := []; // Titles
      v := Vertical; // DrawMode
      st := Box; // Style
      pw := 40; // PointWidth
      options := [args(2..args(0))];
      s := nops(options);
      sspec := tspec := cspec := vspec := FALSE;
      for i from 1 to nops(options) do
          if type( (p := options[i]) ) <> "_equal" then
              error("specification of the form 'option = value' expected")
          end_if;
          case op(p,1)
          of hold(Colors) do 
              c:= op(p,2);
              if domtype(c) <> DOM_LIST or nops(c) < nops(l) then
                  error("option 'Colors': list with ".nops(l)." RGB color values expected")
              end_if;
              options[i]:= FAIL;
              cspec:= TRUE;
              break
	  of hold(Style) do
	      st:= op(p,2);
	      if (st <> Box and 
                  st <> Points and 
                  st <> Lines and 
                  st <> LinesPoints) then
	          error("option 'Style': 'Box', 'Lines', 'Points', or 'LinesPoints' expected")
	      end_if;
              options[i]:= FAIL;
              sspec:= TRUE;
              break
          of hold(Titles) do 
              t:= op(p,2);
              if not testtype(t,Type::ListOf(DOM_STRING,nops(l) $ 2)) then
                  error("option 'Titles': list with ".nops(l)." strings")
              end_if;
              options[i]:= FAIL;
              tspec:= TRUE;
	          break
	  of hold(DrawMode) do
	      v:=op(p,2);
	      if(v <> Vertical and v <> Horizontal) then
		error("option 'DrawMode': Vertical or Horizontal expected")
	      end_if;			   
	      options[i]:=FAIL;
	      vspec:=TRUE;
			
        end_case;

          if sspec = TRUE and cspec = TRUE and tspec = TRUE and vspec=TRUE 
	  then break 
	  end_if;
      end_for;
      options:= subs( options,FAIL=null() )
    end_if;

    if testargs() 
    then if not testtype(l, Type::ListOf(DOM_LIST)) 
            and not domtype(l) = DOM_ARRAY 
	    and not l::dom::hasProp(Cat::Matrix) = TRUE
         then error("1st argument: list of lists or array or matrix expected")
         end_if;

         if testtype(l, Type::ListOf(DOM_LIST)) 
         then if _or(nops(l[i]) <> nops(l[i + 1])$i = 1..nops(l) - 1) 
              then error("1st argument: lists must have the same number of elements")
              end_if;
         elif domtype(l) = DOM_ARRAY 
         then if op(l,[0,1]) <> 2 and op(l,[0,1]) <> 1
	      then error("1st argument: array must be one- or two-dimensional")
              end_if;
         end_if;
     end_if;

     //--------------------------------------------
     // Matrix, Array --> List of Lists :
     if l::dom::hasProp( Cat::Matrix ) = TRUE then
        l:= expr(l): // Matrix -> Array
     end_if;
     nli := [];
     if domtype(l) = DOM_ARRAY // then Array -> List of Lists
     then dim := op(l,[0,1]): // Dimension of Array
          (a[i] := op(l, [0, i + 1, 1])) $ i = 1..2;
          (b[i] := op(l, [0, i + 1, 2])) $ i = 1..2;
	  if dim = 2 
          then for i from a[1] to b[1] do
                nlj := [];
                for j from a[2] to b[2] do
                  nlj := append(nlj, l[i,j])
                end_for;
                nli := append(nli, nlj)
               end_for;
	       l := nli;
   	  elif dim = 1
          then for i from a[1] to b[1] do
                 nli := append(nli,l[i]);
               end_for;
	       l :=[nli]:
          end_if;
    end_if;
    //---------------------------------------------

     l := float(l);
     if not testtype(l, Type::ListOf(Type::ListOf(Type::Real, 1), 1)) 
     then error("1st argument: values must be real numbers")
     end_if;

    m := nops(l); n := nops(l[1]); nc := nops(c);
    offset := (if m = 1 then 0.2 else 1 end_if);
    t := plot::data_set_titles(t, l, 3*m*n);

   if v=Horizontal then 
         // set options:
         scene_default:= 
           [hold(Axes) = hold(Origin), hold(Scaling) = hold(UnConstrained),
	    hold(Labeling) = TRUE, hold(Labels) = ["",""], 
	    hold(Ticks) = hold([Automatic,0]),hold(Title) = ""];
        if options <> [] then
          if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
          if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
	      if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
		  if has(options,hold(Labels)) then scene_default[4]:= FAIL end_if;
   		  if has(options,hold(Ticks)) then scene_default[5]:= FAIL end_if;
		  if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
          options:= subs( scene_default, FAIL=null()) . options;
        else
         options:= scene_default
        end_if;//of options

        if st = Box 
	then return(plot::Scene((plot::data_rectangle(
                0,j+i*(m+offset),l[j+1][i+1],j+1+i*(m+offset),
               (if m=1 and nc=n then c[i+1] else c[j+1] end_if),t[i*m+j+1],
               (if m=1 then 0.2,0.1,FALSE else null() end_if) )
	        $ j=0..m-1) $ i=0..n-1,
                plot::Polygon(plot::Point(0,0),
	                      plot::Point(0,(if m=1 then(m+offset)*n-offset+0.1 
			                            else(m+offset)*n-offset 
					     end_if)),
	                         hold(Color)=RGB::Black), op(options)  ))
	elif st = Lines
        then return(plot::Scene((plot::Polygon(
                                        plot::Point(0, j + 1/2 + i*(m+offset)),
		                        plot::Point(l[j+1][i+1], j + 1/2 + i*(m+offset)),
		                        hold(Color) = [Flat,(if m=1 and nc=n 
					                     then c[i+1] else c[j+1] 
							     end_if)])
		    $ j=0..m-1) $ i=0..n-1,
                    plot::Polygon(plot::Point(0,0),
	                          plot::Point(0,(if m=1 then(m+offset)*n-offset+0.1 
				                        else(m+offset)*n-offset 
					         end_if)),
	                          hold(Color)=RGB::Black), op(options)  ))
        elif st = Points then 
             return(plot::Scene(
		    (plot::Point(l[j+1][i+1], j + 1/2 + i*(m+offset),
                                hold(Color) = (if m=1 and nc=n then 
                                               c[i+1] else c[j+1] end_if)) 
		    $ j=0..m-1) $ i=0..n-1,
                    plot::Polygon(plot::Point(0,0),
	            plot::Point(0,(if m=1 then(m+offset)*n-offset+0.1 
		                          else(m+offset)*n-offset 
			           end_if)),
	                          hold(Color)=RGB::Black), op(options)  ));
        elif st = LinesPoints
        then return(plot::Scene((plot::Polygon(
                    plot::Point(0, j + 1/2 + i*(m+offset)),
		    plot::Point(l[j+1][i+1], j + 1/2 + i*(m+offset)),
		    hold(Color) = [Flat,(if m=1 and nc=n 
		                         then c[i+1] else c[j+1] 
					 end_if)] ) 
		    $ j=0..m-1) $ i=0..n-1,
		   (plot::Point(l[j+1][i+1], j + 1/2 + i*(m+offset),
	                           hold(PointStyle) = FilledCircles,
			           hold(PointWidth) = pw, 
				   hold(Color) = RGB::Black)
		    $ j=0..m-1) $ i=0..n-1,
                    plot::Polygon(plot::Point(0,0),
	            plot::Point(0,(if m=1 then(m+offset)*n-offset+0.1 
		                          else(m+offset)*n-offset 
			           end_if)),
	                          hold(Color)=RGB::Black), op(options)  ))
	end_if; // st

   elif v=Vertical then 
       
         if m = 1 then d := (max(op(l[1]))-min(op(l[1])))/50.0 end_if;

         // set options:
         scene_default:= 
          [hold(Axes) = hold(Origin), hold(Scaling) = hold(UnConstrained),
	   hold(Labeling) = TRUE, hold(Labels) = ["",""], 
	   hold(Ticks) = plot::data_ticks(l), hold(Title) = ""];
         if options <> [] then
          if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
          if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
          if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
          if has(options,hold(Labels)) then scene_default[4]:= FAIL end_if;
          if has(options,hold(Ticks)) then scene_default[5]:= FAIL end_if;
          if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
          options:= subs( scene_default, FAIL=null()) . options;
         else
         options:= scene_default
         end_if;// of options

    if st = Box 
    then return(plot::Scene(
                     (plot::data_rectangle(
                      j+i*(m+offset),0,j+1+i*(m+offset),l[j+1][i+1],
                      (if m=1 and nc=n then c[i+1] else c[j+1] end_if),t[i*m+j+1],
                      (if m=1 then /*0.1,0.3,TRUE*/ 
		       0.1,d,TRUE else null() end_if))
                      $ j=0..m-1) $ i=0..n-1,
		      plot::Polygon(plot::Point(0,0),
	                      plot::Point((m+offset)*n-offset,0),
	                      hold(Color)=RGB::Black), op(options)   ))
    elif st = Lines
    then return(plot::Scene(
                      (plot::Polygon(plot::Point(j+1/2+i*(m+offset),0),
		                     plot::Point(j+1/2+i*(m+offset),l[j+1][i+1]),
                                     hold(Color) = (if m=1 and nc=n 
				                    then c[i+1] else c[j+1] 
						    end_if)
	               ) $ j=0..m-1) $ i=0..n-1,
		       plot::Polygon(plot::Point(0,0),
	                             plot::Point((m+offset)*n-offset,0),
	                             hold(Color)=RGB::Black), op(options))
	         )

    elif st = Points
    then return(plot::Scene(
		      (plot::Point(j+1/2+i*(m+offset),l[j+1][i+1],
                                  hold(Color) = (if m=1 and nc=n then 
                                                 c[i+1] else c[j+1] end_if))
		       $ j=0..m-1) $ i=0..n-1,
		       plot::Polygon(plot::Point(0,0),
	                      plot::Point((m+offset)*n-offset,0),
	                      hold(Color)=RGB::Black), op(options))
               );
                
    elif st = LinesPoints
    then return(plot::Scene(
                      (plot::Polygon(plot::Point(j+1/2+i*(m+offset),0),
		                     plot::Point(j+1/2+i*(m+offset),l[j+1][i+1]),
                                     hold(Color) = (if m=1 and nc=n 
				                    then c[i+1] else c[j+1] 
						    end_if))
		       $ j=0..m-1) $ i=0..n-1,
		      (plot::Point(j+1/2+i*(m+offset), l[j+1][i+1],
	                           hold(PointStyle) = FilledCircles,
			           hold(PointWidth) = pw, 
				   hold(Color) = RGB::Black)
		       $ j=0..m-1) $ i=0..n-1,
		       plot::Polygon(plot::Point(0,0),
	                             plot::Point((m+offset)*n-offset,0),
	                             hold(Color)=RGB::Black), op(options))
	         )
    end_if;// st

   end_if;// of v


end_proc: /*of main*/


