
/*----------------------------------------------------------------*/
/*---------- Utilities for bars,piechart2d/3d,surface ------------*/
/*----------------------------------------------------------------*/

plot::data_set_titles:= proc(t, l, n)
   name plot::data;
    local i, j, k, m;
begin
    if (k:=contains(t, hold(Values))) <> 0 then
        delete t[k];
        if testtype(l[1], DOM_LIST) then
            t := t.[(expr2text(l[i][j])$i=1..nops(l))$j=1..nops(l[1])]
        else
            t := t.[expr2text(l[i])$i=1..nops(l)]
        end_if
    end_if;
    m:= nops(t);
    if m < n then t := t.[ "" $ i = 1..n-m] end_if;
    return(t)
end_proc:

/*-----------------------------------------------------------*/

plot::data_rectangle := proc(a,b,c,d,co,ti,sx,sy,vertical)
    name plot::data;
begin
   
  plot::Polygon( plot::Point(a,b), plot::Point(a,d), plot::Point(c,d), 
                 hold(Filled) = TRUE, hold(Color) = co, hold(Title) = ti,
	             hold(TitlePosition) =[0,0]  ),
  plot::Polygon( plot::Point(a,b), plot::Point(c,b), plot::Point(c,d), 
                 hold(Filled) = TRUE, hold(Color) = co  ),
	
    (if args(0) = 6 then
        // Draw frame 
        plot::Polygon(plot::Point(a,b), plot::Point(a,d), plot::Point(c,d),
	    plot::Point(c,b), plot::Point(a,b), hold(Color) = RGB::Black)
    else
        if vertical then
            if d < 0 then sy := -sy; end_if;
            plot::Polygon(plot::Point(a+sx,d),plot::Point(a+sx,d+sy),plot::Point(c+sx,d+sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray),
            plot::Polygon(plot::Point(a+sx,d),plot::Point(c+sx,d),plot::Point(c+sx,d+sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray),
			plot::Polygon(plot::Point(c,b),plot::Point(c,d+sy),plot::Point(c+sx,d+sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray),
			plot::Polygon(plot::Point(c,b),plot::Point(c+sx,d+sy),plot::Point(c+sx,b), hold(Filled) = TRUE, hold(Color) = RGB::Gray)
            
		else
            if c < 0 then sx := -sx; end_if;
            plot::Polygon(plot::Point(c,b+sy),plot::Point(c+sx,b+sy),plot::Point(c+sx,d+sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray), 
            plot::Polygon(plot::Point(c,b+sy),plot::Point(c+sx,d+sy),plot::Point(c,d+sy), hold(Filled) = TRUE,hold(Color) = RGB::Gray),
			plot::Polygon(plot::Point(a,d),plot::Point(a,d+sy),plot::Point(c,d), hold(Filled) = TRUE,hold(Color) = RGB::Gray),
			plot::Polygon(plot::Point(a,d+sy),plot::Point(c,d),plot::Point(c,d+sy), hold(Filled) = TRUE,hold(Color) = RGB::Gray)
           
	   end_if
    end_if),
	 plot::Polygon(plot::Point(a,b), plot::Point(a,d), plot::Point(c,d),
	    plot::Point(c,b), plot::Point(a,b), hold(Color) = RGB::Black)

end_proc:

/*-----------------------------------------------------------*/

plot::data_ticks:= proc(l, ymin=FAIL, ymax=FAIL)
    name plot::data;
    local i, tx, ty, tz;
begin
    if type(l[1][1]) = DOM_LIST then
        tx := [ l[i][j][1] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
        tx := plot::data_xyzticks(tx);
        ty := [ l[i][j][2] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
        ty := plot::data_xyzticks(ty, ymin, ymax);
        if nops(l[1][1]) = 2 then
            return([ tx, ty ])
        else
            tz := [ l[i][j][3] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
            tz := plot::data_xyzticks(tz);
            return([ tx, ty, tz ])
        end_if
    else 
        ty := [ op(l[i]) $ i = 1..nops(l) ];
        return([ 0, plot::data_xyzticks( ty ) ])
    end_if
end_proc:

/*-----------------------------------------------------------*/

plot::data_xyzticks:= proc(l, ymin=FAIL, ymax=FAIL)
    name plot::data;
    local d, i, j, n;
begin
    // Convert ty to a set to remove multiple entries
    // and re-convert into a list
    l:= sort( [op({op(l)})] );  
    n:= nops(l);
    // Maximal tick marks: 40 (Internal upper bound!)
    if n > 40 then
       // 10 (arbitrary) tick marks 
       return( 10 )
    end_if;
    if type(ymin) = DOM_FLOAT then
        d:= (ymax-ymin)/20.0
    else
        d:= (l[n]-l[1])/20.0
    end_if;
    j:= l[1];
    for i from 2 to n do
        if l[i] - j < d then
            l[i]:= l[i] = ""
        else
            j:= l[i]
        end_if
    end_for;

    return(l)
end_proc:

/*-----------------------------------------------------------*/

plot::data_curve_min_max := proc(p, a, b, g)
   name plot::data;
    local j, ma, mi, v, x;
begin
    mi:= infinity; ma:= -infinity;
    for j from 1 to nops(p) do
        for x from a[j] to b[j] step (b[j]-a[j])/g do
            v := p[j](float(x));
            if v < mi then 
                mi := v
            elif v > ma then
                ma := v
            end_if
        end_for
    end_for;
    return( [mi, ma] )
end_proc:

/*-----------------------------------------------------------*/

plot::data_point2d_list := proc(l, t, c)
    name plot::data;
    local i, j;
begin
    plot::Pointset(                     
        plot::Point(l[j][i][1],l[j][i][2],hold(Color)=c[j]) $ i = 1..nops(l[j]), 
        hold(Title) = t[j], hold(TitlePosition) = [0,0],
        hold(PointWidth) = 40, hold(PointStyle) = hold(FilledCircles)
    ) $ j=1..nops(l)
end_proc:

/*-----------------------------------------------------------*/

plot::data_point3d_list := proc(l, t, c)
   name plot::data;
    local i, j;
begin
    return(plot::Pointset(
        plot::Point(l[j][i][1],l[j][i][2],l[j][i][3],hold(Color) = c[j])$ i = 1..nops(l[j]),
        hold(Title) = t[j], hold(TitlePosition) = [0,0]
    ) $ j = 1..nops(l))
end_proc:

/*-----------------------------------------------------------*/

plot::data_checkformat := proc(l, dtyp)
    name plot::data;
    local j, check;
begin
    check:= proc(l, dtyp)
        local a, b, dim, i, j, nl, t;
    begin
        if domtype(l) = DOM_ARRAY or 
            l::dom::hasProp( Cat::Matrix ) = TRUE 
        then
            if (dim := op(l,[0,1])) > 2 then 
                error("one- or two-dimensional array or matrix expected")
            end_if
        else
            return( l )
        end_if;

        (a[i] := op(l, [0, i+1, 1])) $ i=1..dim;
        (b[i] := op(l, [0, i+1, 2])) $ i=1..dim;

        nl := [];
        if dim = 1 then
            t := {hold(Points), hold(Lines), hold(Curves), hold(LinesPoints), hold(CurvesPoints)};
            for i from a[1] to b[1] do
                if testtype(l[i], Type::Real) then
                    if contains(t, dtyp) then
                        nl := append(nl, [i, l[i]])
                    else
                        nl := append(nl, l[i])
                    end_if
                end_if
            end_for
        else // dim = 2
            for i from a[1] to b[1] do
                for j from a[2] to b[2] do
                    if testtype(l[i,j], Type::Real) then
                        nl := append(nl, [i, j, l[i,j]])
                    end_if
                end_for
            end_for
        end_if;

        return(nl)
    end_proc:

    if type(l) = DOM_LIST then
        [ check(l[j], dtyp) $ j = 1..nops(l) ]
    else
        [ check(l, dtyp) ]
    end_if
end_proc:
