// $Date: 2002/11/07 17:03:22 $  $Author: ccr $   $Revision: 1.26.2.1 $

/*++
  implplot.mu

  Eine neue Implementation von implicit fr Nullstellenmengen
  glatter Funktionen R^2->R.  Die Funktion mu auf der 
  Nullstellenmenge fast berall regulr sein.

  Beispiele:

  >> f:= plot::implicit(sin(x*y),x=-5..5,y=-5..5):
     plot(f)

  >> f := plot::implicit(sin(5*sin(x)*y),x=-5..5,y=-5..5):
     plot(f)

  >> f := plot::implicit(sin(x^3-x*y),x=-5..5,y=-5..5):
     plot(f)

  >> f := plot::implicit(y*(3*x^2-y^2)-(x^2+y^2)^2, x=-1..1, y=-1..1):
     plot(f)
++*/

// utility functions used a couple of times:
alias(perturb(x, h) =
  ((userinfo(20,"Perturbing...");
    while traperror((x := x + h/(100*(stdlib::frandom()-0.5))))=260
    do end_while;)));


alias(newton2d(f, x, y, Df, x0, y0, h, prec, i, fxy, Df_square, Dfxy, hf) =
  ((i := 0;
    while traperror((fxy := float(subs(f,[x=x0,y=y0])))) <> 0
    do perturb(x0, h); perturb(y0, h); end_while;
      while traperror((
      Dfxy := [ float(subs(Df[1],[x=x0,y=y0])), 
		float(subs(Df[2],[x=x0,y=y0])) ];
      )) <> 0 do
         perturb(x0, h); perturb(y0, h);
         while traperror((fxy := float(subs(f,[x=x0,y=y0])))) <> 0
         do perturb(x0, h); perturb(y0, h); end_while;
      end_while;
      Df_square := rdmult(Dfxy[1],Dfxy[1],1)+rdmult(Dfxy[2],Dfxy[2],1);
    while
//(specfunc::abs(fxy)>prec)
         not iszero(fxy)
         and (iszero(Df_square) 
              or specfunc::abs(fxy)/specfunc::sqrt(Df_square) > prec)
         and i<500 do
      while traperror((
      Dfxy := [ float(subs(Df[1],[x=x0,y=y0])), 
		float(subs(Df[2],[x=x0,y=y0])) ];
      )) <> 0 do
         perturb(x0, h); perturb(y0, h);
         while traperror((fxy := float(subs(f,[x=x0,y=y0])))) <> 0
         do perturb(x0, h); perturb(y0, h); end_while;
      end_while;
      Df_square := rdmult(Dfxy[1],Dfxy[1],1)+rdmult(Dfxy[2],Dfxy[2],1);
      if iszero(fxy) then break; end_if;
      if specfunc::abs(Df_square/fxy)<1e-15 then 
        // Rang(Df) not maximal. Make a small perturbation instead.
        perturb(x0, h);
        perturb(y0, h);
      else
	hf := h * fxy/Df_square;
        [x0, y0] := [x0 - hf*Dfxy[1], y0 - hf*Dfxy[2]];
      end_if:
      while traperror((fxy := float(subs(f,[x=x0,y=y0])))) <> 0
      do perturb(x0, h); perturb(y0, h); end_while;
      i := i+1;
    end_while:
))):


plot::implicit:= proc(F,xrx,yry)
  local f,c,x,y,xmin,xmax,ymin,ymax,plotfactor,
        Lf,Lstart,clean_startpoints,PlotList,
        currcolor, Options, i, x0, y0, h0, fxy,
        i_n2d, fxy_n2d, Df_square_n2d, Dfxy_n2d, hf_n2d;
begin
  if not(testtype(F,DOM_LIST)) then F := [F] end_if:

  F := map(F, x -> if type(x) = "_equal" 
                     then eval(subsop(x, 0=_subtract))
                     else x
                   end_if);

  if testargs() then
    if testtype(xrx, Type::Equation()) <> TRUE
       then error("expecting DOM_IDENT=Range as second argument");
    end;
    if testtype(yry, Type::Equation()) <> TRUE
       then error("expecting DOM_IDENT=Range as third argument");
    end;
  end;

  // testtype(xrx,"_equal")...
  ([x,xmin])    := [op(xrx)];
  ([xmin,xmax]) := map([op(xmin)],float);
  ([y,ymin])    := [op(yry)];
  ([ymin,ymax]) := map([op(ymin)],float);

  if testargs() then
    if testtype(x, Type::Union(DOM_IDENT, DOM_VAR)) = FALSE
       then error(expr2text(x)." is not an identifier");
    end;
    if testtype(y, Type::Union(DOM_IDENT, DOM_VAR)) = FALSE
       then error(expr2text(y)." is not an identifier");
    end;
    if testtype([xmin, xmax, ymin, ymax], 
                Type::ListOf(DOM_FLOAT, 4, 4)) <> TRUE
       then error("Ranges must be numerical and real");
    end;
    if xmin > xmax then [xmin, xmax] := [xmax, xmin]; end;
    if ymin > ymax then [ymin, ymax] := [ymax, ymin]; end;
    x0 := (xmin + xmax)/2; y0 := (ymin + ymax)/2;
    h0 := min(xmax - xmin, ymax - ymin)/1e2;
    while traperror((fxy := float(subs(F,[x=x0,y=y0])))) <> 0
    do perturb(x0, h0); perturb(y0, h0); end_while;
    if testtype(fxy,
	Type::Union(DOM_FLOAT, Type::ListOf(DOM_FLOAT))) <> TRUE
       then error("first argument must depend only on ".expr2text(x).
                  " and ".expr2text(y));
    end;
  end;

  // assumptions for fast simplifications below
  if x::dom = DOM_IDENT then
    save x;
    assume(xmin <= x <= xmax);
  end_if;

  if y::dom = DOM_IDENT then
    save y;
    assume(ymin <= y <= ymax);
  end_if;

  fxy := hull(_union(op(map(F, f->hull(abs(interval(subs(f, [x=xmin...xmax, y=ymin...ymax]))))))));

  if rhs(fxy) > 1 then fxy := 0...1; end_if;

  Options := table(
    Grid = [7,7],
    Colors = [[1.0,0.0,0.0],[0.0,1.0,0.0],[0.0,0.0,1.0],
                 [1.0,1.0,0.0],[1.0,0.0,1.0],[0.0,1.0,1.0]],
    MinStepsize = min(xmax-xmin,ymax-ymin)*1.0e-6,
    MaxStepsize = min(xmax-xmin,ymax-ymin)*1.0e-2,
    StartingStepsize = min(xmax-xmin,ymax-ymin)*1.0e-3,
    Precision = 10.0^(2-DIGITS) * rhs(fxy),
    Ticks=[10,10],
    Contours=[0],
    Splines=FALSE,
    Factor=FALSE,
    args(4..args(0)));

  if nops(Options[Grid]) < 2 then
    Options[Grid] := [Options[Grid],Options[Grid]];
  end_if;

  if Options[Grid][1] < 2 or Options[Grid][2] < 2 then
    error("Grid must be at least 2 in each direction!");
  end_if;

// xxx Argument checking!!

  currcolor := 1;

  clean_startpoints :=
  proc(last_x, last_y, cur_x, cur_y, epsilon)
    name plot::implicit::clean_startpoints;
    local i, min_x, max_x, min_y, max_y, a, b, dist;
    option escape;
  begin
    min_x := min(last_x, cur_x) - epsilon;
    max_x := max(last_x, cur_x) + epsilon;
    min_y := min(last_y, cur_y) - epsilon;
    max_y := max(last_y, cur_y) + epsilon;
    for i from 1 to nops(Lstart) do
      if(min_x <= Lstart[i][1]) then
        break;
      else
        if Lstart[i][1]>max_x then return(); end_if;
      end_if;
    end_for;

    a := last_y - cur_y;
    b := cur_x - last_x;
    dist := sqrt(rdmult(a,a,1)+rdmult(b,b,1));
    if iszero(dist) then return() end_if;
    a := a/dist;
    b := b/dist;

    while(i<=nops(Lstart) and Lstart[i][1]<max_x) do
      if (Lstart[i][2]>min_y and Lstart[i][2]<max_y
          and specfunc::abs(a*(Lstart[i][1]-cur_x)+b*(Lstart[i][2]-cur_y))
		<epsilon )
      then
        userinfo(25,"Deleting ".expr2text(Lstart[i]));
        delete Lstart[i];
      else
        i := i+1;
      end_if;
    end_while;
  end_proc;
  
  plotfactor :=
  proc(f)
    name plot::implicit::plotfactor;
    local Df, plotsingle, startx, starty, x_i, y_i,
          cur_x, cur_y, dummy, Lsol;
  begin
    plotsingle :=
    proc(cur_x, cur_y)
      name plot::implicit::plotfactor::plotsingle;
      local direction, h, Df_x, Df_y, Df_xy, dx, dy,
            res, res_x, res_y, old_nres, old_Dx, old_Dy,
            last_x, last_y, n, t, minmax;
    begin
      minmax := proc() begin [min(args()),max(args())]; end;
      res := [[startx,starty,0.0]];
      for direction from -1 to 1 step 2 do
        old_nres := nops(res);
        cur_x := startx;
        last_x := startx;
        last_y := starty;
        cur_y := starty;
        h := Options[StartingStepsize];
        t := 0.0;
        old_Dx:=0;
        old_Dy:=0;
        traperror((
// xxx Schutz vor periodischen Lsungen verbessern!
        while xmin <= cur_x and cur_x <= xmax
          and ymin <= cur_y and cur_y <= ymax
          and (nops(res)-old_nres<5
               or ((rdmult(cur_x-startx,cur_x-startx,1)
                   +rdmult(cur_y-starty,cur_y-starty,1))>h^2))
          and nops(res) < 1000 do

          Df_x := float(subs(Df[1],x=cur_x,y=cur_y));
          Df_y := float(subs(Df[2],x=cur_x,y=cur_y));
// Mit der QR-Zerlegung von numeric::factorQR ist det(Q)=
//
//                 3                            2
//                b                            a  b
//  - --------------------------- - ---------------------------,
//                 /    2    \1/2                /    2    \1/2
//      2    2 3/2 |   b     |        2    2 3/2 |   b     |
//    (a  + b )    | ------- |      (a  + b )    | ------- |
//                 |  2    2 |                   |  2    2 |
//                 \ a  + b  /                   \ a  + b  /
//
// also sign(det(Q))=-sign(b), falls (a,b)<>(0,0).  Weiterhin gilt
//      R = ((Df_xy[1]^2+Df_xy[2]^2)^(1/2)),
// also det(R)>=0.  Somit ist
//     ( diff(f,x)  diff(f,y) )
// sign(   dx          dy     ) = direction
//
// erfllt fr

          Df_xy := sqrt(rdmult(Df_x,Df_x,1)+rdmult(Df_y,Df_y,1));
          if iszero(Df_xy) then
            break;
          end_if;
          dx    := direction * -Df_y/Df_xy;
          dy    := direction * Df_x/Df_xy;

          old_Dx := old_Dx * Df_x;
          old_Dy := old_Dy * Df_y;
          if old_Dx <= 0 and old_Dy <=0
            and not(iszero(old_Dx) and iszero(old_Dy))
            then
            // possible bifurcation point
                break;
          end_if;
          old_Dx := Df_x;
          old_Dy := Df_y;

          last_x := cur_x;
          last_y := cur_y;

          cur_x := float(cur_x + h*dx);
          cur_y := float(cur_y + h*dy);

          newton2d(f, x, y, Df,cur_x,cur_y,1, Options[Precision],
                   n, fxy_n2d, Df_square_n2d, Dfxy_n2d, hf_n2d);
//        t := t + direction * h * Df_xy;
          t := t + direction * sqrt(rdmult(cur_x-last_x,cur_x-last_x,1)
                                   +rdmult(cur_y-last_y,cur_y-last_y,1));
          clean_startpoints(last_x,last_y,cur_x,cur_y,h);
          res := res.[[cur_x,cur_y,t]];
          h := h*5/(n+1);
//          traperror((h := float(100000 * h * specfunc::abs(
//                ((Dfxy_n2d[1] + Dfxy_n2d[2]) * fxy_n2d * Df_xy^2)
//                /((Df_x + Df_y)*subs(f, x=last_x+h*dx, y=last_y+h*dy) * Df_square_n2d)))));
          h := min(max(h,Options[MinStepsize]),Options[MaxStepsize]);
       end_while;
      ));
      end_for; // direction
      if nops(res) = 2 then
          cur_x := (res[1][1]+res[2][1])/2.0;
          cur_y := (res[1][2]+res[2][2])/2.0;
          newton2d(f, x, y, Df,cur_x, cur_y,
                        1, Options[Precision],
                        n, fxy_n2d, Df_square_n2d, Dfxy_n2d, hf_n2d);
          res := res.[[cur_x,cur_y,(res[1][3]+res[2][3])/2.0]];
      end_if;
        
      userinfo(20, "Found ".expr2text(nops(res))." points on component.");
// xxx this is too much work.  --> collect the points differently
      res := sort(res, (a,b)->bool(op(a,3)<op(b,3)));
      if Options[Splines] then
        res_y := numeric::cubicSpline(op(map(res, 
              x -> [op(x,3),op(x,2)])), Natural);
        res_x := numeric::cubicSpline(op(map(res,
              x -> [op(x,3),op(x,1)])), Natural);
        t := genident();
        dx := minmax(op(map(res,op,1)));
        dx := dx[2] - dx[1];
        dy := minmax(op(map(res,op,2)));
        dy := dy[2] - dy[1];
        dx := min(5*nops(res),50,
                          ceil(100*(max(dx/(xmax-xmin),dy/(ymax-ymin)))));
        dx := max(dx, 3);
          userinfo(20, "Using ".expr2text(dx)." points on curve");
        dummy := plot::Curve2d([res_x(t),res_y(t)],
                t=minmax(op(map(res,op,3))));
        dummy::Grid  := [dx];
	dummy::Title := "";
//             ,map(res,x->plot::Point(op(x,1),op(x,2),
//             hold(Options)=[Color=Options[Colors][currcolor]]))
      else
      
        dummy := plot::Polygon(_seqgen(plot::Point(res[i][1],
                   res[i][2]),i,1..nops(res)));
//        dummy := plot::Group(plot::Polygon(_seqgen(plot::Point(res[i][1],
//                   res[i][2]),i,1..nops(res))),
//                   plot::Pointlist(_seqgen(plot::Point(res[i][1],
//                   res[i][2]),i,1..nops(res))));             
      end_if;
      dummy::Color := Options[Colors][currcolor];
      return(dummy);
    end_proc:

// back to plotfactor

    while op(f, 0) = hold(_power) and is(op(f,2)>0) do
      f := op(f,1);
    end_while;

    Df:=[diff(f,x), diff(f,y)];
    if op(Df[1], 0)=hold(diff) or op(Df[2], 0)=hold(diff) then
      context(hold(error("Functions must be differentiable w.r.t. the indeterminates.")));
    end_if;

    Lstart := [];
    userinfo(10,"Calculating startpoints");
    for x_i from xmin to xmax step ((xmax-xmin)/(Options[Grid][1]-1)) do
      for y_i from ymin to ymax step ((ymax-ymin)/(Options[Grid][2]-1)) do
        cur_x := x_i; cur_y := y_i;
        newton2d(f, x, y, Df,cur_x,cur_y,1, Options[Precision],
                 i_n2d, fxy_n2d, Df_square_n2d, Dfxy_n2d, hf_n2d);
        Lstart := Lstart.[[cur_x, cur_y]];
      end_for;
    end_for;

    Lstart := select(Lstart, a->testtype(a,Type::ListOf(DOM_FLOAT,2,2)));
    Lstart := select(Lstart, a->a[1] >= xmin and a[1] <= xmax
                                and a[2] >= ymin and a[2] <= ymax
                                and abs(float(subs(f, [x=a[1], y=a[2]]))) 
                                    <= Options[Precision]);
// print(map(Lstart, a-> [a[1], a[2], abs(float(subs(f, [x=a[1], y=a[2]])))]));
    Lsol :=[];
    userinfo(15,"Calculated ".expr2text(nops(Lstart))." points");
    while nops(Lstart)>0 do
      [startx,starty] := Lstart[1];
      delete Lstart[1];
      if(startx>=xmin and startx <= xmax
         and starty >= ymin and starty <= ymax) then
        userinfo(20,"Starting at (".
                 expr2text(startx).",".expr2text(starty).")");
        Lsol := Lsol.[plotsingle(startx,starty)];
      end_if;
    end_while;
    return(Lsol);
  end_proc:

  PlotList := [];
  for f in F do
    userinfo(5,"Plotting ".expr2text(f));
    for c in Options[Contours] do
      if Options[Factor] then
        Lf := factor(f-c);
        Lf := Factored::convert_to(Lf,DOM_LIST);
        Lf := [Lf[2*i]$i=1..(nops(Lf)-1)/2];
      else
        Lf := f-c;
        if op(Lf, 0) = hold(_power) and is(op(Lf, 2) > 0) = TRUE then
           Lf := op(Lf, 2);
        end_if;
        if op(Lf, 0) = hold(_mult) then
           Lf := [op(Lf)];
        else
           Lf := [Lf];
        end_if;
      end_if;
      PlotList := PlotList.[op(map(Lf,op@plotfactor))];
      currcolor := modp(currcolor,nops(Options[Colors])) + 1;
    end_for;
  end_for;
  plot::Group(op(PlotList), plot::NullObject2d$2); // , ViewingBox=[xmin,xmax,ymin,ymax]);
end_proc:


