/*  Diese Routine entstand durch Aufteilung der frueheren Routine "plot::data"
    ($Date: 2002/02/28 18:29:01 $  $Author: dellaere $   $Revision: 1.5 $)
    Die Subroutinen "Utilities", welche sich vorher direkt in "plot::data" befanden,
    sind nun in "plot::datatools" abgelegt.
    Aufruf: piechart2d(data,<options>)
            data: List, 1-dimensional Array or, Matrix whith one row or one column
	Neue Option: Moves=[k[1]=M[1],k[2]=M[2],..,k[i]=M[i]], 
	             wobei k[1]..k[i] Indizenummern von 1.. nops(l)und
 	             M[1]..M[i] Zahlen zwischen 0 und 10 sein mssen.
	             
    Alessandro Dell Aere
*/

/*++

    Examples:
    >> export(plot): export(RGB):

    >> plot(piechart2d( [5,12,38,14,25], Moves=[4=6],Titles=[4="Neue Option: Moves"] ))

    >> plot(piechart2d( [5,12,38], Colors=[RoyalBlue,VioletRed,GreenPale] ))
    >> plot(piechart2d( [5,6,12,4],Title="CICHLIDENZUCHT:",
              Titles=[1="Aulonocara",2="Cynotilapia",3="Pseudotropheus",4="Steatocranus"]))
    >> plot(piechart2d(matrix([1,5,3,6,5]))
    >> plot(piechart2d(array(1..5, (1)=9, (2)=8, (3)=5, (4)=8, (5)=5 )))

  
++*/

plot::piechart2d:= proc(l)
    local s, c, i, j, t, m, p, tspec, cspec, mspec, options, a, b, db, obj,
    pi, nl, rgb1, R, ss, scene_default;
begin
   
    if  args(0) < 1 then
	    error("1st argument: expecting data")
    end_if;

        // get options:
        c:= RGB::ColorList; // default color list
	    m:= []; // Moves
        t:= []; // Titles
        options:= [args(2..args(0))];
        s:= nops(options);
        tspec:= cspec:= FALSE;
        for i from 1 to s do
            if type( (p:= options[i]) ) <> "_equal" then
                error("specification of the form 'option = value' expected")
            end_if;
            case op(p,1)
            of hold(Colors) do 
                c:= op(p,2);
                if domtype(c) <> DOM_LIST or nops(c) < nops(l) then
                    error("option 'Colors': list with ".nops(l)." RGB color values expected")
                end_if;
                options[i]:= FAIL;
                cspec:= TRUE;
                break
			of hold(Moves) do
			    m:=[0$nops(l)];
				if max(op(op(op(p,2),j),1)$j=1..nops(op(p,2)))> nops(l) then
				error("option 'Moves': entries of list must have left side values between 1 and ".nops(l))
				end_if;

				for j from 1 to nops(op(p,2)) do
			    	m[op(op(op(p,2),j),1)]:=op(op(op(p,2),j),2);
                end_for; //j

                if domtype(m) <> DOM_LIST or nops(op(p,2)) > nops(l)then
				error("option 'Moves': list with max. ".nops(l)." entries of the form 'a=b' expected")
				end_if;
			    if min(op(op(op(p,2),j),2)$j=1..nops(op(p,2)))< 0
				      or max(op(op(op(p,2),j),2)$j=1..nops(op(p,2)))> 10 then
				error("option 'Moves': entries of list must have right side values between 0 and 10")
                end_if;
                options[i]:= FAIL;
                mspec:= TRUE;
                break
            of hold(Titles) do 
			    t:=[""$nops(l)];
				if max(op(op(op(p,2),j),1)$j=1..nops(op(p,2)))> nops(l) then
				error("option 'Titles': entries of list must have left side values between 1 and ".nops(l))
				end_if;

			    for j from 1 to nops(op(p,2)) do 
                    t[op(op(op(p,2),j),1)]:=op(op(op(p,2),j),2);
                end_for; //j

                if not testtype(t,Type::ListOf(DOM_STRING,1,nops(l) )) then
                error("option 'Titles': list with max. ".nops(l)." entries of the form 'a=<string>' expected")
                end_if;
				
                options[i]:= FAIL;
                tspec:= TRUE;
            end_case;

            if cspec = TRUE and tspec = TRUE and mspec = TRUE then break end_if;
        end_for;
        options:= subs( options,FAIL=null() );
    

    if testargs() 
    then if not domtype(l)=DOM_LIST
            and not domtype(l) = DOM_ARRAY 
	    and not l::dom::hasProp(Cat::Matrix) = TRUE
         then error("1st argument: list or array or matrix expected")
         end_if;
	 if domtype(l)=DOM_LIST
	 then if not testtype(float(l), Type::ListOf(DOM_FLOAT)) 
              then error("1st argument: entries of list must be numerical values")
              end_if:
	 elif domtype(l) = DOM_ARRAY 
         then if op(l,[0,1]) <> 1 
	      then error("1st argument: array must be one-dimensional")
              end_if;
	 elif l::dom::hasProp( Cat::Matrix ) = TRUE  
	 then if op(l,[0,2,2]) <> 1 and op(l,[0,3,2]) <> 1
	      then error("1st argument: matrix must have one row or one column")
	      end_if;
	 end_if;

    end_if; 
 
     //--------------------------------------------
     // Matrix, Array --> List :
     nl := [];
     if domtype(l) = DOM_ARRAY // Array -> List 
     then for i from op(l, [0, 2, 1]) to op(l, [0, 2, 2]) do
              nl := append(nl, l[i])
	  end_for;
          l := nl;
     elif l::dom::hasProp( Cat::Matrix ) = TRUE // Matrix -> List
     then if op(l,[0,2,2]) <> 1 	   
	  then l := linalg::transpose(l);
          end_if;
	  for i from 1 to op(l,[0,3,2]) do 
	     nl:=append(nl, l[1,i])
	  end_for;
              l := nl;          
     end_if;
    //-----------------------------------------------

    if  (_or(is(l[i] < 0) $i=1..nops(l) )) 
    then error("1st argument: entries of list must be non-negative")
    end_if;

    l := float(l):
    rgb1 := c -> [max(0,c[1]-0.2), max(0,c[2]-0.2), max(0,c[3]-0.2)];
    ss := _plus(op(l)); obj := null(); 
    pi := float(PI); b := 0; 
    t := plot::data_set_titles(t, l, nops(l));
	R:=[0$nops(l)];
    
    if m=[] 
	then
	    for i from 1 to nops(l) do
          db:= l[i]/ss*2*pi;
          a := b; b := b+db; 	
	      R[i]:=plot::Ellipse2d([0,0],1,1, 
			                    Filled=TRUE, hold(Grid)=[ceil(db*6)], hold(Color)=[hold(Flat),
			                	c[i]], hold(Title) = t[i],
			                    hold(TitlePosition) = [5+4*cos((a+b)/2),5.4-4.6*sin((a+b)/2)] );
	      obj:=obj,R[i];
		  (R[i])::range:=a..b;
        end_for;//i
	else

        for i from 1 to nops(l) do
          db:= l[i]/ss*2*pi;
          a := b; b := b+db; 	
	      R[i]:=plot::Ellipse2d([m[i]/25*cos((a+b)/2),m[i]/25*sin((a+b)/2)],1,1,
		                       Filled=TRUE, hold(Grid)=[ceil(db*6)], hold(Color)=[hold(Flat),
							   c[i]], hold(Title) = t[i],
							   hold(TitlePosition) = [5+(3.5+m[i]/50)*cos((a+b)/2),5.4-(4+m[i]/50)*sin((a+b)/2)] );
	    	 obj:=obj,R[i];
		     (R[i])::range:=a..b;
        end_for;//i
    end_if;//m

    // set options:
    scene_default:= 
        [hold(Axes) = hold(None), hold(Scaling) = hold(Constrained),hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[3]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(obj, op(options)) )
end_proc:





