/*  Diese Routine entstand durch Aufteilung der frueheren Routine "plot::data"
    ($Date: 2002/02/28 18:29:01 $  $Author: dellaere $   $Revision: 1.8 $)
    Die Subroutinen "Utilities", welche sich vorher direkt in "plot::data" befanden,
    sind nun in "plot::datatools" abgelegt.
    Aufruf: piechart3d(data,<options>)  
            data: List, 1-dimensional Array or, Matrix whith one row or one column

    Alessandro Dell Aere 2001/01/28
*/

/*++

    Examples:
    >> export(plot): export(RGB):

    >> plot(piechart3d( [5,12,38,14,25] ))

    >> plot(piechart3d( [5,12,38], Colors=[RoyalBlue,VioletRed,GreenPale] ))
    >> plot(piechart3d( [5,6,12,4,8,8,11],Title="CICHLIDENZUCHT:",
            Titles=[ 1="Aulonocara",     2="Cynotilapia",
		     3="Pseudotropheus", 4="Steatocranus",
                     5="Cyrtocara",      6="Labidochromis",
					 7="Andere"                       ] ))


  ++*/

plot::piechart3d:= proc(l)
   local s, c, i, j, t, p, tspec, cspec, options, a, b, db, fsin, fcos, 
	      grid, obj, nl, pi, rgb1, ss, u, v, segment_angle, scene_default; 
begin
    if args(0) < 1 then
        error("1st argument: expecting data")
    end_if;
        

        // get options:
        c:= RGB::ColorList; // default color list
        t:= []; // Titles
        options:= [args(2..args(0))];
        s:= nops(options);
        tspec:= cspec:= FALSE;
        for i from 1 to nops(options) do
            if type( (p:= options[i]) ) <> "_equal" then
                error("specification of the form 'option = value' expected")
            end_if;
            case op(p,1)
            of hold(Colors) do 
                c:= op(p,2);
                if domtype(c) <> DOM_LIST or nops(c) < nops(l) then
                    error("option 'Colors': list with ".nops(l)." RGB color values expected")
                end_if;
                options[i]:= FAIL;
                cspec:= TRUE;
                break
            of hold(Titles) do 
                t:=[""$nops(l)];
				if max(op(op(op(p,2),j),1)$j=1..nops(op(p,2)))> nops(l) then
				error("option 'Titles': entries of list must have left side values between 1 and ".nops(l))
				end_if;

			    for j from 1 to nops(op(p,2)) do 
                    t[op(op(op(p,2),j),1)]:=op(op(op(p,2),j),2);
                end_for; //j

                if not testtype(t,Type::ListOf(DOM_STRING,1,nops(l) )) then
                error("option 'Titles': list with max. ".nops(l)." entries of the form 'a=<string>' expected")
                end_if;
                options[i]:= FAIL;
                tspec:= TRUE;
            end_case;

            if cspec = TRUE and tspec = TRUE then break end_if;
        end_for;
        options:= subs( options,FAIL=null() );


    if testargs() 
    then if not domtype(l)=DOM_LIST
            and not domtype(l) = DOM_ARRAY 
	    and not l::dom::hasProp(Cat::Matrix) = TRUE
         then error("1st argument: list or array or matrix expected")
         end_if;
	 if domtype(l)=DOM_LIST
	 then if not testtype(float(l), Type::ListOf(DOM_FLOAT)) 
              then error("1st argument: entries of list must be numerical values")
              end_if:
	 elif domtype(l) = DOM_ARRAY 
         then if op(l,[0,1]) <> 1 
	      then error("1st argument: array must be one-dimensional")
              end_if;
	 elif l::dom::hasProp( Cat::Matrix ) = TRUE  
	 then if op(l,[0,2,2]) <> 1 and op(l,[0,3,2]) <> 1
	      then error("1st argument: matrix must have one row or one column")
	      end_if;
	 end_if;       
    end_if; 

     //--------------------------------------------
     // Matrix, Array --> List :
     nl := [];
     if domtype(l) = DOM_ARRAY // Array -> List 
     then for i from op(l, [0, 2, 1]) to op(l, [0, 2, 2]) do
              nl := append(nl, l[i])
	  end_for;
          l := nl;
     elif l::dom::hasProp( Cat::Matrix ) = TRUE // Matrix -> List
     then if op(l,[0,2,2]) <> 1 	   
	  then l := linalg::transpose(l);
          end_if;
	  for i from 1 to op(l,[0,3,2]) do 
	     nl:=append(nl, l[1,i])
	  end_for;
              l := nl;          
     end_if;

    //-----------------------------------------------

    if (_or(is(l[i] < 0) $i=1..nops(l)))
    then error("1st argument: entries of list must be non-negative")
    end_if;

    l := float(l):
    rgb1 := proc(c)
    begin
        c[1] := max(0,c[1]-0.2);
        c[2] := max(0,c[2]-0.2);
        c[3] := max(0,c[3]-0.2);
        c
    end_proc:

    [u,v] := [genident(),genident()];
     
    fsin:= slot(sin, "float"): 
    fcos:= slot(cos, "float"): 
 
    ss:= _plus(op(l)); obj := null(); 
    pi:= float(PI); b := 0; 
    t:= plot::data_set_titles(t, l, nops(l)*2);
    // divide each slice into segements of 6 degrees each.
    // This angle is small enough to generate a smooth slice
    segment_angle:= 0.1047197551; // = float(6/180*PI) = 6 degrees

    for i from 1 to nops(l) do
        db:= l[i]/ss*2*pi;
        // use at least 2 segments per slice, i.e., grid >=3
        grid:= max(3, round(db/segment_angle)); 
        a:= b; b:= b + db;
        obj:= obj, plot::Surface3d(
            [v*fcos(u),v*fsin(u), 0],u=[a,b],v=[0,1],
            hold(Style) = [hold(ColorPatches),hold(Only)], hold(Color) = [hold(Flat), c[i]],
            hold(Grid) = [grid,2], hold(Title)=t[i],
			hold(TitlePosition)=[5+4*cos((a+b)/2),5-4*sin((a+b)/2)]
        );
        obj:= obj, plot::Surface3d([fcos(u),fsin(u), v], u=[a,b], v=[-0.2,0],
            hold(Style) = [hold(ColorPatches),hold(Only)],
		    hold(Color) = [hold(Flat), rgb1(c[i])],
            hold(Grid)=[grid,2]
	  
        );
    end_for;

    // set options:
    scene_default:= [hold(Axes) = hold(None), hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[2]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(obj, op(options),CameraPoint=[0,-10,10]) );

end_proc:





