/*  Diese Routine entstand durch Aufteilung der frueheren Routine "plot::data"
    ($Date: 2001/11/10 16:12:16 $  $Author: walter $   $Revision: 1.6 $)
   
    Aufruf: surface(data, <options>)

    data: List of Lists, Matrix or Array
    options: DataPoints = RGB-Color  
             alle Optionen von plot::Function3d 
             alle Scene Optionen (siehe plotOptions3d.tex)

    2001/10.Okt. dellaere

 Examples:
    >> export(RGB): export(plot): 
       L := [[2,1,2,1,2],[1,2,1,2,1],[2,1,3,1,2],[1,2,1,2,1],[2,1,2,1,2]]:
 
    >> plot(plot::surface(L, Color=[Height,RGB::Green,RGB::Blue],
                             Smoothness=[12,12], Axes=None, Grid=[2,12]))
    >> plot(surface(L));
    >> plot(surface(matrix([[5,2,2],[2,3,2],[8,2,0]])));
    >> plot(surface(array(1..2, 1..3,
                          (1,1) = sqrt(2), (1,2) = 2, (1,3) = 2, 
                          (2,1) = sin(20), (2,2) = 3, (2,3) = 2)));

    >> plot(surface(L, Style=[WireFrame, Mesh]));
    >> plot(surface(L, Style=[WireFrame, ULine], DataPoints=Black));
    >> plot(surface(L, Style=[ColorPatches,AndMesh],
                       CameraPoint=[7,20,25], Color=[Flat, RGB::Yellow],
                       LineWidth=40, DataPoints=Black));

    >> plot(surface([[5,7,1],[3,4,1],[4,5,1],[3,-2,1],[0,0,1]]));
    >> plot(surface([[0,1,1.5,1,0],[1,1.5,1,0,1],[1.5,1,0.5,1,1.5]
                    ,[1,1.5,1,0,1],[0,1,1.5,1,0]]));
    >> plot(surface([[5,10,24,-3],[6,5,2,18],[19,45,12,-10]]));
    >> plot(surface([[5,10,24,-3],[6,5,2,18]]);
    >> plot(surface([[float(sin(i/10*4*PI)*cos(j/10*2*PI))$i=0..10]$j=0..20]));


Call(s):

     Probleme:
     -- array(nestedlist) wird anders (falsch) verarbeitet als
        nestedlist.
	--> done, dellaere 17.10.2001

     -- Rueckgabe waere (eigentlich) von plot::Scene auf
        plot::Group umzustellen. Problem dann: nicht alle
        Szene-Optionen sind dann mehr zulaessig.
        (Empfehlung: momentan nicht anpassen)



     -- Implementierungsbedarf fuer "Surface": wir brauchen
        noch eine 2d-Spline-Approximation von Daten, um das
        unselige (und langsame) numeric::lagrange ersetzen
        zu koennen. In "Curves" ist dies bereits geschehen:
        altes numeric::lagrange --> numeric::cubicSpline.  
        Zudem stimmt die Farbverwaltung nicht!
        --> done 2001/10.Okt. dellaere
*/

plot::surface := proc(l)

    local i, j, m, n, x, y, a, b, nli, nlj, p, L, P,
          interpoly, scene_default, f_default,
	  pcolor, options, scene_options, f_options;

begin
    if args(0) < 1 then
        error("expecting data")
    end_if;
   
    if testargs() 
    then if not testtype(l, Type::ListOf(DOM_LIST)) 
            and not domtype(l) = DOM_ARRAY 
	    and not l::dom::hasProp(Cat::Matrix) = TRUE
         then error("1st argument: list of lists or two-dimensional array or matrix expected")
         end_if;

         if testtype(l, Type::ListOf(DOM_LIST)) 
         then if _or(nops(l[i]) <> nops(l[i + 1])$i = 1..nops(l) - 1) 
              then error("1st argument: lists must have the same number of elements")
              end_if;
         elif domtype(l) = DOM_ARRAY 
         then if op(l,[0,1]) <> 2 
	      then error("1st argument: array must be two-dimensional")
              end_if;
         end_if;
     end_if;

     //--------------------------------------------
     // Matrix, Array --> List of Lists :
     if l::dom::hasProp( Cat::Matrix ) = TRUE then
        l:= expr(l): // Matrix -> Array
     end_if;
     nli := [];
     if domtype(l) = DOM_ARRAY // then Array -> List of Lists
     then (a[i] := op(l, [0, i + 1, 1])) $ i = 1..2;
          (b[i] := op(l, [0, i + 1, 2])) $ i = 1..2;
          for i from a[1] to b[1] do
            nlj := [];
            for j from a[2] to b[2] do
               nlj := append(nlj, l[i,j])
            end_for;
            nli := append(nli, nlj)
          end_for;   
          l := nli;
     end_if;
    //---------------------------------------------

     l := float(l);
     if not testtype(l, Type::ListOf(Type::ListOf(Type::Real, 1), 1)) 
     then error("1st argument: values must be real numbers")
     end_if;
       

     m := nops(l); 
     n := nops(l[1]); 
     scene_options := [];
     f_options := []; 
     P := null();
     // get options:
     options := [args(2..args(0))];
     pcolor := FAIL; //No Points
     for i from 1 to nops(options) do

        if type( (p := options[i]) ) <> "_equal" 
        then error("specification of the form 'option = value' expected")
	 
        //----------------------------------------------------------------
	// Split List of Options (options) into 2 Lists,
        // one for Scene-Options (scene_options)
        // and one for Function3d-Options (f_options).
	// If Option DataPoints is used, handle it seperatly (pcolor). : 
        elif contains({Arrows, Axes, AxesOrigin, AxesScaling, AxesInFront, BackGround,
	               CameraPoint, FocalPoint, FontFamily, FontSize, 
                       FontStyle, ForeGround, Labeling, Labels, /*LineStyle,
		       LineWidth,*/ PlotDevice, /*PointStyle, PointWidth,*/ 
		       Scaling, Ticks, Title, TitlePosition, ViewingBox},
		       op(p,1)) 
        then scene_options := append(scene_options, p)
	elif contains ({Color, Grid, LineStyle, LineWidth, PointStyle,
                        PointWidth, Smoothness, Style /*, Title, TitlePosition*/ },
		        op(p,1)) 
	then f_options := append(f_options, p)
	elif op(p,1) = hold(DataPoints) 
	then pcolor := op(p,2);
	     if domtype(pcolor) <> DOM_LIST 
	        or nops(pcolor) <> 3 or pcolor[1] < 0 or pcolor[1] > 1
		or pcolor[2] < 0 or pcolor[2] > 1 or pcolor[3] < 0 
		or pcolor[3] > 1 
	     then error("option 'DataPoints': RGB Color expected")
             end_if;
	else error("Unknown Option '".op(p,1)."'")
	end_if;
     end_for;
     //----------------------------------------------------------------
  
    // set options:   
    scene_default:= [hold(Axes) = hold(Box), hold(Scaling) = hold(UnConstrained), 
                     hold(Ticks) = 0, hold(Title) = "", hold(Labeling) = FALSE];
    f_default := [hold(Grid) = [m,n],
                 hold(Smoothness) = [min(4, 100 div m), min(4, 100 div n)]];

	  
    if scene_options <> [] then
       if has(scene_options, hold(Axes)) then scene_default[1] := FAIL end_if;
       if has(scene_options, hold(Scaling)) then scene_default[2] := FAIL end_if;
       if has(scene_options, hold(Ticks)) then scene_default[3] := FAIL end_if;
       if has(scene_options, hold(Title)) then scene_default[4] := FAIL end_if;
       if has(scene_options, hold(Labeling)) then scene_default[5] := FAIL end_if;
       scene_options := subs( scene_default, FAIL = null()) . scene_options;
    else scene_options := scene_default
    end_if;

    if f_options <> [] then
       if has(f_options, hold(Grid)) then f_default[1] := FAIL end_if;
       if has(f_options, hold(Smoothness)) then f_default[2] := FAIL end_if;
       f_options := subs( f_default, FAIL = null()) . f_options;
    else f_options := f_default;
    end_if;

    interpoly := numeric::cubicSpline2d(
                   [i$i = 0..m - 1],
                   [j$j = 0..n - 1],
                   array(0..m - 1, 0..n - 1, l)); 
    x := genident();
    y := genident();
    L := plot::Function3d(interpoly(x, y),
                          x = 0..m - 1,
			  y = 0..n - 1,
		          op(f_options));

    if pcolor <> FAIL
    then P := (plot::Point(i, j, l[i + 1][j + 1],
                           hold(Color) = [Flat, pcolor],
		           hold(PointWidth) = 40,
			   hold(PointStyle) = hold(FilledCircles))
		          $ i = 0..m - 1) $ j = 0..n - 1 

    end_if;
 // return(plot::Group(P, L, op(scene_options))); 
    return(plot::Scene(P, L, op(scene_options))); 
     

end_proc: 

