
plotlib::gen_color_func := proc(dim, obj, MinV = [], MaxV = [])
   local applyArray, theColor, MinVals, MaxVals;
begin
   [applyArray, theColor, MinVals, MaxVals] 
      := plotlib::setupPlotItem(obj, dim, MinV, MaxV);
   Color = plotlib::plotItem(applyArray, theColor, MinVals, MaxVals)
end_proc:

plotlib::plotItem := proc(applyArray, theColor,theMinVals,theMaxVals)
  option escape;
begin
  if (theColor <> 0) and (op(theColor,1) = Function) and (contains(applyArray,plotlib::resetColor) <> 0) then 
    subsop(theColor,2=plotlib::resetColorProc(theMinVals,theMaxVals,op(theColor,2),[0.1,0.9]))
  else
    theColor
  end_if
end_proc:

plotlib::isColorVal := proc(colorValue)
begin
  if (domtype(colorValue) = DOM_LIST) and (nops(colorValue) = 3) 
          and (testtype(colorValue,Type::ListOf(Type::Constant))) then
    if (float(colorValue[1]) <= 1.0 and float(colorValue[1]) >= 0.0 and
        float(colorValue[2]) <= 1.0 and float(colorValue[2]) >= 0.0 and
        float(colorValue[3]) <= 1.0 and float(colorValue[3]) >= 0.0) then
      return(TRUE):
    end_if:
  end_if:
  return(FALSE):
end_proc:

plotlib::HSLtoRGB := proc(hueTriple)
local maxSL, maxS, bErrorInput, rgbTriple, Lfactor, diagVal, i;
begin
  bErrorInput := FALSE:
  if domtype(hueTriple) <> DOM_LIST or nops(hueTriple) <> 3 then
    bErrorInput := TRUE:
  else
    for i from 1 to 3 do
      if (hueTriple[i] < 0.0) or (hueTriple[i]>1.0) then
        bErrorInput := TRUE:
      end_if:
    end_for:
  end_if:
  
  if bErrorInput = TRUE then
    error("Input to HSLtoRGB must be a triple of color values between 0 and 1!"):
  end_if:
  
  maxS := [0.0,0.0,0.0]:
  if (hueTriple[1] <= float(1/6)) then
    maxS := [1.0, 6*hueTriple[1], 0.0]:
  elif (hueTriple[1] <= float(1/3)) then
    maxS := [2.0 - 6*hueTriple[1], 1.0, 0.0]:
  elif (hueTriple[1] <= 0.5) then
    maxS := [0.0, 1.0, 6*hueTriple[1]-2.0]:
  elif (hueTriple[1] <= float(2/3)) then
    maxS := [0.0, 4.0 - 6*hueTriple[1], 1.0]:
  elif (hueTriple[1] <= float(5/6)) then
    maxS := [6*hueTriple[1] - 4.0, 0.0, 1.0]:
  else
    maxS := [1.0, 0.0, 6.0 - 6*hueTriple[1]]:
  end_if:
  Lfactor := 2.0 * hueTriple[3]:
  diagVal := (1.0 - hueTriple[2]) * hueTriple[3]:
  if (Lfactor > 1.0) then
    diagVal := diagVal + (hueTriple[2] * (Lfactor - 1.0)):
    Lfactor := 2.0 - Lfactor:
  end_if:
  Lfactor := Lfactor * hueTriple[2]:
  rgbTriple := [diagVal, diagVal, diagVal]:
  for i from 1 to 3 do
    rgbTriple[i] := diagVal + Lfactor * maxS[i]:
  end_for:
  rgbTriple;
end_proc:

plotlib::RGBtoHSL := proc(rgbTriple)
local maxSL, maxS, outH, outS, outL, bErrorInput, i;
begin
  bErrorInput := FALSE:
  if domtype(rgbTriple) <> DOM_LIST or nops(rgbTriple) <> 3 then
    bErrorInput := TRUE:
  else
    for i from 1 to 3 do
      if (rgbTriple[i] < 0.0) or (rgbTriple[i]>1.0) then
        bErrorInput := TRUE:
      end_if:
    end_for:
  end_if:
  
  if bErrorInput = TRUE then
    error("Input to RGBtoHSL must be a triple of color values between 0 and 1!"):
  end_if:

  minIndex := 1:
  maxIndex := 1:
  for i from 2 to 3 do
    if (rgbTriple[i] < rgbTriple[minIndex]) then
      minIndex := i:
    elif (rgbTriple[i] > rgbTriple[maxIndex]) then
      maxIndex := i:
    end_if:
  end_for:
  outL := 0.5*(rgbTriple[minIndex] + rgbTriple[maxIndex]):
  if (rgbTriple[minIndex] = rgbTriple[maxIndex]) then
    outH := float(2/3):
    outS := 0.0:
  else
    maxS := [0.0,0.0,0.0]:
    for i from 1 to 3 do
      maxS[i] := (rgbTriple[i] - rgbTriple[minIndex])/(rgbTriple[maxIndex] - rgbTriple[minIndex]):
    end_for:
    if minIndex = 1 then
      if maxIndex = 2 then
        outH := float(1/3) + maxS[3]/6:
      else
        outH := float(2/3) - maxS[2]/6:
      end_if:
    elif minIndex = 2 then
      if maxIndex = 1 then
        outH := 1.0 - maxS[3]/6:
      else
        outH := float(2/3) + maxS[1]/6:
      end_if:
    else
      if maxIndex = 1 then
        outH := maxS[2]/6:
      else
        outH := float(1/3) - maxS[1]/6:
      end_if:
    end_if:
    if outL <= 0.5 then
      outS := (rgbTriple[maxIndex] - rgbTriple[minIndex])/(2.0 * outL):
    else
      outS := (rgbTriple[maxIndex] - rgbTriple[minIndex])/(2.0 - (2.0 * outL)):
    end_if: 
  end_if:
  [outH,outS,outL];  
end_proc:

plotlib::linearColorProc := proc()
local ii,proportion,minColor,maxColor;
begin
  minColor := [0.0,0.0,0.0]:
  maxColor := [0.0,0.0,0.0]:
  proportion := plotlib::computeProportion(args(ii) $ ii=1..args(0)):
  [minColor[ii] + proportion * (maxColor[ii] - minColor[ii]) $ ii = 1..3];
end_proc:

plotlib::HSLColorProc := proc()
local ii, proportion, minHSLColor, maxHSLColor,colorVals;
begin
  minColor := [0.0,0.0,0.0]:
  maxColor := [0.0,0.0,0.0]:
  proportion := plotlib::computeProportion(args(ii) $ ii=1..args(0)):
  colorVals := [minColor[ii] + proportion * (maxColor[ii] - minColor[ii]) $ ii = 1..3]:
  if colorVals[1] > 1.0 then
    colorVals[1] := colorVals[1] - 1.0:
  end_if:
  plotlib::HSLtoRGB(colorVals);
end_proc:


//Expecting color specification of one of the following forms:
//   (1) [Flat,[0.9,0.8,0.7]]
//   (2) [Height,[0.9,0.8,0.7],[0.1,0.2,0.3]]
//   (3) [Rainbow,[0.9,0.8,0.7],[0.1,0.2,0.3]]
//   (4) [Monochrome,[0.9,0.8,0.7]]
//   (5) [Function,colorProcName]
//   (6) [Height,[0.9,0.8,0.7],[0.1,0.2,0.3],ColorAxes=[X,Y,Z]]
//   (7) [Rainbow,[0.9,0.8,0.7],[0.1,0.2,0.3],ColorAxes=[Y,U,V]]
//   (8) [Monochrome,[0.9,0.8,0.7],ColorAxes=[X,U]] 
plotlib::presetColorSpec := proc(colorSpec,theDim,bUseProc)
local bError, outType, outSpec, startVal, endVal, i, j, colorAxes, minColorVal, maxColorVal, colorAxesArray, bUsesAxes, refArray, inType;
option escape;
begin
  //First, read in values and check for malformed params:
  bError := TRUE:
  colorAxes := []:
  colorAxesArray := []:
  outType := Height:
  bUsesAxes := FALSE:
  if (args(0) = 3) and (domtype(colorSpec) = DOM_LIST) then
    inType := op(colorSpec,1):
    case inType
      of Rainbow do
        outType := Function:  //and fall through
        bUsesAxes := TRUE:
        if (nops(colorSpec) = 3) and (plotlib::isColorVal(op(colorSpec,2)) = TRUE) and (plotlib::isColorVal(op(colorSpec,3)) = TRUE) 
                   and (bUseProc <> TRUE) then
          bError := FALSE:
          startVal := op(colorSpec,2):
          endVal := op(colorSpec,3):
        elif (nops(colorSpec) = 4 and plotlib::isColorVal(op(colorSpec,2)) = TRUE and plotlib::isColorVal(op(colorSpec,3)) 
                     and (type(op(colorSpec,4)) = "_equal") and (op(colorSpec,[4,1]) = ColorAxes)) or (bUseProc=TRUE) then
          bError := FALSE:
          outType := Function:
          bUsesAxes := TRUE:
          if nops(colorSpec) = 4 then
            colorAxes := op(colorSpec,[4,2]):
          end_if:
          startVal := op(colorSpec,2):
          endVal := op(colorSpec,3):
        end_if:
        if bError = FALSE then
          if inType = Rainbow then
            minColorVal := plotlib::RGBtoHSL(startVal):
            maxColorVal := plotlib::RGBtoHSL(endVal):
            if abs(maxColorVal[1] - minColorVal[1]) < 0.5 then
              if maxColorVal[1] >= minColorVal[1] then
                minColorVal[1] := minColorVal[1] + 1.0:
              else
                maxColorVal[1] := maxColorVal[1] + 1.0:
              end_if:
            end_if:
          else
            minColorVal := startVal:
            maxColorVal := endVal:
          end_if:
        end_if:
      break:
      of Monochrome do
        if (nops(colorSpec) = 2) and (plotlib::isColorVal(op(colorSpec,2)) = TRUE) and (bUseProc <> TRUE) then
          bError := FALSE:
          startVal := op(colorSpec,2):
          outType := Height:
        elif (nops(colorSpec) = 3 and (plotlib::isColorVal(op(colorSpec,2)) = TRUE) and (type(op(colorSpec,3)) = "_equal") 
                     and (op(colorSpec,[3,1]) = ColorAxes)) or (bUseProc = TRUE) then
          bError := FALSE:
          outType := Function:
          bUsesAxes := TRUE:
          if nops(ColorSpec) = 3 then
            colorAxes := op(colorSpec,[3,2]): 
          end_if:
          startVal := op(colorSpec,2):
        end_if:
        if bError = FALSE then
          minColorVal := plotlib::RGBtoHSL(startVal):
          maxColorVal := minColorVal:
          minColorVal[3] := 0.1:
          maxColorVal[3] := 0.9:
          minColorVal := plotlib::HSLtoRGB(minColorVal):
          maxColorVal := plotlib::HSLtoRGB(maxColorVal):
        end_if:
      break:
    end_case:
  end_if:
  if bError = TRUE then
    error("Invalid arguments to setupColorProc!");
  end_if:
    
  if theDim = 3 then
    refArray := [X,Y,Z,U,V]:
  else
    refArray := [X,Y,U]:
  end_if:
  for i from 1 to nops(colorAxes) do
    j := contains(refArray,colorAxes[i]):
    if j <> 0 then
      colorAxesArray := append(colorAxesArray,j):
    end_if:
  end_for:
  if nops(colorAxesArray) = 0 then
    colorAxesArray := [theDim]:
  end_if:

  case outType
    of Function do
      if colorSpec[1] = Rainbow then
        outSpec := [Function,subsop(plotlib::HSLColorProc,[4,1,2]=minColorVal,[4,2,2]=maxColorVal, [4,3,2,0]=subsop(plotlib::computeProportion, [4,1,2] = colorAxesArray))]:
      else
        outSpec := [Function,subsop(plotlib::linearColorProc,[4,1,2]=minColorVal,[4,2,2]=maxColorVal,[4,3,2,0]=subsop(plotlib::computeProportion, [4,1,2] = colorAxesArray))]:
      end_if:
    break:
  end_case:
  
  [outSpec,bUsesAxes];
end_proc:


plotlib::resetProportionProc := proc(theProc,minVals,maxVals,compressionCoefficients)
option escape;
local ii,offset,dampTerm,srcDiff,theStart,theEnd,theDampTerm,retProc,theAxes;
begin
  offset := [0.0,0.0,0.0,0.0,0.0]:
  dampTerm := [0.0,0.0,0.0,0.0,0.0]:
  theStart := 2.0 * float(compressionCoefficients[1]) - 1.0:
  theEnd := 2.0 * float(compressionCoefficients[2]) - 1.0:
  if theEnd >= 1.0 then
    theEnd := 0.999:
  end_if:
  srcDiff := 1.0:
  theAxes := op(theProc,[4,1,2]):
  theDampTerm := (theEnd / sqrt(1 - theEnd * theEnd)) - (theStart / sqrt(1 - theStart * theStart)):
  if theDampTerm < 1.0e-06 then
    theDampTerm := 1.0e-06:
  end_if:
  for ii in theAxes do
    if (ii <= nops(minVals)) and (ii <= nops(maxVals)) and (minVals[ii] <> []) and (maxVals[ii] <> []) then
      srcDiff := maxVals[ii] - minVals[ii]:
      dampTerm[ii] := srcDiff / theDampTerm:
      offset[ii] := maxVals[ii] - dampTerm[ii] * (theEnd / sqrt(1 - theEnd * theEnd)):
    end_if:
  end_for:
  
  retProc := subsop(theProc, [4,2,2] = offset, [4,3,2] = dampTerm);
end_proc:


plotlib::resetColorProc := proc(minVals,maxVals,colorFunc,compressionCoefficients)
local retProc;
option escape;
begin
  retProc := eval(op(colorFunc,[4,3,2,0])):
  retProc := plotlib::resetProportionProc(retProc,minVals,maxVals,compressionCoefficients):
  retProc := subsop(colorFunc,[4,3,2,0]=retProc); 
end_proc:

plotlib::computeProportion := proc()
local ii,jj,theAxes,theDiff,dampTerm,offset,proportion;
option escape;
begin
  theAxes := [1,2,3]:
  offset := [0.0,0.0,0.0,0.0,0.0]:
  dampTerm := [1.0,1.0,1.0,1.0,1.0]:
  proportion := 0.0:
  theDiff := 0.0:
  jj := 0.0:
  for ii in theAxes do
    if ii <= args(0) then
      theDiff := float(args(ii)) - offset[ii]:
      if (theDiff <> 0.0 or dampTerm[ii] <> 0.0) then
        proportion := proportion + 0.5 + 0.5 * (theDiff / sqrt((theDiff * theDiff) + (dampTerm[ii] * dampTerm[ii]))):
        jj := jj + 1.0:
      end_if:
    end_if:
  end_for:
  if jj <> 0.0 then
    proportion := proportion/jj:
  end_if:
  proportion;
end_proc:

//This proc expects: 
//  (i)"theFunc" - A function (expression or proc), which will likely though not necessarily return a List;
// (ii)"rangeList" - A List of Range specifications, each of the form <varname> = [<minimum>,<maximum>];
//(iii)"gridList" - A corresponding List of grid specifications, simply as numbers, as in [<numGridPoints1>,<numGridPoints2>,...];
//(iv) optionally, a specification "computeExtrema = [<indexExpr1>,<indexExpr2>,...]", where each index expression is either a
//         nonnegative integer or an array of such. Each of these will be used as the second argument to a call "op(functionResult,indexExpr)"
//         so they should represent valid indices into the result Lists generated by applying "theFunc" to the source
//         grid points (or had better just be "1" if the results are scalars!).
//
//It returns a List of arrays (DOM_LISTs), the members of which are:
//   (i) an array of the result values of applying "theFunc";
//  (ii) an array of the points at which "theFunc" was applied;
// (iii) an array of 0's and 1's, with 0's appearing where errors occurred during evaluation;
// (iv) optionally (if "computeExtrema" was passed in), an array of minimum values - one for each element of "computeExtrema";
//  (v) optionally (if "computeExtrema" was passed in), an array of maximum values - one for each element of "computeExtrema".
//Note that each of the first three arrays should be precisely the same shape.
plotlib::computeGridPoints := proc(theFunc,rangeList,gridList)
local ii, jj, kk, startPt, endPt, currPt, currVal, currValArray, incrementList, bDone, 
      epsValArray, currPtsArray, defValArray, extremaIndices, minArray, maxArray, compareVal, bFlatten;
begin
  if (domtype(rangeList) <> DOM_LIST) or (type(rangeList[1]) <> "_equal")
       or (domtype(op(rangeList,[1,2])) <> DOM_LIST)
       or (domtype(gridList) <> DOM_LIST) 
       or (nops(gridList) <> nops(rangeList)) then
    error("Mismatched arguments to computeGridPoints!");
  end_if:
  if (domtype(theFunc)<>DOM_PROC) and (domtype(theFunc)<>DOM_FUNC)
        and (domtype(theFunc)<>DOM_EXEC) then
    if (domtype(theFunc) = DOM_LIST) and (contains([domtype(op(theFunc,jj)) $ jj=1..nops(theFunc)],piecewise) <> 0) then
      theFunc := [fp::unapply(theFunc[jj],op(rangeList,[ii,1]) $ ii=1..nops(rangeList)) $ jj=1..nops(theFunc)]:
    else
      theFunc := fp::unapply(theFunc,op(rangeList,[ii,1]) $ ii=1..nops(rangeList));
    end_if:
  end_if:
  extremaIndices := []:
  bFlatten := FALSE:
  for ii from 4 to args(0) do
    if type(args(ii)) = "_equal" then
      case op(args(ii),1)
        of computeExtrema do
          extremaIndices := op(args(ii),2):
//An array of indices into our output data; which output indices we're supposed to return extrema of.
        break:
        of flatten do
          bFlatten := op(args(ii),2):
        break:
      end_case:
    end_if:
  end_for:
  jj := nops(extremaIndices):
  minArray := [[] $ ii=1..jj]:
  maxArray := [[] $ ii=1..jj]:
  jj := nops(rangeList):
  bDone := FALSE:
  startPt := float([op(rangeList,[ii,2,1]) $ ii = 1..jj]):
  endPt := float([op(rangeList,[ii,2,2]) $ ii = 1..jj]):
  incrementList := [1.0 $ ii = 1..jj]:
  for ii from 1 to jj do 
    if gridList[ii] < 2 then
      incrementList[ii] := (endPt[ii] - startPt[ii]) / 15.0: //arbitrary
    else
      incrementList[ii] := float((endPt[ii] - startPt[ii]) / (gridList[ii])):
    end_if:
  end_for:
  currPt := startPt:
  while bDone = FALSE do
    if not ((traperror( (currVal := float(theFunc(op(currPt)))) ) <> 0) 
            or ( (map({op(currVal)},domtype)<>{DOM_FLOAT}) and (map({op(currVal)},testtype,Type::ListOf(DOM_FLOAT))<>{TRUE}))) then
      for kk from 1 to nops(extremaIndices) do
        if traperror( (compareVal := op(currVal,extremaIndices[kk]) )) <> 0 then
          error( "Invalid extrema specification in computeGridPoints" );
        end_if:
        if minArray[kk] = [] then 
          minArray[kk] := compareVal:
          maxArray[kk] := compareVal:
        elif compareVal < minArray[kk] then
          minArray[kk] := compareVal:
        elif compareVal > maxArray[kk] then
          maxArray[kk] := compareVal:
        end_if:
      end_for:
    end_if:
    for ii from jj downto 1 do
      currPt[ii] := currPt[ii] + incrementList[ii]:
      if currPt[ii] > endPt[ii] then //we've completed a pass at level ii
        if ii = 1 then
          bDone := TRUE:
        else
          currPt[ii] := startPt[ii]:
        end_if:        
      else   //found one!
        break:
      end_if:
    end_for:
  end_while:
  [minArray,maxArray]
end_proc:

//We expect:
//  (i) a List "inList" containing the plotting parameters as passed in - mostly key=value pairs;
// (ii) two Lists "currMinVals" and "currMaxVals" which are used to accumulate min and max vals for color functions;
plotlib::setupPlotItem := proc(inList,dim, MinV, MaxV)
  local i, j, kk, applyArray, rangePos, gridPos, colorPos,
        throwAway, theColor, localMinVals, localMaxVals;
  option escape;
begin
  localMinVals := []:
  localMaxVals := []:
  rangePos     := []:
  applyArray   := []:

  for j from 1 to nops(inList) do
    i := op(inList,j):

    if type(i) = "_equal" then
      case op(i,1)
        of Grid do
          gridPos := j:
        break:
        of Color do
          colorPos := j:
        break:
	of Smoothness do
	break:
        otherwise  //is this a Range statement?
          if nops(op(i,2)) = 2 then
            if (testtype(op(i,2),Type::ListOf(Type::Constant))) and (nops(op(i,2)) = 2) then
              rangePos := append(rangePos, j):
            end_if:
          end_if:
        break: 
      end_case:
    end_if:
  end_for:
  
  throwAway := 0:
  if colorPos <> 0 then
    [theColor,throwAway] := plotlib::presetColorSpec( op(inList,[colorPos,2]), dim, TRUE ):
    if throwAway = TRUE then
      applyArray := append(applyArray,plotlib::resetColor):
    end_if:
  end_if:

  if not ( MinV = [] or MaxV = [] or
           nops(MinV) <> dim or nops(MaxV) <> dim or
           has(map(MinV, testtype, Type::Real), FALSE) or
           has(map(MaxV, testtype, Type::Real), FALSE) 
	 ) then
     [localMinVals,localMaxVals] := [MinV,MaxV]
  else
     [localMinVals,localMaxVals] := plotlib::computeGridPoints(
        inList[2],
	[inList[rangePos[kk]] $ kk=1..nops(rangePos)],
	[5 $ kk=1..nops(rangePos)],
	computeExtrema=[kk $ kk=1..dim]
      )
  end_if;

  [applyArray, theColor, localMinVals,localMaxVals];
end_proc:


