
plotlib::hatch := proc(hatch, hatch_obj, hnr, LogXY, baseline) 
  local color, h, ha, l, lines, style, width;
begin

  // Default-Einstellungen:
  lines  := 100;
  width  :=   1;
  style  := SolidLines;
  color  := RGB::LawnGreen;

  // Typpruefung und Konvertierung der Argumente

  if domtype(hatch) = DOM_INT and hatch > 0 then
     ha := [[hatch, BaseLine, baseline, lines, [Flat, color], width, style]]
  elif domtype(hatch) = DOM_LIST then
     ha := plotlib::check_hatch_list(hatch, baseline, lines, color, width, style);
  else
     error("Illegal Hatch.")
  end_if;
 
  l := [];
  for h in ha do
      l := l . plotlib::calc_hatch(h, hatch_obj, LogXY);
  end_for;
  
  return( l )

end_proc:

plotlib::check_hatch_list := proc(hatch, baseline, lines, color, width, style)
  local a, b, bl, c, h, ha, l, n, n1, n2, ret, s, t, w; 
begin

  if hatch = [] then
     error("Hatch = [...]: Non-empty list expected.")
  elif domtype(hatch[1]) <> DOM_LIST then
     return(plotlib::check_hatch_list([hatch], baseline, lines, color, width, style))
  end_if;

  // hatch = [l1,...,lN], l1,...,lN list, N >= 1

  ret := [];

  for h in hatch do

      c := color; w := width; s := style; l := lines; bl := baseline; 

      if domtype(h) <> DOM_LIST then
         error("Hatch = [ arg_1, ..., arg_N ], arg_i should be lists.")
      end_if;
      if not ((domtype(h[1]) = DOM_INT and h[1] > 0) or h[1] = BaseLine) then
         error("Hatch: [n1, n2, ...], n1 an integer with n1 > 0 or n1 = BaseLine.")
      end_if;
      if not ((domtype(h[2]) = DOM_INT and h[2] > 0) or h[2] = BaseLine) then
         error("Hatch: [n1, n2, ...], n2 an integer with n2 > 0 or n1 = BaseLine.")
      end_if;
      if h[1] = h[2] then
         error("Hatch: [n1, n2, ...], n1 and n2 should be different.")
      end_if;
      n1 := h[1]; n2 := h[2]; 
      // Loesche ersten beide Elemente in der Liste h
      h[1] := null(); h[1] := null();
      if n1 = BaseLine then n := n1; n1 := n2; n2 := n end_if;

      t := select(h, has, Color);
      if t = [] then
         c := [Flat, color]
      else
         if type(t[1]) <> "_equal" then
            error("Hatch: Illegal Color.")
         end_if;
         c := op(t[1],2);
	 h := subs(h, op(t)=null());
         if testtype(c, Type::ListOf(Type::Real, 3)) then
	    c := [Flat, c]
         else
            error("Hatch: Illegal Color.")
	 end_if
      end_if;

      t := select(h, has, LineStyle);
      if t <> [] then
         if type(t[1]) <> "_equal" then
            error("Hatch: Illegal LineStyle.")
         end_if;
         s := op(t[1],2);
	 h := subs(h, op(t)=null());
	 if s <> SolidLines and s <> DashedLines then
	    error("Hatch: Illegal LineStyle.")
         end_if;
      end_if;

      t := select(h, has, LineWidth);
      if t <> [] then
         if type(t[1]) <> "_equal" then
            error("Hatch: Illegal LineWidth.")
         end_if;
         w := op(t[1],2);
	 h := subs(h, op(t)=null());
	 if domtype(w) <> DOM_INT or w < 1 then
	    error("Hatch: Illegal LineWidth.")
         end_if;
      end_if;

      t := select(h, has, Lines);
      if t <> [] then
         if type(t[1]) <> "_equal" then
            error("Hatch: Illegal Lines.")
         end_if;
         l := op(t[1],2);
	 h := subs(h, op(t)=null());
	 if domtype(l) <> DOM_INT or l < 1 then
	    error("Hatch: Illegal Lines.")
         end_if;
      end_if;

      t := select(h, has, BaseLine);
      if t <> [] then
         if type(t[1]) <> "_equal" then
            error("Hatch: Illegal BaseLine.")
         end_if;
         bl := float(op(t[1],2));
	 h := subs(h, op(t)=null());
	 if domtype(bl) <> DOM_FLOAT then
	    error("Hatch: Illegal BaseLine.")
         end_if;
      end_if;

      ha := [n1, n2, bl, l, c, w, s];

      case nops(h)
         of 0 do
	    break;
	 of 1 do
	    l := h[1];
	    if type(l) = "_equal" then
	       l := op(l, 2);
	    end_if;
	    if type(l) = "_range" then
	          a := float(op(l,1)); b := float(op(l, 2));
		  if domtype(a) <> DOM_FLOAT or
		     domtype(b) <> DOM_FLOAT or a >= b then
		     error("Hatch: Illegal range.")
		  else
		     ha := ha . [a, b]
		  end_if;
	     else
	       error("Hatch: Illegal range.") 
	     end_if;
             break;
	  otherwise
	     error("Hatch: Illegal arguments.")
       end_case;

       ret := ret . [ha]

   end_for;

   return( ret )

end_proc:



/*

Schraffiert die Bereiche zwischen der Funktion f(x) und g(x) 
im Intervall x=a..b
Der Rueckgabewert der Prozedur ist ein Objekt, dass in plot2d 
und plotfunc2d benutzt werden kann.

f - ein Ausdruck oder eine Prozedur in einer Veraenderlichen
a - ein Ausdruck, der sich zu einer reellen Zahl evaluiert
b - ein Ausdruck, der sich zu einer reellen Zahl evaluiert

Es muss gelten: b-a > 0

Optional:
g - ein Ausdruck oder eine Prozedur in einer Veraenderlichen.
    Default ist die konstante Funktion f(x) = 0
n - eine ganze Zahl > 2. Gibt die Anzahl der Schraffuren an.
    Default ist 50.
c - ein RGB-Farbe. Siehe ?RGB bzgl. vordefinierter Farbwerte.
w - eine ganze Zahl > 0. Strichstaerke der Schraffuren.
    Default ist 5.
    s - SolidLines oder DashedLines. Schraffuren werden als 
    durchgezogene oder gestrichelelte Linien gezeichnet.
    Default ist SolidLines.
d - eine ganze Zahl >= 0. Ist d > 0, dann wird die erste
    Linie der Schraffur nicht an der Stelle a, sondern 
    an der Stelle a+s/(d+1), wobei s=(b-a)/n ist. Der 
    Abstand der Schraffur-Linien ist in jedem Fall s.
    d gibt also nur einen Offset bzgl a an, an dem mit
    den Schraffur-Linien begonnen wird. Diese Option
    ist dann sinnvoll, wenn mehrere Schraffuren im 
    Intervall a..b gezeichnet werden sollen, und die
    Schraffuren nicht bereinander liegen sollen.

Soll z.B. c spezifiziert werden, so muss auch g und n angegeben
werden, z.B. schraffur_plot2d(sin(x), -1, 1, 0.25, 50, RGB::Green)


Beispiele:

f:=sin(x): a:=-1: b:=1: y:=0.25: n:=50:
plotfunc2d(Axes=Corner, AxesInFront=TRUE,
           schraffur_plot2d(f,a,b,y,n,RGB::Green), f, y,
           x=a..b):
	   
f:=sin(x): a:=-1: b:=1: y:=0.25: n:=50:
plot2d(Axes=Corner, AxesInFront=TRUE,
       schraffur_plot2d(f,a,b,y,n,RGB::Green),
       [Mode=Curve, [x, f], x=[a,b]],
       [Mode=Curve, [x, y], x=[a,b]]):

f:=sin(x): a:=-1: b:=1: y:=0.25: n:=50:
plot2d(Axes=Corner, AxesInFront=TRUE,
       schraffur_plot2d(f,a,b,y,n,RGB::Green),
       [Mode=Curve, [x, f], x=[a,b], LineWidth=12],
       [Mode=Curve, [x, y], x=[a,b], LineWidth=12]):

f:=sin(x): a:=-5: b:=5: y:=x: n:=50:
plotfunc2d(Axes=Corner,AxesInFront=TRUE,
           schraffur_plot2d(f,a,b,y,n,RGB::Green),f,y,
           x=a..b):

f:=sin(x): a:=-0.5: b:=1.25: y:=x^2: n:=50:
plotfunc2d(Axes=Corner,AxesInFront=TRUE,
           schraffur_plot2d(f,a,b,y,n,RGB::Green),f,y,
           x=a..b):

f1:=sin(x): f2:=cos(x): a:=-1: b:=2: y:=0.25: n1:=49: n2:= 50:
plotfunc2d(Axes=Corner, AxesInFront=TRUE,
           schraffur_plot2d(f1,a,b,y,n1,RGB::Purple), f1, 
           schraffur_plot2d(f2,a,b,y,n2,RGB::Gold), f2,
           y,
           x=a..b):

f1:=sin(x): f2:=cos(x): a:=-1: b:=2: y:=0.25: n:=50:
plotfunc2d(Axes=Corner, AxesInFront=TRUE,
           schraffur_plot2d(f1,a,b,y,n,RGB::Purple , 5, SolidLines , 0), f1, 
           schraffur_plot2d(f2,a,b,y,n,RGB::Magenta, 5, DashedLines, 1), f2,
           y,
           x=a..b):

// Ist f unstetig, so sollte der vertikale Sichtbarkeitsbereich 
// vorgegeben werden.

// Ohne Vorgabe eines vertikalen Sichtbarkeitsbereichs:
f:=tan(x): a:=-2: b:=2: y:=0: n:=50:
plotfunc2d(Axes=Corner, AxesInFront=TRUE,
           schraffur_plot2d(f,a,b,y,n,RGB::Green), f, y,
           x=a..b):

// Vorgabe eines vertikalen Sichtbarkeitsbereichs (yr=-10..10):
f:=tan(x): a:=-2: b:=2: y:=0: n:=50:
plotfunc2d(Axes=Corner, AxesInFront=TRUE,
           schraffur_plot2d(f,a,b,y,n,RGB::Green), f, y,
           x=a..b, yr=-10..10):

*/

plotlib::calc_hatch := proc(h, ht, LogXY)
  local a, b, baseline, c, color, d, fx1, fy1, 
        l, lines, mina, maxb, s, style, width;
begin
  
  // h = [ n1, n2, baseline, lines, color, width, style <, a, b> ]

  if domtype(ht[h[1]]) <> DOM_LIST then
     return( [] )
  end_if;

  // ht[h[i]][1] = [
  //                [Mode=Curve, [x(u),y(u)], u=[a1,b1],...], 
  //                ...
  //                [Mode=Curve, [x(u),y(u)], u=[aN,bN],...]
  //               ]
  // a <= a1 < b1 < a2 < b2 < ... < aN < bN <= b
  //
  // ht[h[i]][2] = [
  //                [Mode=Curve, [x(u),y(u)], u=[a,b],...]
  //               ]

  baseline := h[3]; lines := h[4]; color := h[5]; 
  width := h[6]; style := h[7];

  mina := float(op(ht[h[1]][2][3], [2,1]));
  maxb := float(op(ht[h[1]][2][3], [2,2]));

  if nops(h) = 9 then
     a := h[8]; b := h[9];
     if a < mina then a := mina end_if;
     if b > maxb then b := maxb end_if;
  else
     a := mina; b := maxb;
  end_if;

  // mina <= a < b <= maxb
  
  fx1 := ht[h[1]][2][2][1];
  fy1 := ht[h[1]][2][2][2];

  fx1 := plotlib::convert2exec(fx1, LogXY[1]);
  fy1 := plotlib::convert2exec(fy1, LogXY[2]);

  s := float((b-a)/lines);
  c := ht[h[1]][1];    
  c := plotlib::combine_intervals([float(op(c[i][3],2))$i=1..nops(c)],
                                  [[a,b]],mina,maxb,[]);
  // c = [[c1,d1],...,[cM, dM]]

  if h[2] = BaseLine then   
     l := plotlib::hatching(c, a, b, s, fx1, baseline, fx1, fy1);
     return( [[Mode = GridLines, l , 
               Color=color, LineWidth=width, LineStyle=style]] )
  else
     if domtype(ht[h[2]]) <> DOM_LIST then
        return( [] )
     end_if;
     fx2 := ht[h[2]][2][2][1];
     fy2 := ht[h[2]][2][2][2];
     fx2 := plotlib::convert2exec(fx2, LogXY[1]);
     fy2 := plotlib::convert2exec(fy2, LogXY[2]);

     d := ht[h[2]][1];
     mina := max(float(op(ht[h[2]][2][3], [2,1])), mina);
     maxb := min(float(op(ht[h[2]][2][3], [2,2])), maxb);
     d := plotlib::combine_intervals([float(op(d[i][3],2))$i=1..nops(c)],
                                     c, mina, maxb, []);
     l := plotlib::hatching(c, a, b, s, fx1, fy1, fx2, fy2);
     return( [[Mode = GridLines, l , 
               Color=color, LineWidth=width, LineStyle=style]] )
  end_if;
end_proc:

plotlib::hatching := proc(c, mina, maxb, s, fx1, fy1, fx2, fy2)
  local ai, bi, m, n, l, x;
begin
  x := mina; n := 1; m := nops(c); l := [];
  repeat
  ai := c[n][1]; bi := c[n][2];
  while x < ai do x := x + s end_while;
    while x <= bi do  
      l := l . [polygon(point(fx1(x), fy1(x)), point(fx2(x), fy2(x)))];
      x := x + s;
    end_while;
    n := n + 1
  until x > maxb or n > m end_repeat;
  return( l )
end_proc:

plotlib::hatching := proc(c, mina, maxb, s, fx1, fy1, fx2, fy2)
  local n, l, r, x;
begin
  l := [];
  // Berechne die urspruengliche Schrittweite s = (b-a)/lines 
  // fuer jedes Interval [n[1], n[2]] neu. Es soll fr r gelten, 
  // dass n[1]+k*r=n[2] ist. Da interne Rundungsfehler in der
  // for-Schleife dazu fuehren koennen, dass n[1]+k*r > n[2]
  // ist, wird n[2] um eps vergroessert. Hier ist allerdings
  // Vorsicht geboten, denn x=n[2]+eps kann ggf. bereits einen
  // Fehler ausloesen, da fx(x) bzw. fy(x) singulaere Punkte
  // seien koennen.
  for n in c do
      r := float((n[2]-n[1])/ceil((n[2]-n[1])/s));
      for x from n[1] to n[2]+10^(-DIGITS-1) step r do
          l := l . [polygon(point(fx1(x), fy1(x)), point(fx2(x), fy2(x)))];
    end_for;
  end_for;
  return( l )
end_proc:

plotlib::convert2exec := proc(f, log_scaling)
  local f1, ft, t;
begin
  t := op(indets(f)); f1 := f;
  // Convert f to an executable object
  if not contains({DOM_PROC, DOM_EXEC, DOM_FUNC_ENV}, domtype(f)) then 
     if contains({DOM_EXPR,DOM_IDENT,DOM_INT,DOM_RAT,DOM_FLOAT,piecewise}, domtype(f)) then
	if log_scaling then
           f1 := fp::unapply(ln(f), t);
	else
	   f1 := fp::unapply(f, t);
	end_if
      else
        error("Illegal argument: ".expr2text(f))
      end_if
  else
      if log_scaling then
        f2 := f; f1 := fp::unapply(ln(hold(f2)(t)), t);
     end_if
  end_if;
  return(f1)
end_proc:
