// $Date: 2001/09/03 16:21:11 $ $Author: stefanw $ $Revision: 1.2 $

// prog::getOptions -- extract options from argument sequence
// 
// prog::getOptions(ARGS, <allOptions>, <index>)
// 
// ARGS          - expression sequence of arguments
//                 Options can be identifiers
//                 or expression 'option = value'
// allOptions    - list or set of all options
// index         - integer position to skip the 'index'-th arguments of 'ARGS'
//              
// Return value: a table of all options and corresponding values;
//               if an option is a single option without a value,
//               the table contains 'NIL' as the corresponding value
//
//
// Example:
// prog::getOptions((1, Y, All, Status = "test"), [hold(All), hold(Status), hold(None)], 3)
//
//              table(All    = NIL,
//                    Status = "test")
//
// The using procedure can ask for options with 'contains'
//   OPTS:= getOptions(args(), ...)
// if contains(OPTS, hold(Status)) then
//   if OPTS[hold(Status)] = "test" then
//     ...
// 

prog::getOptions:=
  proc(ARGS = [], allOptions = [], index = 1)
    option hold, noDebug;
    local OPTS;
  begin
    ARGS:= context(ARGS);
    index:= context(index);
    if type(ARGS) <> DOM_LIST and type(ARGS) <> "_exprseq" then
      // expression "sequence" of length 1
      ARGS:= [ARGS]
    end_if;
    if nops(ARGS) = 0 or nops(ARGS) < index then
      return(table())
    else
      allOptions:= context(allOptions);
      OPTS:= select([op(ARGS, index..nops(ARGS))], has, allOptions);
      OPTS:= map(OPTS, X -> if type(X) = DOM_SET then op(X) else X end_if);
      OPTS:= split(OPTS, testtype, Type::Equation());
      table(op(op(OPTS, 1)), op(map(op(OPTS, 2), X -> X = NIL)));
    end_if
  end_proc: