// $Date: 2001/07/25 14:11:14 $ $Author: bij $ $Revision: 1.9 $

/*--

TCOV.README
-----------
Das Datenfile hat die Form:

<int>:<int>:
  ~     ~
  ~     ~
<int>:<int>:
-1:-1:
<str>
  ~
<str>

<int> ganze Zahl >= 0
<str> Zeichenkette in " " eingeschlossen.

Der erste Eintrag ist der Index des zugehoerigen Dateinamens, der
zweite Eintrag die Zeilennummer. Nach der Zeile -1:-1: folgen die
Dateinamen. 

Bsp.:

0:2:
1:4:
-1:-1:
"test1":
"test2":

Zeile 2 der Datei test1, die den Index 0 hat, und Zeile 4 der
Datei test2, die den Index 1 hat, wurden durchlaufen.

--*/

// 

// Option: Pretty, File, Stat

prog::tcov :=
proc(STMT, Opt = hold(Pretty), Opt1 = null())
  local fd, getline1, getline2, lpr, i, j, l, file,
        line, cnt, filename, lines, max_file, result, 
        fname, format, chName, IFILE, OFILE;
  save PRETTYPRINT;
  option hold;
begin
  //if debug() <> TRUE then
  //  error("MuPAD kernel must be started in debug modus (see '?debug')")
  //end_if;

  if testargs() then
    case args(0)
      of 1 do
        // one statement
        break
      of 2 do // Option: Pretty, File, Stat
        if contains([hold(Pretty), hold(File), hold(Stat)], Opt) = 0 then
          if Opt = hold(Append) then
            Opt1:= hold(Append)
          else
            error("2nd argument must be option 'File', 'Pretty', 'Stat' or 'Append'")
          end_if
        end_if;
        break
      of 3 do
        if Opt1 <> hold(Append) then
          error("3rd argument must be option 'Append'")
        end_if;
        break
      otherwise
        error("wrong number of arguments")
    end_case
  end_if;

  getline1 := proc()
      local a, b;
  begin
        if finput(fd, a, b) = null() then
           fclose(fd); error("Unexpected EOF while reading '".fname."'")
        elif a = -1 then
           null()
        else
           a, b
        end_if
  end_proc:

  getline2 := proc()
      local a;
  begin
        if finput(fd, a) = null() then
           fclose(fd); null()
        else
           a
        end_if
  end_proc:

  // Erzeugt Zeichenkette der L"ange width.               
  // Ist length(string) < width, so wird der rechte Teil  
  // der Zeichenkette mit fill_pattern aufgef"ullt, falls 
  // falls right = TRUE, sonst der linke Teil.            

  format := proc(str, width, fill_pattern, right)
     local i;
  begin
     str := "".str;
     for i from 1 to width - length(str) do
         if right then
            str := str.fill_pattern
         else
            str := fill_pattern.str
         end_if
     end_for;
     str
  end_proc:

  lpr := proc(str)
    begin format(str, 6, " ", FALSE) end_proc:

  chName:= proc(X, ch)
             local Y;
           begin
             if (Y:= stringlib::subs(X, ch)) <> X then
               stringlib::subs(X, ch)
             else
               FAIL
             end_if
           end_proc;

  //------------------ Hauptprogramm -------------------

  if args(0) = 0 then
     return();
  end_if;

  fname := "/tmp/tcov".getpid().".dat";

  // Aufruf des internen generischen test coverage (gcov). 
  // Dieser schreibt die in der Datei TCOV.README          
  // beschriebenen Daten in die Datei tcov.dat, die        
  // nun im folgenden aufbereitet werden.                  

  result := gcov(context(STMT), fname);

  fprint(Unquoted, 0, "Info: Evaluation of '".
                      _if(length(expr2text(STMT)) > 35,
                          stringlib::format(expr2text(STMT), 35)."...",
                          stringlib::format(expr2text(STMT))).
                      "' finished.");
  fprint(Unquoted, 0,  "Info: Now evaluate coverage information.");

  if (fd := fopen(fname)) = FAIL then 
     error("could not open file '".fname."'") 
  end_if;

  // Der wesentliche Algorithmus, der das Datenfile  
  // ausliest und aufbereitet.                       

  // Datenfile einlesen 

  lines:= table();
              
  while domtype((l := getline1())) <> DOM_NULL do
	 file := l[1] ; line := l[2] ;
     if domtype(lines[file]) <> DOM_LIST then
        lines[file] := []
     end_if;
     if contains(lines[file], line) > 0 then
        cnt[file][line] := cnt[file][line] + 1
     else
        cnt[file][line] := 1;
        lines[file] := append(lines[file], line)
     end_if
  end_while;

  max_file := -1;
  while domtype((l := getline2())) <> DOM_NULL do
     max_file := max_file +1;
     filename[max_file] := l;
  end_while;

  // 

  // reads in existing status files
  if Opt1 = hold(Append) then // extend existing files
    for i from 1 to max_file do
      if traperror((IFILE:= fopen(chName(filename[i], ".mu" = ".tcov")))) = 0
         and IFILE <> FAIL then
        // file exists
        while (finput(IFILE, l, j)) <> null() do
          if contains(lines, i) then
            if contains(cnt, i) and contains(cnt[i], l) then
              cnt[i][l]:= cnt[i][l] + j
            else
              lines[i]:= append(lines[i], l);
              cnt[i][l]:= j
            end_if
          else
            lines[i]:= [l]
          end_if
        end_while;
        fclose(IFILE)
      end_if
    end_for
  end_if;

  if Opt = hold(Pretty) then // screen output
    PRETTYPRINT := FALSE;
    for i from 1 to max_file do
      if domtype(lines[i]) = DOM_LIST then
        print(Unquoted, "File: ".filename[i]);
        for j in sort(lines[i]) do
          print(Unquoted,
                "      Line:".lpr(j).":".lpr(cnt[i][j])." time(s)")
        end_for
      end_if
    end_for
  // 
  else // write status files
    for i from 1 to max_file do
      // open status file for writing
      if traperror((IFILE:= fopen(hold(Text),
                                  chName(filename[i], ".mu" = ".tcov"),
                                  hold(Write)))) <> 0
         or IFILE = FAIL then
        if chName(filename[i], ".mu" = ".tcov") <> FAIL then
          warning("cannot open status file '".chName(filename[i], ".mu" = ".tcov")
                  ."' for writing")
        else
          // tmp file
        end_if;
        next
      end_if;
      if domtype(lines[i]) = DOM_LIST then
        for j in sort(lines[i]) do
          fprint(IFILE, j, cnt[i][j])
        end_for
      end_if;
      fclose(IFILE)
    end_for;
    fprint(Unquoted, 0, "Info: '.tcov' status files written.");
    
    if Opt = hold(File) then // write file with commented lines
      // open source files and files for output
      for i from 1 to max_file do
        if traperror((IFILE:= fopen(filename[i]))) <> 0
           or IFILE = FAIL then
          warning("cannot open source file '".filename[i]."'");
          next
        end_if;
        if traperror((OFILE:= fopen(hold(Text),
                                    chName(filename[i], ".mu" = "_tcov.mu"),
                                    hold(Write)))) <> 0
           or OFILE = FAIL then
          if chName(filename[i], ".mu" = "_tcov.mu") <> FAIL then
            warning("cannot open file '".chName(filename[i], ".mu" = "_tcov.mu")
                    ."' for writing")
          else
            // tmp file
          end_if;
          next
        end_if;
        l:= 0;
        fprint(Unquoted, OFILE, "// !!! Three lines are inserted in front of this file !!! -*-tcov-*- ");
        fprint(Unquoted, OFILE, "// Generated by ".
                                _if(Opt1 = hold(Append), "*appending*", "")
                                ." 'prog::tcov' call: ".stringlib::format(expr2text(STMT), 35)."\n");
        while (l:= l + 1; line:= ftextinput(IFILE)) <> null() do
          if contains(cnt, i) and contains(cnt[i], l) then
            fprint(Unquoted, OFILE, "/*".
                   stringlib::format(expr2text(min(cnt[i][l], 9999)), 4, hold(Right)).
                   " */".line)
          else
            fprint(Unquoted, OFILE, "/* --- */".line)
          end_if
        end_while;
        fclose(IFILE);
        fclose(OFILE)
      end_for;
      fprint(Unquoted, 0, "Info: '_tcov.mu' files written.")
    end_if
  end_if;

  gcov(NIL, fname); // Unlink tcov file 

  print(Unquoted, "");
  result

end_proc:
