// $Date: 2002/02/11 14:25:27 $ $Author: bij $ $Revision: 1.66 $

/*++
prog::testinit -- init testing

prog::testinit(f, <expected_ntime>, <Option>)

f                - file name (default is "test")
expected_ntime   - normalized!! Time (with 'prog::ntime') in
                   milli seconds (DOM_INT) or seconds (DOM_FLOAT)
Option           - _in(Quiet, Timer = t)
                   Quiet = suppress 'prog::testexit' messages
                   Timer = use 't' instead of 'prog::ntime' to
                           compute the normalization factor
                   All   = measure time of whole test

Opens the test result file 'f.res'. If the variable 'TESTPATH' has a
value, it is used as a prefix for the path name.
++*/

alias(AllOptions = [hold(Quiet), // no messages
                    hold(Timer), // = func/value instead of 'prog::ntime()'
                    hold(All),   //count difference between 'testinit' and 'testexit'
                                 // otherwise: only time for evaluating of first
                                 // arguments inside of 'prog::test' calls is counted
                    // architectures: times
                    "UNIX", "MSDOS", "MACOS", "linux", "solaris"
                    ]):

prog::testinit:=
  proc(f = "test", TIME = 0, TIMEUNIT = 0)
    local UOPT, LogFile, inp, S, NM, OPTS;
  begin
    if testargs() then
      if args(0) = 0 then
        // ...
      elif domtype(f) <> DOM_STRING then
          error("first argument must be a string")
      end_if
    end_if;
    
    if TESTPATH <> hold(TESTPATH) then
      f:= TESTPATH.f
    end_if;
    sysassign(prog::TestFile, fopen(Text, f.".res", Write));

    OPTS:= prog::getOptions(args(), AllOptions, 2);
    if contains(OPTS, hold(Arch)) then
      OPTS:= table(op(OPTS), op(OPTS[hold(Arch)]))
    end_if;

    // table containing older results
    UOPT:= table(text2expr(Pref::userOptions()));
    // MuPAD Option: -U "TESTPATH=\"PROP/TEST/property.err\",NAME=\"galois\"
    //                                                      ,TIMING=is,NEWTIMINGS=TRUE"
    if UOPT <> table() then
      // gesetzte 'Useroption'
      if nops(UOPT) > 1 then
        if contains(UOPT, hold(TESTPATH)) then
          // Dateiname relativ zum LIBPATH
          if contains(UOPT, hold(NAME)) then
            NM:= UOPT[hold(NAME)]
          else
            NM:= "noname"
          end_if;
          S:= table("Name" = NM);
          // Name der Log-Datei
          if traperror((LogFile:= fopen(UOPT[TESTPATH]))) = 0 and LogFile <> FAIL then
            while (inp:= finput(LogFile)) <> null() do
              if op(inp, 1) = "Time" then
                // save global timing information
                // Line format:
                //    const   comp-name  expect time  yesterdays time  yesterdays difference 
                //   ["Time", "local",   14490,       14650,           320]
                //   time in (unnormalized) ms
                if op(inp, 2) = NM then // this computer
                  S["Time", op(inp, 2)]:= (op(inp, 3..nops(inp)), 0, 0) // fill up with dummies
                                                                        // for first set of empty times
                else // uncomment this to keep all information
                  //S["Time", op(inp, 2)]:= (op(inp, 3..nops(inp)))
                end_if
              else
                S[inp]:= FALSE
              end_if
            end_while;
            fclose(LogFile);
          else
            userinfo(1, "Yesterdays log-file not found");
            // set default values
            S["Time", 0]:= [0, 0];
            UOPT[hold(RESETTIMINGS)]:= TRUE; // set initial timing
          end_if;
        end_if;
        // perform timings of one function
        if contains(UOPT, hold(TIMING)) then
          // perform timings
          f:= genident("_INT");
          evalassign(f, eval(UOPT[hold(TIMING)]), 2);
          protected(f, hold(Error));
          inp:= UOPT[hold(TIMING)];
          prog::init(eval(UOPT[hold(TIMING)]));
          sysevalassign(inp,
                        subsop(proc()
                                 local RES;
                               begin
                                 sysassign(prog::TestTimings,
                                           prog::TestTimings + time((RES:= NIL(args()))));
                                 RES
                               end_proc, [4, 1, 2, 2, 1, 2, 0] = eval(f)), 2,    // f
                                         6 = text2expr(prog::getname(eval(f)))); // name of f
          sysassign(prog::TestTimingsFunc, UOPT[hold(TIMING)]);
          sysassign(prog::TestTimings, 0)
        end_if;
        // reset old timings inside of logfiles
        if contains(UOPT, hold(RESETTIMINGS)) then
          S[hold(RESETTIMINGS)]:= TRUE
        end_if
      end_if;
      // write older results
      if contains(S, ("Time", NM)) then
        S["Time", 0]:= S["Time", NM]; // expected Timings from logfile: unnormalized!
        delete S["Time", NM]
      else
        S["Time", 0]:= 0
      end_if;
      sysassign(prog::TestOldErrors, S);
      sysassign(prog::TestErrors, table());
      sysassign(prog::TestLogFile, UOPT[hold(TESTPATH)])
    end_if;
    
    // evaluate time unit -> inside of 'prog::testexit'
    // expected time for whole test
    if domtype(TIME) = DOM_INT then // integer number as second argument
      sysassign(prog::TestExpectedTime, TIME)  // expected cumputing time in milli seconds
                                               // normalized with prog::normalizeTime
    elif domtype(TIME) = DOM_FLOAT then // real number as second argument
      // expected cumputing time in seconds
      sysassign(prog::TestExpectedTime, TIME)
    elif contains(OPTS, sysname(hold(Arch))) then
      sysassign(prog::TestExpectedTime, OPTS[sysname(hold(Arch))])
    elif contains(OPTS, sysname()) then
      sysassign(prog::TestExpectedTime, OPTS[sysname()])
    end_if;
    if domtype(prog::TestExpectedTime) = DOM_FLOAT then
      // keep time in seconds
      sysassign(prog::TestExpectedTime, specfunc::trunc(prog::TestExpectedTime*1000))
    end_if;
    
    if prog::TestFile = FAIL then
      error("could not open results file")
    end_if;
    sysassign(prog::TestNo, 1);
    sysassign(prog::TestFunc, FAIL);
    sysassign(prog::TestInitTime, time()); // gesamte Zeit
    sysassign(prog::TestTime, 0); // Rechenzeit
    sysassign(prog::TestErrorCount, 0);
    sysassign(prog::TestCount, 0);
    if contains(OPTS, hold(Timer)) then
      if domtype(OPTS[hold(Timer)]) = DOM_FLOAT then
        sysassign(prog::TestNTime, specfunc::trunc(OPTS[hold(Timer)]*1000))
      else
        sysassign(prog::TestNTime, OPTS[hold(Timer)])
      end_if
    end_if;
    sysassign(prog::TestQuiet, contains(OPTS, hold(Quiet)));
    sysassign(prog::TestNTime, [prog::TestNTime, contains(OPTS, hold(All))]);
    null()
  end_proc:

/*++
prog::testexit -- exit testing

prog::testexit()

Close test results file.
++*/

prog::TestFile:= FAIL:
prog::TestLogFile:= FAIL:
prog::TestOldErrors:= FAIL:
prog::TestErrors:= FAIL:
prog::TestErrorCount:= 0:
prog::TestExpectedTime:= 0:  // expected time factor
prog::TestNTime:= 1:         // list: [normalize factor, measure whole test time : DOM_BOOL]
prog::TestQuiet:= FALSE:

prog::testexit:=
  proc()
    local LogFile, TIME, ntime, ntimearg, ntimelog, k, NAME, F;
  begin
    TIME:= time() - prog::TestInitTime; // gesamte Test-Zeit
    F:= X -> stringlib::formatf(float(X), 1);
    if domtype(prog::TestFile) = DOM_INT then
      fclose(prog::TestFile)
    end_if;

    // normalized time factor
    if op(prog::TestNTime, 1) = 1 then // compute only when not given
      ntime:= prog::ntime()
    else
      ntime:= op(prog::TestNTime, 1)/1000
    end_if;
    // take whole test time instead of netto time using by all 'prog::test' calls
    // if option 'All' is given
    if op(prog::TestNTime, 2) = TRUE then
      sysassign(prog::TestTime, TIME) // whole test time
    end_if;
    ntimearg:= prog::TestExpectedTime; // expected time given to 'prog::testinit'

    // interaktiv: Meldung ausgeben
    if not prog::TestQuiet /*and prog::TestLogFile = FAIL*/ then
      //Info: property::assume - 26 examples,  2 errors, 12 warnings, time: 18.10 s (expected: 25.00 s)
      fprint(Unquoted, 0, "Info: ",
             expr2text(prog::TestCount),
             " test"._if(prog::TestCount = 1, "", "s").", ",
             expr2text(prog::TestErrorCount),
             " error"._if(prog::TestErrorCount = 1, "", "s")
//             );
//      fprint(Unquoted, 0, "Info: "
             ,", runtime factor ", F(prog::TestTime/1000/ntime),
             _if(ntimearg <> 0,
                 ("  (expected ", F(ntimearg/1000), ")"),
                 "  (nothing expected)")
             //," (testing and loading overhead", F((TIME - prog::TestTime)/1000/ntime), ")"
             );
      // old or expected timings exists and are measurable
      if prog::TestTime >= 350 and ntimearg <> 0 then 
        if prog::TestTime/ntimearg/ntime >= 1.55 then
          warning("expected time and test time differ "
                  .F(float(((prog::TestTime/ntimearg/ntime) - 1)*100))." %")
        elif prog::TestTime/ntimearg/ntime < 0.6 then
          warning("expected time and test time differ "
                  .F(float(((prog::TestTime/ntimearg/ntime) - 1)*100))." %")
        end_if
      end_if;

      if prog::TestTimings <> FAIL then
        fprint(Unquoted, 0, "Info: '",
               expr2text(prog::TestTimingsFunc),
               "' takes ", F(prog::TestTimings/1000), " s")
      end_if;
    end_if;
    // Aufruf von Test: Log-Datei schreiben
    // Achtung: Meldungen nicht ndern! werden von Test eingelesen
    if prog::TestLogFile <> FAIL then
      ntimelog:= prog::TestOldErrors["Time", 0]; // expected time stored in '.err' file
      // format:
      //  expect time  yesterdays time  yesterdays difference 
      //  14490,       14650,           320
      //   time in (unnormalized) ms

      // Meldung wird von Test ausgewertet
      fprint(Unquoted, 0, "Info: All tests ", expr2text(prog::TestCount), " [prog::testexit]");
      NAME:= prog::TestOldErrors["Name"]; // Rechnername bergeben von Test
      if traperror((LogFile:= fopen(Text, prog::TestLogFile, Write))) = 0
         and LogFile <> FAIL then
        for k in prog::TestErrors do // alle aktuellen Fehler in Logdatei merken
          if op(k, 2) = TRUE then
            fprint(LogFile, op(k, 1))
          end_if
        end_for;
        for k in prog::TestOldErrors do // old state information
          if op(op(k, 1), 1) = "Time" and op(k, [1, 2]) <> 0 then
            fprint(LogFile, [op(op(k, 1)), op(k, 2)])
          end_if
        end_for;
        // neu setzen einer Erwartungs-Zeit
        if prog::TestOldErrors[hold(RESETTIMINGS)] = TRUE then
          fprint(LogFile, ["Time", NAME, prog::TestTime, prog::TestTime, op(ntimelog, 2) - op(ntimelog, 1)])
        else // keep old timing
          //                             old timing       akt timing      yesterdays diff
          fprint(LogFile, ["Time", NAME, op(ntimelog, 1), prog::TestTime, op(ntimelog, 2) - op(ntimelog, 1)])
        end_if;
        fclose(LogFile)
      else
        warning("could not open log file '".prog::TestLogFile."'")
      end_if;
    // Time check
      if prog::TestExpectedTime <> 0 then
        ntime:= prog::TestExpectedTime*ntime;
      else
        ntime:= op(ntimelog, 1); // old timing from '.err'-Datei
      end_if;
      if ntime >= 350 then // old or expected timings exists and are measurable
        F:= proc(x) // x in [-1.0...-0.4[ union ]0.5...infinity[
              local x2;
            begin
              x2:= 10^trunc(log(10, abs(x)));
              op(["-", "", "+"], sign(x) + 2).expr2text(round(abs(x)/x2)*x2)
            end;
        if prog::TestTime/ntime > 1.5 or prog::TestTime/ntime < 0.6 then
          warning("Time ".F(((prog::TestTime/ntime) - 1)*100)._if(prog::TestExpectedTime = 0, " X", " Y"))
        end_if
      else
        fprint(0, "Info: time not measurable [prog::testexit]")
      end_if
    end_if;
    // reset global values for next test
    sysassign(prog::TestFile, FAIL);
    sysassign(prog::TestLogFile, FAIL);
    sysassign(prog::TestNo, 1);
    sysassign(prog::TestOldErrors, FAIL);
    sysassign(prog::TestFunc, FAIL);
    sysassign(prog::TestErrorCount, 0);
    sysassign(prog::TestCount, 0);
    sysassign(prog::TestExpectedTime, 0);
    sysassign(prog::TestNTime, 1);
    null()
  end_proc:

/*++
prog::testfunc -- init testing of function

prog::testfunc(f)

f - expression

prog::testfunc(f) initializes the testing of function f. The argument
is not evaluated. If the test result file has not been opened it is
opened by prog::testfunc. If the variable 'TESTPATH' has a
value, it is used as a prefix for the path name.
++*/

prog::testfunc:=
  proc(FUNC)
    option hold;
  begin
    if args(0) <> 1 then
      error("wrong no of args")
    end_if;
    if prog::TestFile = FAIL then
      prog::testinit(expr2text(FUNC))
    end_if;
    sysassign(prog::TestNo, 1);
    sysassign(prog::TestFunc, FUNC);
    sysassign(prog::TestFuncName, expr2text(FUNC));
    null()
  end_proc:

prog::TestNo:= 1:

// testnum() returns the current test number
// testnum(n) sets the current test number to n
prog::testnum:=
  proc(NR)
  begin
    if args(0) = 0 then
      return(prog::TestNo)
    elif args(0) > 1 then
      error("wrong no of args")
    elif domtype(NR) = DOM_INT then
      if NR < 1 then
        error("number must be positive")
      end_if
    else
      error("argument must be a positive integer")
    end_if;
    sysassign(prog::TestNo, NR);
    null()
  end_proc:

/*++
prog::test -- test function

prog::test()
prog::test(a)
prog::test(a, b)
prog::test(a, b, c)

a, b - expressions
c    - string

prog::test prints - depending upon the parameters - the current function
and test number onto the test results file 'test.res'. This signals
an error in the current test.

prog::test() prints the current function and test number unconditional.

prog::test(a) prints the current function and test number if a is not
equal to TRUE or if an error occures during the evaluation of a.

prog::test(a, b) prints the current function and test number if a is not
equal to b (in the sense of _equal) of if an error occures during the
evaluation of a or b.

prog::test(a,b,c) works like prog::test(a, b) but append string c to 
the output.
++*/

prog::test:=
  proc(a, b = TRUE, c = "")
    option hold;
    local r, res, ERR, ERRes, i, lprint, TE, PREF, TIME, OPTS;
  begin
    // evaluate the third argument
    c := context(c):
    lprint:=
      proc(COND, MSG)
        name prog::test;
        local MESS;
      begin
        if COND then // error!
          // remove all newlines in input or result
          MESS:= stringlib::subs(stringlib::subs(_concat(map(args(i)$i=2..args(0),
                                                             X -> _if(domtype(X) = DOM_STRING,
                                                                      X,
                                                                      expr2text(X)))),
                                                 "\n" = " "), "  " = " ");
          if domtype(prog::TestFile) = DOM_INT then
            fprint(Unquoted, prog::TestFile, prog::TestFunc, " ",
                   prog::TestNo, ": ", MSG)
          end_if;

          // change status to '*new*' if error is new but not registered
          if prog::TestOldErrors <> FAIL and
               not contains(prog::TestOldErrors, [prog::TestFunc, expr2text(MSG)]) and OPTS[hold(Status)] = "*tst*" then
            OPTS[hold(Status)]:= "*new*"
          elif prog::TestLogFile = FAIL then  // if FAIL then no flags!
            MESS:= MESS
                   .(op(["", " // Status: Registered", " // Status: Analyzed", " // Status: ToDo"],
                        1 + contains(["*reg*", "*ana*", "*tod*"], OPTS[hold(Status)])));
            OPTS[hold(Status)]:= ""
          end_if;

          fprint(Unquoted, 0, "Error", OPTS[hold(Status)], " in test '", prog::TestFunc, "' ",
                 prog::TestNo, ": ", MESS);
          if prog::TestErrors <> FAIL then
            sysassign(prog::TestErrors[[prog::TestFunc, expr2text(MSG)]], TRUE)
          end_if;
          sysassign(prog::TestErrorCount, prog::TestErrorCount + 1)
        else // no error
          if prog::TestOldErrors <> FAIL then
            if contains(prog::TestOldErrors, [prog::TestFunc, expr2text(MSG)]) then
              fprint(0, "Info: Test '".expr2text(prog::TestNo)."' corrected [prog::test]")
            end_if;
            if contains(["*reg*", "*ana*", "*tod*"], OPTS[hold(Status)]) > 0 then
              fprint(Unquoted, 0, "Error".OPTS[hold(Status)]
                     .": test '", prog::TestNo, "' error fixed; remove status flag [prog::test]")
            end_if
          else // interaktiv
            if contains(["*reg*", "*ana*", "*tod*"], OPTS[hold(Status)]) > 0 then
              warning("test '".expr2text(prog::TestNo)."' error fixed; remove status flag")
            end_if
          end_if
        end_if
      end_proc:

    // patch warning to suppress several warnings
//     WARN:= warning;
//     sysassign(warning,
//               proc()
//               begin
//                 if not hastype(map(prog::NOWARN,
//                                    (X, Y) -> stringlib::match(Y, "[".
//                                                               _if(type(X) = DOM_STRING, X, expr2text(X)).
//                                                               "]"), args()),
//                                DOM_INT) then
//                   WARN(args())
//                 end_if
//               end_proc);

    if domtype(prog::TestFile) <> DOM_INT then
      //warning("results file has not been opened")
    end_if;
    if prog::TestFunc = FAIL then
      //warning("no function to test");
      sysassign(prog::TestFunc, "interactive");
      sysassign(prog::TestNo, 1)
    end_if;
    // Tests global zahlen
    sysassign(prog::TestCount, prog::TestCount + 1);
    
    PREF:= Pref::typeCheck(Always), testargs(TRUE);

    // collect additional arguments
    OPTS:= (hold(Status) = "tst", hold(Test));
    if args(0) > 2 then
      if domtype(args(3)) <> DOM_STRING then
        if eval(args(3)) <> hold(Quiet) and
           (type(args(3)) <> "_equal" or eval(op(args(3), 1)) <> hold(Status)) then
          error("illegal 3rd argument in test '".expr2text(prog::TestNo)."'")
        end_if;
        OPTS:= args(3..args(0));
        c:= ""
      elif args(0) > 3 then
        if eval(args(4)) <> hold(Quiet) and
           (type(args(4)) <> "_equal" or eval(op(args(4), 1)) <> hold(Status)) then
          error("illegal 4th argument in test '".expr2text(prog::TestNo)."'")
        end_if;
        OPTS:= args(4..args(0))
      end_if;
      OPTS:= split(map([OPTS], eval), testtype, Type::Equation());
      OPTS:= table(op(op(OPTS, 1)), op(map(op(OPTS, 2), X -> X = NIL)));
      if contains(OPTS, hold(Status)) then
        if domtype(OPTS[hold(Status)]) <> DOM_STRING then
          OPTS[hold(Status)]:= expr2text(OPTS[hold(Status)])
        end_if;
        OPTS[hold(Status)]:= "*".stringlib::lower(stringlib::format(OPTS[hold(Status)], 3))."*"
      elif contains(OPTS, hold(Quiet)) then // old option
        OPTS[hold(Status)]:= "*reg*"
      else
        OPTS:= table(hold(Status) = "*tst*")
      end_if;
    else
      OPTS:= table(hold(Status) = "*tst*")
    end_if;
    // is the status string valid? - //no: message only if log file exists//
    if /*prog::TestOldErrors <> FAIL and*/ contains(["*tst*", "*ana*", "*reg*", "*tod*"], OPTS[hold(Status)]) = 0 then
      fprint(Unquoted, 0, "Error: unknown status of test '".prog::TestNo."': ".expr2text(a = b));
      OPTS[hold(Status)]:= "*tst*" // unknown status
    end_if;
    
    
//     if type(args(args(0) - ARGM)) = hold(NoWarning) then
//       sysassign(prog::NOWARN, [prog::TestFuncName]);
//       ARGM:= ARGM + 1
//     elif type(args(args(0) - ARGM)) = "_equal"
//          and op(args(args(0) - ARGM), 1) = hold(NoWarning) then
//       sysassign(prog::NOWARN, op(args(args(0) - ARGM), 2));
//       if not type(prog::NOWARN) in {DOM_SET, DOM_LIST} then
//         sysassign(prog::NOWARN, [prog::NOWARN])
//       end_if;
//       ARGM:= ARGM + 1
//     end_if;
    
    if type(b) = "_equal" and op(b, 1) = hold(TrapError) then
      TE:= op(b, 2);
    else
      TE:= 0
    end_if;
    
    TIME:= 0;
    r:= FAIL; res:= FAIL;
    userinfo(Text, 1, "testing '", prog::TestFunc, "' ", prog::TestNo);
    if args(0) = 0 then
      print(prog::TestNo);
      sysassign(prog::TestNo, prog::TestNo - 1)
    else // standard
      userinfo(Text, 2,"  for ",a," = ",_if(TE > 0,  "Error No. ".expr2text(TE), b));
      ERR:= traperror((TIME:= time((r:= context(a)))));
      ERRes:= traperror((res:= context(b)));
      lprint(ERR <> TE or ERRes <> 0 or (ERR + ERRes = 0 and r <> res),
             hold(_equal)(a, b),  " [",
             _if(ERRes > 0, ("TrapError ", ERRes, " result"),
                 _if(ERR > 0, ("TrapError ", ERR), ("<> ", r))), "]", _if(c <> "", " (".c.")", c))
    end_if;
    sysassign(prog::TestNo, prog::TestNo + 1);
    sysassign(prog::TestTime, prog::TestTime + TIME);
    Pref::typeCheck(op(PREF, 1));
    testargs(op(PREF, 2));
//     sysdelete(prog::NOWARN);
    // rebuild system warning
//     sysassign(warning, WARN);
    null()
  end_proc:

// end of file 
