// $Date: 2001/12/11 00:14:23 $  $Author: jngerhar $  $Revision: 1.43 $

//
//  Complex - Domain of property
//            'Properties of complex identifiers'
//

alias(Complex = property::Complex):
alias(RE(x) = op(extop(x, 5), 1)):
alias(IM(x) = op(extop(x, 5), 2)):
alias(_RE = [5, 1]):
alias(_IM = [5, 2]):

Complex:= newDomain("property::Complex"):
Complex::create_dom:=property::hold(Complex):
Complex::info:= "Properties of complex identifiers";
Complex::domtype:= ()->(Type):
Complex::type:= "Complex":
Complex::prop:= TRUE:
Complex::propdom:= slot(property, "Complex"):
Complex::testtype:= Type::testtype:

Complex::noTrace:= {hold(print), hold(testtype), hold(domtype)}:

//Complex::ax:= []:

Complex::new:= 
  proc(re = Type::Real, im = Type::Real)
  begin
    if contains({Type::Imaginary, Type::Complex}, re) then
      error("the real part can only be real")
    elif contains({Type::Imaginary, Type::Complex}, im) then
      error("the imaginary part can only be real")
    end_if;
    new(Complex, "Complex", TRUE, FALSE, FALSE, [re, im])
  end_proc:

Complex::_new:=
  proc(re = Type::Real, im = Type::Real)
  begin
    if contains({Type::Imaginary, Type::Complex}, re) then
      re:= re and Type::Real
    end_if;
    if contains({Type::Imaginary, Type::Complex}, im) then
      im:= im and Type::Real
    end_if;
    new(Complex, "Complex", TRUE, FALSE, FALSE, [re, im])
  end_proc:

Complex::is:= 
  proc(FU1, FU2)
    local i;
  begin
    if type(FU1) = "NRange" then
      FU1:= property::NRange::Complex(FU1)
    end_if;

    if type(FU2) = "NRange" then
      FU2:= property::NRange::Complex(FU2):
      return(property::implies(RE(FU1), RE(FU2)) and property::implies(IM(FU1), IM(FU2)))
    end_if;

    if type(FU2) <> "Complex" then
      return(UNKNOWN)
    end_if;
    if (i:= contains([RE(FU2), IM(FU2)], FAIL)) > 0 then
      if i = 2 then
        return(property::implies(RE(FU1), RE(FU2)))
      else
        return(property::implies(IM(FU1), IM(FU2)))
      end_if
    end_if;
    
    property::implies(RE(FU1), RE(FU2)) and property::implies(IM(FU1), IM(FU2))
  end_proc:

// Uberladung
Complex::contains:=
        proc(CP, nm)
        begin
          if type(CP) <> "Complex" then
            FALSE
          else
            bool(property::_is(nm, CP) = TRUE)
          end_if
        end_proc:

Complex::_and:=
  proc()
  begin
    property::_Op
    (Complex, _and,
     proc(FU1, FU2)
       local re, im, ret;
     begin
       ret:= _if(RE(FU1) = Type::Real, null(), RE(FU1)),
             _if(RE(FU2) = Type::Real, null(), RE(FU2));
       re:= eval(_if((re:= _and(ret)) = TRUE, Type::Real, re));
       ret:= _if(IM(FU1) = Type::Real, null(), IM(FU1)),
             _if(IM(FU2) = Type::Real, null(), IM(FU2));
       im:= eval(_if((im:= _and(ret)) = TRUE, Type::Real, im));
       if property::implies(im, Type::Zero) = TRUE then return(re) end_if;
       if property::Fail(re, im) then
         return(property::Null)
       else
         return(Complex::_new(re, im))
       end_if
     end_proc, args())
  end_proc:

Complex::_or:=
  proc()
  begin
    //return(hold(_or)(args()));
    property::_Op
    (Complex, _or,
     proc(FU1, FU2)
       local re, im, ret;
     begin
       ret:= _if(RE(FU1) = Type::Real, null(), RE(FU1)),
             _if(RE(FU2) = Type::Real, null(), RE(FU2));
       re:= eval(_if((re:= _or(ret)) = FALSE, Type::Real, re));
       ret:= _if(IM(FU1) = Type::Real, null(), IM(FU1)),
             _if(IM(FU2) = Type::Real, null(), IM(FU2));
       im:= eval(_if((im:= _or(ret)) = FALSE, Type::Real, im));
       //if property::implies(im, Type::Zero) then return(re) end_if;
       return(Complex::_new(re, im))
     end_proc, args())
  end_proc:

Complex::_plus:=
  proc()
    local sp, const;
  begin
    if op((sp:= split([args()], testtype, Type::Complex)), 1) <> [] then
      const:= _plus(op(op(sp, 1)));
      if nops(op(sp, 2)) = 1 // 
         and type((sp:= Complex::_plus(op(op(sp, 2))))) = "Complex" then
        return(property::Complex(RE(sp)+Re(const), IM(sp) + Im(const)))
      end_if
    end_if;
    property::_Op
    (Complex, _plus,
     proc(FU1, FU2)
     begin
       Complex::_new(RE(FU1) + RE(FU2), IM(FU1) + IM(FU2))
     end_proc, args())
  end_proc:

Complex::_subtract:=
  proc(FU1, FU2)
  begin
    if testtype(FU2, Type::Real) then
      return(Complex::_plus(FU1, -FU2))
    end_if;

    property::_Op
    (Complex, _subtract,
     proc(FU1, FU2)
     begin
       Complex::_plus(FU1, Complex::_negate(FU2))
     end_proc, args())
  end_proc:

Complex::_negate:=
  proc(FU)
  begin
    subsop(FU, _RE = _negate(RE(FU)), _IM = _negate(IM(FU)))
  end_proc:

Complex::_mult:=
  proc()
    local sp, const;
  begin
    if op((sp:= split([args()], X -> testtype(X, Type::Constant)
                                     and testtype(X, Type::Arithmetical))), 1) <> [] then
      const:= _mult(op(op(sp, 1)));
      sp:= op(op(sp, 2));
      if type((sp:= property::convert_to(Complex::_mult(sp),
                                            Complex))) = "Complex" then
        // split [complex, non real, real]
        if type(const) = "_mult" then
          const:= split(const, X -> _lazy_or(domtype(X) = DOM_COMPLEX,
                                             _lazy_and(testtype(X, Type::Real),
                                                       UNKNOWN)))
        else
          const:= subsop([1, 1, 1],
                         contains([TRUE, UNKNOWN, FALSE],
                                  _lazy_or(domtype(const) = DOM_COMPLEX,
                                           _lazy_and(testtype(const, Type::Real),
                                                     UNKNOWN))) = const)
        end_if;
        if op(const, 3) <> 1 then // real number
          sp:= property::Complex(RE(sp)*op(const, 3), IM(sp)*op(const, 3))
        end_if;
        if op(const, 2) <> 1 then // symbolic constant or expression
          sp:= property::Complex(RE(sp)*op(const, 2), IM(sp)*op(const, 2))
        end_if;
        if (const:= op(const, 1)) <> 1 and const <> null() then // complex number
          if op(const, 1) = 0 then
            sp:= property::Complex(-IM(sp)*op(const, 2), RE(sp)*op(const, 2))
          else
            sp:= property::Complex(RE(sp)*op(const, 1) - IM(sp)*op(const, 2),
                                   RE(sp)*op(const, 2) + IM(sp)*op(const, 1))
          end_if
        end_if;
        return(sp)
      end_if
    end_if;

    property::_Op
    (Complex, _mult,
     proc(FU1, FU2)
     begin
       Complex::_new(RE(FU1)*RE(FU2) - IM(FU1)*IM(FU2),
                     RE(FU1)*IM(FU2) + IM(FU1)*RE(FU2))
       //Type::Complex
       //hold(_mult)(FU1, FU2)
     end_proc, args())
  end_proc:

Complex::_divide:=
  proc(FU1, FU2)
  begin
    if testtype(FU2, Type::Real) then
      return(Complex::_mult(FU1, 1/FU2))
    end_if;

    property::_Op
    (Complex, _divide,
     proc(FU1, FU2)
     begin
       Complex::_mult(FU1, Complex::_invert(FU2))
     end_proc, args())
  end_proc:

Complex::_invert:=
  proc(FU)
    local x;
  begin
    x:= RE(FU)^2 + IM(FU)^2;
    if property::implies(RE(FU), property::Rel(hold(_unequal), 0)) = TRUE then
      Complex::_new(RE(FU)/x and property::Rel(hold(_unequal), 0), IM(FU)/x)
    elif property::implies(IM(FU), property::Rel(hold(_unequal), 0)) = TRUE then
      Complex::_new(RE(FU)/x, IM(FU)/x and property::Rel(hold(_unequal), 0))
    else
      Complex::_new(RE(FU)/x, IM(FU)/x)
    end_if
    //hold(_invert)(FU1)
  end_proc:

Complex::_power:=
  proc(FU1, POW)
  begin // 
    if type(FU1) <> "Complex"
       and type((FU1:= property::convert_to(FU1, Complex))) <> "Complex" then
      return(hold(_power)(FU1, POW))
    end_if;
    if property::_is(POW, Type::Even) = TRUE then
      if property::_is(RE(FU1), property::Rel(hold(_equal), 0)) = TRUE then
        if property::_is(POW, Type::Residue(0, 4)) = TRUE then
          return(IM(FU1)^POW)
        elif property::_is(POW, Type::Residue(2, 4)) = TRUE then
          return(-IM(FU1)^POW)
        end_if;
        return(hold(_or)(IM(FU1)^2, -IM(FU1)^2))
      else
        return(Type::Complex)
      end_if
    elif property::_is(POW, Type::Real) = TRUE then
      // ???
      return(Type::Complex)
    end_if;
    hold(_power)(FU1, POW)
  end_proc:

// Complex::simplify:=
//   proc(CP)
//   begin
//     
//   end_proc:

Complex::NRange:= 
  proc(CP)
  begin
    if property::implies(IM(CP), Type::Zero) = TRUE then
      return(property::convert_to(RE(CP), property::NRange))
    elif property::implies(RE(CP), Type::Zero) = TRUE then
      case property::implies(IM(CP), Type::NonZero)
        of TRUE do
          return(property::NRange("Imaginary", "Zero"));
          break
        of FALSE do
          return(Type::Zero)
        otherwise
          return(Type::Imaginary)
      end_case
    end_if;
    FAIL
  end_proc:

//Complex::Residue:= 
//  proc(CP)
//  begin
//    FAIL
//  end_proc:

Complex::Rel:=
  proc()
  begin
    property::Null
  end_proc:

alias(SB = solvelib::BasicSet):
alias(DI = Dom::ImageSet):
Complex::prop2set:=
  proc(CP)
    local ID1, ID2;
  begin
    if IM(CP) = Type::Zero then
      property::prop2set(RE(CP))
    elif RE(CP) = Type::Real and IM(CP) = Type::Real then
      SB(Dom::Complex)
    else
      ID1:= genident("X");
      ID2:= genident("Y");
      DI(ID1 + ID2*I, [ID1, ID2], [property::prop2set(RE(CP)), property::prop2set(IM(CP))])
    end_if
  end_proc:
unalias(DI):
unalias(SB):

Complex::Re:=
  proc(CP)
  begin
    RE(CP)
  end_proc:

Complex::Im:=
  proc(CP)
  begin
    IM(CP)
  end_proc:

Complex::exp:=
  proc(CP)
  begin
    CP
  end_proc:

Complex::abs:=
  proc(CP)
  begin
    Type::NonNegative
  end_proc:

// Complex::sign:=
//   proc(CP)
//   begin
// //noch nicht fertig!!!
//     Complex::_new(sign(RE(CP)), sign(IM(CP)))
//   end_proc:

Complex::hull:=
  proc(CP)
  begin
    hull(RE(CP)) + I*hull(IM(CP))
  end_proc:

Complex::print:=
  proc(XX)
    local REAL, ZERO;
  begin
    // 
    REAL:= (_if(property::convert_to(RE(XX), Type::Real, hold(Exact)) = Type::Real, 0, 1)
            + (_if(property::convert_to(IM(XX), Type::Real, hold(Exact)) = Type::Real, 0, 2)));
    ZERO:= (_if(property::convert_to(RE(XX), Type::Zero, hold(Exact)) = Type::Zero, 0, 1)
            + (_if(property::convert_to(IM(XX), Type::Zero, hold(Exact)) = Type::Zero, 0, 2)));
    if REAL = 0 then
      Type::Complex
    elif ZERO = 0 then 
      Type::Zero
    elif ZERO = 2 and REAL = 2 then
      Type::Imaginary
    elif ZERO = 1 then
      RE(XX)
    else
      (_if(REAL mod 2 = 1, ("Re(.) " .
                            (_if((RE(XX))::dom::Z <> FAIL, (RE(XX))::dom::Z(RE(XX)), ""))
                            . expr2text(RE(XX))), "")) .
      (_if(REAL = 3, " and ", "")).
      (_if(REAL - 1 > 0, ("Im(.) " .
                          (_if((IM(XX))::dom::Z <> FAIL, (IM(XX))::dom::Z(RE(XX)), ""))
                          . expr2text(IM(XX))), ""))
    end_if
  end_proc:

unalias(Complex):
