// $Date: 2001/11/13 13:03:17 $ $Author: jngerhar $ $Revision: 1.19 $ 

/*--
	Ei/Series -- the function attribut "series" for Ei
--*/

// Note: Ei is equal to E_1 in Chapter 5 of Abramowitz & Stegun

Ei::series := proc(f,x,n,dir) // series(Ei(f),x=0,n) 
    local s,t,i,k,tt,a;
begin
  if f = x then // expansion of Ei(x) for x = 0 
     if n = 0 then
       Series::error("order too small")
     end_if;
     // 5.1.11
     Series::Puiseux::create(1, 0, n, [-ln(x) - EULER] . Series::gen["Ei"](n),
                                x, 0, dir)
   else
     t := Series::series(f, x, n, dir);
     if domtype(t) = Series::Puiseux then
       k := Series::Puiseux::ldegree(t);
       if k = FAIL then // t = O(..)
         Series::error("order too small");
       elif k > 0 then // f goes to zero 
         if n = 0 then
           Series::error("order too small")
         end_if;
         // 5.1.11
         s := Series::Puiseux::create(1, 1, n, Series::gen["Ei"](n),
                                      x, 0, dir);
         return((s @ t)
                - Series::series(EULER + ln(f), x, n + 1, dir))
       elif k < 0 then // f goes to +/-infinity 
         // 5.1.51
         tt := 1;
         s:=Series::Puiseux::create(1,1,n+1,[1,(tt:=-tt*i)$i=1..n-1],x,0,dir);
         s:=Series::Puiseux::_fconcat(s, 1/t);
         return(s * Series::series(exp(-f), x, n, dir))
       elif k = 0 then // expansion around a finite point <> 0
         a := lcoeff(t);
         if is(a < 0) = TRUE then
           // expansion on the branch cut
           return(Series::Puiseux::const((1 - signIm(f))*PI*I, x, n, dir)
                  + Series::unknown(Ei(f),x,n,dir))
         end_if
       end_if;
       Series::unknown(Ei(f),x,n,dir)
     else
        userinfo(2, "unable to compute series expansion");
        FAIL
     end_if
  end_if
end_proc:

// ensure that domain Series is loaded
Series:

// expansion of Ei(x) + EULER + ln(x) around x = 0, formula 5.1.11
// = -x/1*1! + x^2/2*2! - x^3/3*3! +- ...
Series::gen["Ei"] := proc(n)
  local t;
begin
  [(t := 1), (t := -t*(i - 1)/i^2) $ i = 2..(n - 1)]
end_proc:
Series::gen["Ei"](1):=[]:
Series::gen["Ei"](2):=[1]:

// end of file 
