
// handles the call series(arccoth(f), x, n, dir)

arccoth::series := proc(f,x,n,dir)
   local a, l, t, d;
begin

   if dir <> Undirected then // directional expansion
     l:= limit(f, x, dir);
     if has(l, hold(limit)) or l = undefined then
       return(FAIL)
     elif iszero(l) or has(l, infinity) then
        // use arccoth(x) = arctanh(1/x)
        return(arctanh::series(1/f, x, n, dir))
     end_if
   end_if;

   // recursively expand the argument
   t := Series::series(f, x, n, dir);

   if domtype(t) = Series::Puiseux then
     d := ldegree(t);
     if d = FAIL then
	d:=Series::Puiseux::order(t);
        if d > 0 then
          return(Series::Puiseux::zero(x, 0, dir)) // O(1)
        else
	  Series::error("order too small");
          return(FAIL);
        end_if
     elif d = 0 then
       // expansion around a finite point a <> 0
       a := lcoeff(t);
       if a = -1 then
         // branch point
         return(Series::series(1/2*(ln((1/f + 1)*1/2) - ln(1 - (1/f + 1)*1/2)),
                               x, n, dir))
       elif a = 1 then
         // branch point
         return(Series::series(-1/2*(ln(-(1/f - 1)*1/2) - ln(1 + (1/f - 1)*1/2)),
                               x, n, dir))

       elif is(a, Type::Real)  = TRUE then
         // expansion on the branch cut:
         // add correction term; the conjugate is necessary in order
         // to make the expansion correct for x on the imaginary axis
         // as well
         if is(0 < a < 1) = TRUE then
           return(I*(signIm(conjugate(f)) + 1)*PI/2
                  + Series::unknown(arccoth(f), x, n, dir))
         elif is(-1 < a < 0) = TRUE then
           return(-I*(-signIm(conjugate(f)) + 1)*PI/2
                  + Series::unknown(arccoth(f), x, n, dir))
         end_if
       end_if
     else // d > 0 or d < 0
       // expansion around 0 or +-infinity: use arccoth(x) = arctanh(1/x)
       return(arctanh::series(1/f, x, n, dir))
     end_if
   end_if;

   Series::unknown(arccoth(f),x,n,dir);

end_proc:
