
// handles the call series(arcsin(f), x, n, dir)
arcsin::series := proc(f,x,n,dir)
   local s, l, k, t, d, a;
begin

   if dir <> Undirected then // directional expansion
     l:=limit(f, x, dir);
     if has(l, hold(limit)) or l = undefined then
       return(FAIL)
     elif iszero(l) then
       if n<1 
         then s:= []
         else s:= 1; 
              s:=[s,(0,(s:=s*(2*k-1)/2/k)/(2*k+1))$k=1..(n+1 div 2)-1]
       end_if;
       s:=Series::Puiseux::create(1,1,2*((n+1) div 2)+1,s,x,0,dir);
       if f=x
         then return(s)
         else return(Series::Puiseux::_fconcat(s,Series::series(f,x,n,dir)))
       end_if
     elif iszero(l-1) then
       s:=1; l:=1; f:=1-f;
       for k from 1 to n-1 do
         l:=l*f*(2*k-1)/4/k;
         s:=s+l/(2*k+1)
       end_for;
       return(PI/2 - Series::series(sqrt(2*f)*s,x,n,dir));
     elif iszero(l+1) then
        s:=1; l:=1; f:= 1+f;
        for k from 1 to n-1 do
           l:=l*f*(2*k-1)/4/k;
           s:=s+l/(2*k+1)
        end_for;
        return(-PI/2 + Series::series(sqrt(2*f)*s,x,n,dir));
     elif has(l, infinity) then
        if l = -infinity then
          // due to cancellation in the argument of ln, we need to increase
          // the order
          return(Series::series(-I*ln(I*f+sqrt(1-f^2)), x, n + 1, dir));
        else
          return(Series::series(-I*ln(I*f+sqrt(1-f^2)), x, n, dir));
        end_if
     end_if;
  end_if;

  // recursively expand the argument
  t := Series::series(f, x, n, dir);

  if domtype(t) = Series::Puiseux then
    d := ldegree(t);
    if d=FAIL then // t = O(..)
	d:=Series::Puiseux::order(t);
        if d > 0 then return(t)
        else
	  Series::error("order too small");
          return(FAIL);
        end_if
    end_if;
    if d >= 0 then
      a := coeff(t, x, 0);
      case a
      of 0 do
        if n<1 
          then s:= []
          else s:= 1; 
               s:=[s,(0,(s:=s*(2*k-1)/2/k)/(2*k+1))$k=1..(n+1 div 2)-1]
        end_if;
        s:=Series::Puiseux::create(1,1,2*((n+1) div 2)+1,s,x,0,dir);
        if f=x
          then return(s)
          else return(Series::Puiseux::_fconcat(s, t));
        end_if
      of 1 do
        s:=1; l:=1; f:=1-f;
        for k from 1 to n-1 do
          l:=l*f*(2*k-1)/4/k;
          s:=s+l/(2*k+1)
        end_for;
        return(PI/2 - Series::series(sqrt(2*f)*s,x,n,dir));
      of -1 do
        s:=1; l:=1; f:= 1+f;
        for k from 1 to n-1 do
           l:=l*f*(2*k-1)/4/k;
           s:=s+l/(2*k+1)
        end_for;
        return(-PI/2 + Series::series(sqrt(2*f)*s,x,n,dir));
      end_case;
      // expansion on the branch cut:
      // add correction term for the upper/lower half plane, respectively
      if is(a > 1) = TRUE then
        return(Series::Puiseux::const(PI/2, x, n, dir)
               + signIm(f)*Series::unknown(arccos(f),x,n,dir))
      elif is(a < -1) = TRUE then
        return(Series::Puiseux::const(-PI/2, x, n, dir)
               + signIm(f)*Series::unknown(arccos(-f),x,n,dir))
      end_if
    else // ldegree(t) < 0
      return(FAIL);
    end_if
  end_if;

  Series::unknown(arcsin(f),x,n,dir);

end_proc:
