// $Date: 2001/12/08 22:41:06 $ $Author: jngerhar $ $Revision: 1.26 $ 

/*--
	ln/Series -- the function attribut "series" for ln
--*/

ln::series := proc(f, x, n, dir)
   local s, t, a, k;
begin
   // s := series(ln(1 + x), x)
   s := Series::Puiseux::create(1, 1, n + 1, Series::gen["ln"](n), x, 0, dir);

   // handle special cases first
   if f = 1 + x then
      s
   elif f = 1 - x then
      Series::Puiseux::create(1, 1, n + 1,
        map(Series::gen["ln"](n), _negate@abs), x, 0, dir)
   else

      // recursively expand the argument into a series
      t := Series::series(f, x, n + 1, dir);
      if domtype(t) <> Series::Puiseux then
        return(FAIL)
      end_if;

      k := Series::Puiseux::ldegree(t);
      a := Series::Puiseux::lcoeff(t);
      // f = t = a*x^k + higher order terms

      if k = FAIL then // t = O(..)
        Series::error("order too small")

      elif k = 0 then // expansion around the finite point a <> 0
         //
         // f = a*(1 + higher order terms)
         //

         // t := series(ln(1 + higher order terms))
         t := Series::Puiseux::lmonomial(t, Rem)[2]; // t := t - a
         if a <> 1 then
           t := Series::Puiseux::scalmult(t, 1/a)    // t := t/a
         end_if;
         t := Series::Puiseux::_fconcat(s, t);

         if is(a < 0) = TRUE then
           //
           // expansion on the branch cut:
           // ln(f) = signIm(f)*I*PI + ln(-f)           
           //       = signIm(f)*I*PI + ln(-a) + ln(1 + higher order terms)
           //
           Series::Puiseux::const(signIm(f)*I*PI + ln(-a), x,
                                  Series::Puiseux::order(t), dir) + t
         else
           //
           // ln(f) = ln(a) + ln(t)
           //
// This may still be wrong if a contains x or a can be simplified to 0
// or a negative number!
           Series::Puiseux::const(ln(a), x, Series::Puiseux::order(t), dir) + t
         end_if

      else // expansion around 0 or +-infinity
         //
         // f = a*x^k*(1 + higher oder terms)
         // ln(f) = ln(a*x^k) + ln(1 + higher order terms)
         // This is valid only when a*x^k ist not real negative!
         //

         if n = 0 then
           Series::error("order too small")
         end_if;
         
         // t := series(ln(1 + higher order terms))
         t := Series::Puiseux::lmonomial(t, Rem)[2]; // t := t - a*x^k
         t := Series::Puiseux::scalmult(t, 1/a, -k); // t := t/(a*x^k) 
         t := Series::Puiseux::_fconcat(s, t);

         if not has(a, x) then
           // try to split ln(a*x^k), if possible

           if dir = Right or is(x >= 0) = TRUE then
             return(t + Series::Puiseux::const(ln(a) + k*ln(x), x, n, dir))
           elif dir = Left or is(x <= 0) = TRUE then
             // we might pull out (-1)^k as well here
             return(t + Series::Puiseux::const(ln((-1)^k*a) + k*ln(-x), x, n, dir))
           elif is(a >= 0) = TRUE then
             if k > -1 and k < 1 then
               return(t + Series::Puiseux::const(ln(a) + k*ln(x), x, n, dir))
             else
               return(t + Series::Puiseux::const(ln(a) + ln(x^k), x, n, dir))
             end_if
           elif dir = Real or is(x, Type::Real) = TRUE then
             if domtype(k/2) = DOM_INT then // k is even
               return(t + Series::Puiseux::const(ln(a) + ln(x^k), x, n, dir))
             elif domtype((k - 1)/2) = DOM_INT then // k is odd
               return(t + Series::Puiseux::const(ln(-a) + ln(-x^k), x, n, dir))
             end_if;
           elif is(a <= 0) = TRUE then
             return(t + Series::Puiseux::const(ln(-a) + ln(-x^k), x, n, dir))
           end_if;
         end_if;

         // default: no splitting
         t + Series::Puiseux::const(ln(a*x^k), x, n, dir)
      end_if
   end_if
end_proc:

// ensure that domain Series is loaded
Series:

// expansion for ln(1 + x) around x = 0
Series::gen["ln"] := proc(n)
   local t, i;
begin
   t := -1;
   [((t := -t)/i $ i = 1..n)]
end_proc:

// end of file 
