// $Date: 2001/11/15 14:49:30 $ $Author: jngerhar $ $Revision: 1.18 $ 

/*--
	sin/Series -- the function attribut "series" for sin
--*/

sin::series := proc(f,x,n,dir)
   local t,k,a,ts,tc;
begin
   if f=x then
      Series::Puiseux::create(1,1,2*((n+1) div 2)+1,Series::gen["sin"](n),x,0,dir)
   else
      t:=Series::series(f,x,n,dir);
      if domtype(t)<>Series::Puiseux then
         return(FAIL)
      end_if;
      k:=ldegree(t);
      if k=FAIL then // t = O(..)
	 k:=Series::Puiseux::order(t);
         if k > 0 then
           return(t)
         else
	   Series::error("order too small");
         end_if
      elif k>0 then
         Series::Puiseux::_fconcat(
            Series::Puiseux::create(1,1,2*((n+1) div 2)+1,Series::gen["sin"](n),x,0,dir)
         ,t)
      elif k=0 then
         a:=lcoeff(t);
         t:=f-a;
         // sin(a+t)=sin(a)*cos(t)+cos(a)*sin(t) 
         if iszero(sin(a)) then
           cos(a)*sin::series(t,x,n,dir)
         else
           sin(a)*cos::series(t,x,n,dir)+cos(a)*sin::series(t,x,n,dir)
         end_if
      else // expansion around infinity
        if dir <> Undirected then
          // directed expansion
          // a := non-positive order part of t
          a := expr(Series::Puiseux::truncate(t, 0));
          if coeff(t, 0) <> FAIL then
            a := a + coeff(t, 0)
          end_if;
          if contains({infinity, -infinity}, limit(a, x, dir)) then
            f := normal(f - a);
            if iszero(f) then
              Series::Puiseux::const(sin(a), x, n, dir)
            else
              // cos(a)*sin(f) + sin(a)*cos(f)
              ts := Series::series(sin(f), x, n, dir);
              tc := Series::series(cos(f), x, n, dir);
              Series::Puiseux::scalmult(ts, cos(a))
              + Series::Puiseux::scalmult(tc, sin(a))
            end_if;
          else
            userinfo(2, "unable to compute series expansion");
            FAIL
          end_if
        else // undirected expansion
          userinfo(2, "unable to compute series expansion");
          FAIL
        end_if;
      end_if
   end_if
end_proc:

// ensure that domain Series is loaded
Series:

Series::gen["sin"]:=proc(n) local i,t;
begin
   t:=1;
   [1,(0,(t:=-t/(2*i)/(2*i+1)))$i=1..(n-1) div 2]
end_proc:
Series::gen["sin"](0):=[]:
Series::gen["sin"](1):=[1]:

// end of file 
