/* $Date: 2002/02/18 16:20:10 $ $Author: tonner $ $Revision: 1.11 $ */

/* solvelib::Union


   Given a set1 whose elements contain a parameter x,
   and a set2, return the set of all objects that can
   be obtained by substituting x in some element of set1
   by some element of set2

*/

solvelib::Union:=
proc(set1, x, set2)
local i, l, S, imgsets;

begin

  if type(set2) = DOM_SET then
    return(_union(eval(subs(set1, x = i)) $i in set2))
  end_if;
  
  if type(set2)=piecewise then
     // do case analysis w.r.t. parameter set
    if has(set2, x) then 
      error("Index set of Union must not depend on index variable")
    end_if;
    return(piecewise::extmap(set2, z-> solvelib::Union(set1, x, z)))
  end_if;
  
  // try overloading
  if set1::dom::Union <> FAIL then
    return(set1::dom::Union(args()))
  end_if;
  
  case type(set1)
    of DOM_SET do
      l:=[];
      imgsets:=[];
      for i from 1 to nops(set1) do
        S:=solvelib::substituteBySet(op(set1,i), x, set2);
        if type(S) = Dom::ImageSet then
          imgsets := append(imgsets, S)
        else
          l:= append(l, S)
        end_if;
      end_for;
      l:= _union(op(l));
      case nops(imgsets)
        of 0 do
          return(l)
        of 1 do
          return(l union op(imgsets, 1))
        otherwise
          // avoid computing a union of ImageSets
          return(_union(l, op(imgsets)))
      end_case
    of "solve" do
    of "RootOf" do
      return(procname(args()))
    of "_union" do
    of "_intersect" do
      return(map(set1, solvelib::Union, x, set2))
    of "_minus" do
      return(procname(args()))
    otherwise
      // error("unknown set type ".expr2text(type(set1)))
      return(procname(args()))
  end_case
 end_proc:

solvelib::Union:= funcenv(solvelib::Union):


solvelib::Union::type := "Union":
solvelib::Union::info := "solvelib::Union --- union of a system of sets":
solvelib::Union::print:= "solvelib::Union":

solvelib::Union::Content := proc(Out, data)
                     begin
                       Out::Capply(Out::CUnion,
                                   Out(extop(data, 1)),
                                   Out(extop(data, 2)),
                                   Out(extop(data, 3)));
                     end_proc:





