/* */
/* $Date: 2002/02/26 10:15:05 $ $Author: stefanw $ $Revision: 1.33 $ */
/* */
/* zimmerma, 14/12/95 */

/*++
        solvelib::algebraic  --  solve algebraic equations 

	algebraic(sys,vars <,domain>)

	sys   : set of algebraic equations
	vars  : set of unknowns
	domain: (optional) coefficient domain
++*/

solvelib::algebraic :=
proc(syss, unkk, domm)
  local t, sys, unk, dm, options;
begin
  if testargs() then
    if domtype(syss) <> DOM_SET then
      error("expecting set of equations")
    end_if;
    case args(0)
      of 1 do
        break
      of 2 do
        if domtype(unkk) <> DOM_SET and domtype(unkk) <> DOM_DOMAIN then
          error("invalid second argument")
        end_if;
        break
      of 3 do
        if domtype(unkk) <> DOM_SET or domtype(domm) <> DOM_DOMAIN then
          error("invalid arguments")
        end_if;
        break
      otherwise
        error("wrong no of args")
    end_case
  end_if;

  sys:=syss;

  if args(0) = 1 then
    dm:= null();
    unk:= indets(sys);
    options:= solvelib::getoptions()
  elif args(0) = 2 then
    unk:=unkk;
    if domtype(unk) = DOM_SET then
      t:= indets(sys);
      if nops(unk) <> nops(t) then
        userinfo(1,"found ",t," as unknowns");
        error("wrong no of unknowns given")
      end_if;
      dm:= null()
    else
      dm:= unk;
      unk:= indets(sys)
    end_if;
    options:= solvelib::getoptions()
  else
    unk:=unkk;
    dm:=domm;
    options:= solvelib::getoptions(args(4..args(0)))
  end_if;


  sys:=map
        (sys,
         proc(x)
           name equationsToExpressions;
         begin
           if type(x)="_equal" then
             op(x,1) - op(x,2)
           else
             x
           end_if
         end_proc
         );

  solvelib::solve_algebraic(sys, unk, dm, options)
end_proc:


/*--
	solvelib::solve_algebraic(sys,unk,dm, options) solves the algebraic
	system 'sys' with respect to the unknowns 'unk' over the
	coefficient field 'dm'.

	If 'unk' is a list, then it gives the elimination order.
--*/

solvelib::solve_algebraic:=
proc(syss, unkk, domm) 
  local g, gb, reord, sys, unk, dm, options, storeLeadingCoeffs,
  lcoefflist: DOM_LIST, linear, lc, i, inds,
  indepSets, 
  monic, cond, dimen;
  
begin

  lcoefflist:= [];
  
  // procedure storeLeadingCoeffs
  storeLeadingCoeffs:=
  proc(f, o)
    local l;
  begin
    l:= lcoeff(f, o);
    if testtype(expr(l), Type::Constant) <> TRUE then
      lcoefflist:= append(lcoefflist, l)
    end_if;
    multcoeffs(f, 1/l)
  end_proc;
    
  sys:= syss;
  unk:= unkk;
  if args(0) >= 3 then
    dm:= domm
  else
    dm:= NIL
  end_if;
  options:= solvelib::getoptions(args(4..args(0)));
  if type(unk) = DOM_SET then
    unk:= [op(unk)];
    reord:= Reorder
  else
    reord:= null()
  end_if;
  userinfo(3, "Entering solver for algebraic systems");
  if dm= NIL then
    if map(sys, testtype, Type::PolyExpr(unk,Type::Rational)) = {TRUE}
      then
      dm:=Expr;
      monic:= groebner::primpart
    else
      dm:=Dom::ExpressionField(normal, iszero@normal);
      monic:= storeLeadingCoeffs
    end_if
  end_if;
  
  sys:= map([op(sys)], poly, unk, dm);
  /* first compute a basis wrt DegInvLexOrder */
  gb:=groebner::gbasis(sys, DegInvLexOrder, reord, hold(Monic) = monic);
  /* then transform it in a LexOrder basis (the Grobner walk) */
  g:=groebner::gbasis(gb, LexOrder, reord, hold(Monic)= monic );

  if nops(g) = 1 and degree(g[1])=0 then
    if iszero(g[1]) then
      userinfo(20, "Zero equation: everything is a solution")
    else
      g:= null() /* no solution */
    end_if;
    indepSets:= {};
    
  else  

  // it may be that the order of unknowns has to be exchanged
    indepSets:= groebner::stronglyIndependentSets(g);
    userinfo(10, "Independent sets are :".expr2text(indepSets[3]));
    dimen:= indepSets[1];
    indepSets:= indepSets[2];
  // the number of indep. var. should equal the dimension of the variety
    assert(dimen = nops(indepSets));
  // now indepSets is a set x1, x2, ..., xn of variables,
  // such that the solutions to the system can be expressed as
  // [x1 = ..., x2 = ..., ..., xn = ...]

    unk:= split(unk, v -> not contains(indepSets, v));
    assert(unk[3] = []);

    if nops(unk[2])= 0 then
      userinfo(10, "Zero-dimensional system")
    else
      dm:=Dom::ExpressionField(normal, iszero@normal);
      monic:= storeLeadingCoeffs;
      gb:= map(gb, poly, unk[1], dm);
    // the grobner basis has to be re-computed, and the order of variables
    // must not be changed
      reord:= null();
      lcoefflist:= [];
      g:=groebner::gbasis(gb, LexOrder, hold(Monic)= monic );
    end_if;
      
  
    userinfo(2,"Groebner basis is",map(g, expr));
    userinfo(2,"Groebner basis is valid if the following leading coefficients ".
    "are all nonzero:".expr2text(lcoefflist));
  
    g:=groebner::factor(g, reord);
    g:=map(g, op@groebner::triangular_set);
    g:=op(map(g, solvelib::allvals, {op(unk[1])}, options));
    unk:= unk[1]
  end_if;

  
  if nops(lcoefflist) = 0 then
    return(solvelib::free2Bound({g}, indepSets))
  else
    lcoefflist:= map(lcoefflist, expr);
    cond:= _and(op(map(lcoefflist, _unequal, 0)));
    // search for degree 1 leading coefficient:
    // if it exists, we also solve the special case
    linear:= FAIL;
    for lc in lcoefflist do
      inds:= indets(lc);
      for i in inds do
        if testtype(lc, Type::PolyExpr([i])) and
          degree(lc, [i]) = 1 then
          linear:= [lc, i];
          break
        end_if
      end_for;
      if linear <> FAIL then break; end_if
    end_for;
    
    
    if linear <> FAIL then
      poly(linear[1], [linear[2]]);
      return(solvelib::free2Bound
             (
              piecewise([cond, {g}],
                [linear[1] = 0,
                 solve(subs(syss,
                            linear[2] = -coeff(%,0) / coeff(%, 1)),
                       unk)] ),
             indepSets))
    end_if;
    
    // if nothing helps, accept also higher degree polynomials

    for lc in lcoefflist do
      inds:= indets(lc);
      for i in inds do
        if testtype(lc, Type::PolyExpr([i])) then
          linear:= [lc, i];
          break
        end_if
      end_for;
      if linear <> FAIL then break; end_if
    end_for;

    // attach the condition ,,leading coeff = 0'' as
    // additional equation to the polynomial system, and start over
    if linear <> FAIL then
      return(solvelib::free2Bound
             (
              piecewise([cond, {g}],
                        [not cond, solvelib::bound2Free
                         (solvelib::solve_algebraic
                          (syss union {linear[1]}, unk.[i]), {i})]),
             indepSets))
    else  
      // only non-polynomial leading coefficients
      solvelib::free2Bound(piecewise([cond, {g}]), indepSets)
    end_if
  end_if 
end_proc:


solvelib::allvals :=
proc(g, unkk, options) 
  local x, sol, i, j, r, n, foo, unk;
  
begin
  unk:=unkk;
  sol:=[[]];
  /* foo(l,x,j) adds x=j in front of partial solution l */
  if options["BackSubstitution"]=TRUE then
    foo:=proc(l, x, j)
           name makeEqAndBackSubstitute;
         begin
           select([op(l)], not hastype, "RootOf");
           [x=subs(j,op(%)),op(l)]
         end_proc
  else
    foo:=proc(l,x,j)
           name makeEq;
         begin
           // better: ?
           // if type(j)="RootOf" then
           //  [x in j, op(l)]
           // else  
             [x=j,op(l)]
           // end_if  
         end_proc
  end_if;

  // if the ideal is <1>, there is no solution
  if nops(g)=1 and degree(g[1])=0 then
    return(null())
  end_if;

  
  for i from nops(g) downto 1 do
    r:=g[i];
    x:=indets(op(r,1)) intersect unk;
    if nops(x)=0 then
      next
    end_if;
    /* choose the largest unknown */
    for j from 1 to nops(op(r,2)) do
      if contains(x,op(r,[2, j])) then
        x:=op(r, [2, j]);
        break
      end_if
    end_for;

    unk:=unk minus {x};  // !!!! modified 4.12.00 Walter

    // remove not only x, but also all unknowns after x

    //  unk:= unk minus {op(op(r,2), k) $k=j..nops(op(r,2))}; //!!! modified 4.12.00 Walter

    j:=solvelib::solve_poly(poly(op(r,1),[x]),x,options);
    if type(j)=DOM_SET and property::hasprop(x) then
      // select only those solutions that do not contradict
      // the properties of x
      j:=select(j, s -> is(s, getprop(x))<>FALSE )
    end_if;
    userinfo(5, "Solving polynomial ".expr2text(op(r,1))." gives ".
             expr2text(j));
    // assume that the remaining unknowns can be chosen suitably
    // do not take into account zeros of denominators etc. (to be changed)

    if type(j) = piecewise then
      j:= piecewise::disregardPoints(j);
      if type(j) = piecewise then
        j:= piecewise::assumeIdentsTRUE(j, unk)
      end_if;
    end_if;

    case type(j)
      of "RootOf" do
        sol:=map(sol, foo, x, j);
        break
      of "_union" do
        // given a union of sets and RootOfs
        // {a,b,c} union RootOf(..) union RootOf(..) ...,
        // call foo (x, a), foo(x, b), foo(x, c), foo(RootOf(..)) ...
        // and concatenate all these
        // eval is necessary for flattening
        eval(map([op(j)], S -> (if type(S)=DOM_SET then op(S) else S end_if)));
        sol:=_concat(map(sol, foo, x, op(%, i)) $i=1..nops(%));
        break
      of solvelib::BasicSet do
        break;
      // nothing to do here !
      of DOM_SET do
        if (n:=nops(j))=1 then
          sol:=map(sol,foo,x,op(j))
        elif n=0 then
          // all solutions have been removed, since they
          // were inconsistent with the properties of the variable
          // -> there is no solution to the system
          return(null())
        else
          sol:=_concat(map(sol,foo,x,op(j,i)) $i=1..n)
        end_if;
    end_case;
    userinfo(5, "Solutions so far are ".expr2text(sol));


  end_for;
  op(sol)
end_proc:

/* end of file */

