
// solvelib::free2Bound

// take a solution from the system solver
// and make the elements of vars from parameters to variables  



solvelib::free2Bound:=
proc(sol, vars: DOM_SET)
  local i, doBranch;
begin

  // quick exit
  if vars={} then
    return(sol)
  end_if;
  
  // doBranch: extract the solutions from [cond, object]
  // and add them to solutions
  
  doBranch:=
  proc(cond, object, solutions: Type::Set)
    local inds, i, newobject, originalSolutions;
  begin    
    case type(cond)
      of DOM_BOOL do
        if cond = TRUE then
          return(solutions union object)
        else
          return(solutions)
        end_if;
      of "_not" do
      of "_unequal" do  
        // leave out the condition !? 
        // this is wrong, but agrees with the usual handling
        // of special cases
        return(solutions union object)
      of "_in" do
        if op(cond, 2) = C_ then
          // this is no restriction
          return(solutions union object)
        end_if;
        expand(cond);
        if type(%) <>"_in" then
          return(doBranch(%, object, solutions))
        elif contains(vars, op(cond, 1)) then
          if type(object) = DOM_SET then
            solutions:= solutions union map(object, append, cond)
          else            
            solutions:= solutions union
                        (object intersect solve(cond, op(cond, 1)))
          end_if
        else
          solutions:= solutions union
                        (object intersect solve(cond, op(inds, 1)))
        end_if;
        break
      of "_equal" do
      // of "_unequal" do  
        // usual case
        if contains(vars, lhs(cond)) then
          if type(object) = DOM_SET then
            assert(map(object, type) minus {DOM_LIST} = {});
            solutions:= solutions union map(object, append, cond)
          else
            solutions:= solutions union
                        (object intersect solve(cond, lhs(cond)))
          end_if
        elif (inds:=vars intersect indets(cond)) = {} then
          // this solutions remains piecewise
          solutions:= solutions union piecewise([cond, object], [not cond, {}])
        else
          solutions:= solutions union
                        (object intersect solve(cond, op(inds, 1)))
        end_if;
        break
      of "_or" do
        // accumulate results in solutions
        for i from 1 to nops(cond) do
          solutions:= doBranch(op(cond, i), object, solutions)
        end_for;
        break
      of "_and" do
        originalSolutions:= solutions;
        newobject:= object;
        for i from 1 to nops(cond) do
          // compute new object
          newobject:= doBranch(op(cond, i), newobject, originalSolutions);
        end_for;
        return(newobject)
      otherwise
        // unknown type, return original
        return(solutions union  piecewise([cond, object], [not cond, {}]))
    end_case;
    return(solutions)
  end_proc;
 

  case type(sol)
    of DOM_SET do
      // there is nothing to do!
      return(sol)
    of piecewise do
      // convert conditions into equations
      return(_union(
        doBranch(piecewise::condition(sol, i),
                 piecewise::expression(sol, i), {})
      $i=1..nops(sol)))
    of "solve" do
      // return an unevaluated solve-call, but with additional
      // variables from vars
      case type(op(sol, 2))
        of DOM_SET do
          op(sol, 2) union vars; break
        of DOM_LIST do  
          op(sol, 2).[op(vars)]; break
        of DOM_IDENT do
          {op(sol, 2)} union vars; break
        otherwise
          error("Unknown type of second argument to solve-call")
      end_case;
      return(subsop(sol, 2= %))
    otherwise
      error("Unknown type of solution")
  end_case
  
end_proc:

