// $Date: 2002/02/27 12:42:03 $ $Author: bij $ $Revision: 1.35 $ 

// frankp, 01/02/95 

/*++
	abs -- returns the absolute value of an expression

	abs(x)

	x - expression
++*/

abs := funcenv
(
 proc(x)
   name abs;
   option noDebug;
   local t, aa, a, f;
 begin
   if args(0) = 0 then
     error("no arguments given")
   elif x::dom::abs <> FAIL then
     return( x::dom::abs(args()) )
   elif args(0) <> 1 then
     error("wrong no of args")
   end_if;

 case type(x)
   of DOM_SET do
   of "_union" do
     return(map(x, abs))
 end_case;
 
 if not testtype(x,Type::Arithmetical) then
   error("argument must be of 'Type::Arithmetical'")
 end_if;

 case domtype(x)
   of DOM_INT     do
   of DOM_RAT     do
   of DOM_FLOAT   do
   of DOM_COMPLEX do
     return( specfunc::abs(x) )
   of DOM_IDENT do
     // PI, EULER, CATALAN are handled by assignment below
     if property::hasprop(x) then
       if is(x>=0) = TRUE then
         return(x)
       elif is(x<=0) = TRUE then
         return(-x)
       end_if
     end_if;
     return(procname(x));
     // NOT REACHED
     assert(FALSE)
   of DOM_EXPR    do
     t := op(x,0);
     if domtype(eval(%)) = DOM_FUNC_ENV then
       if slot(eval(%),"abs") <> FAIL then
         return( slot(eval(%),"abs")(op(x)) )
       end_if
     end_if;

     case t
       of hold(_mult) do
         aa := 1;
         a := 1;
         for f in x do
           t := abs(f);
           if type(t)="abs" and op(t)=f then
             aa := aa*f
           else
             a := a*t
           end_if
         end_for;
         if aa <> x then
           if aa <> 1 then
             return( a*procname(aa) )
           else
             return( a )
           end_if
         end_if;
         break
       of hold(_power) do
         if testtype( op(x,2),Type::Real ) then
           return( abs(op(x,1))^op(x,2) )
         end_if;
         break
       of hold(abs) do
         return(x)
     end_case
 end_case;

 if testtype(x, Type::Constant) then
   if not hastype((a:= Re(x)), {"Re", "Im"}) then
     if not hastype((aa:= Im(x)), {"Re", "Im"}) then
       return(sqrt(a^2 + aa^2))
     end_if
   end_if
 end_if;
 
 if property::hasprop(x) then
    // try is
    if is(x >= 0) = TRUE then
      x
    elif is(x <= 0) = TRUE then
      -x
    end_if;
 end_if;

 // try whether sign can handle the problem  
 
 t:= sign(x);
 if t = 0 or t = 1 then
   x
 elif t = -1 then
   -x
 else
   procname(x)
 end_if
end_proc,
NIL,
table( "type"="abs", "print"="abs",
       "info"="abs(x) -- absolute value of x" )
):

abs(I):= 1: abs(PI):=PI: abs(EULER):=EULER:
abs(CATALAN):= CATALAN:

abs:= slot(abs, "diff",
           proc(x)
             local i;
           begin
             sign(op(x,1)) * diff(op(x,1), args(i) $ i=2..args(0))
           end_proc):

abs:= slot(abs,"sign",
           proc(X)
           begin
             case sign(op(X))
               of 0 do
                 return( 0 )
               of -1 do
               of 1 do
                 return( 1 )
               otherwise
                 return( hold(sign)(X) )
             end_case
           end_proc):


abs:= slot(abs,"rectform",
    loadproc(slot(abs,"rectform"),pathname("STDLIB","RECTFORM"),"abs")
):

abs := slot( abs, "Re", hold(abs) ):
abs := slot( abs, "Im", 0 ):

abs:= slot(abs, "series",
    loadproc(slot(abs,"series"),pathname("SERIES"),"abs")
):

abs := slot( abs, "expand", proc() begin
   if type(args(1))="_mult" then map(args(1),abs)
   else abs(args(1)) end_if
end_proc):

abs := slot( abs, "float", specfunc::abs@float ):
abs := slot(abs, "hull", DOM_INTERVAL::abs@hull):
abs:= slot(abs, "realDiscont", {}):
abs:= slot(abs, "undefined", {}):

// minimal property
abs := slot(abs, "minprop", Type::NonNegative):

abs:= slot(abs, "conjugate",
    loadproc(slot(abs,"conjugate"),pathname("STDLIB","CONJ"),"abs")
):

abs:= slot(abs, "simplify",
    loadproc(slot(abs,"simplify"),pathname("STDLIB","SIMPLIFY"),"abs")
):

// end of file 
