binomial := proc(n, k)

   option noDebug;
   local b, i, nk1, tk, tn, fnk, nk;
   name binomial;

begin
    if args(0) <> 2 then error("2 arguments expected") end_if;

    if k = 0 or k = n then    // (1) 
       if domtype(k) = DOM_FLOAT or domtype(n) = DOM_FLOAT then
            return( float(1) )
       else return( 1 )
       end_if;
    end_if;
    if k = 1 or k = n - 1 then
       if domtype(k) = DOM_FLOAT then
            return( float(n) )
       else return( n )
       end_if;
    end_if;

    // try to convert everything to floats if
    // one parameter is a float
    if domtype(k) = DOM_FLOAT then n:= float(n): end_if;
    if domtype(n) = DOM_FLOAT then k:= float(k): end_if;
    if domtype(k) = DOM_COMPLEX and
      (domtype(op(k,1)) = DOM_FLOAT or 
       domtype(op(k,2)) = DOM_FLOAT) then
       n:= float(n):
    end_if;

    if not testtype(n, Type::Numeric) or
       not testtype(k, Type::Numeric) 
    then
        if not testtype(n,Type::Arithmetical) or
           not testtype(k,Type::Arithmetical) 
           then error("arguments must be of type 'Type::Arithmetical'")
           else return( procname(n, k) )
        end_if
    end_if;

    if domtype(n) = DOM_COMPLEX and
      (domtype(op(n,1)) = DOM_FLOAT or 
       domtype(op(n,2)) = DOM_FLOAT) then
       k:= float(k):
    end_if;
    if domtype(k) = DOM_INTERVAL then n:= hull(n): end_if;
    if domtype(n) = DOM_INTERVAL then k:= hull(k): end_if;

    case domtype(k)
    of DOM_INT do
       case domtype(n)
         of DOM_INT do
            if k < 0 then
               if 0 <= n then 
                  return( 0 )
               elif n-k < 0 then
                 // n<0 and k<0, and n < k
                  return( 0 )
               else 
                  return( binomial(n, n-k) )
               end_if
            else // k > 0, see (1) 
               if n < 0 then 
                  return( (-1)^k*binomial(k-n-1, k) )
               elif n-k < k then 
                  return( binomial(n, n-k) )
               end_if
            end_if;
            break;
         of DOM_RAT      do
         of DOM_FLOAT    do
         of DOM_COMPLEX  do
            if k < 0 then 
               return( 0 )
            end_if;
            break;
       end_case;

       // the special cases k = 0 and k = 1 were handled above
       nk1:= n - k - 1;
       b:= n*_mult((nk1+i)/i $ i = 2..k);
       return(b);
       break;

    of DOM_RAT do
        case domtype(n)
          of DOM_FLOAT do
             // cannot really happen, because if n is a float,
             // k would have been converted to a float, too
             return( gamma(n+1)/gamma(k+1)/gamma(n-k+1) );
          of DOM_INTERVAL do
             // cannot really happen, because if n is an interval,
             // k would have been converted to an inteval, too
             return(gamma(n+1)/gamma(k+1)/gamma(n-k+1) );
          of DOM_COMPLEX do
             if domtype(op(n,1)) = DOM_FLOAT or
                // cannot really happen, because if n is a complex float,
                // k would have been converted to a float, too
                domtype(op(n,2)) = DOM_FLOAT then
                return( gamma(n+1)/gamma(k+1)/gamma(n-k+1) )
             end_if;
        end_case;
        return(procname(n, k));
        break;

    of DOM_FLOAT do
       case domtype(n)
       of DOM_FLOAT do
            if iszero(k) then return(float(1)) end_if; //(1)
            if iszero(k - 1) then return(n) end_if;

            // the most important case first:
            // -1 < k < n + 1
            if k > - 1 and n - k > - 1 then
               // no problem here, we do not encounter any
               // singularities in gamma. We just have to
               // care of speed:

               // use binomial(n, k) = binomial(n, n - k)
               // to make k as small as possible

               k:= min(k, n - k);

               // now, -1 <= k <= n/2

               if iszero(k) then return(float(1)) end_if; //(1)
               if iszero(k - 1) then return(n) end_if;

               // speed up the case of small 'integer' k by
               // avoiding calls to gamma. In MuPAD 2.1, the
               // break even point is about tk = 100: 
               // Note that frac(k) is more expensive than
               // the equivalent k - trunc(k):
               tk:= trunc(k):
               if iszero(k - tk) and tk < 100 and tk > 1 then
                  nk1:= n - k - 1:
                  b:= n*_mult( (nk1+i)/i $ i = 2..trunc(k));
                  return(b);
               end_if;
               // generic return
               return(gamma(n+1)/gamma(k+1)/gamma(n-k+1));
            end_if;

            // For the remaining cases
            // a) k <= - 1
            // b) n <= - 1
            // c) n - k <= - 1
            // we have to take care of singularities in
            // binomial(n, k) = gamma(n+1)/gamma(k+1)/gamma(n-k+1).
            // binomial must produce limiting values if possible.

            //---------
            // case a)
            //---------
            if k <= -1 then
                 if n - k > - 1 then
                    return(binomial(n, n-k));
                 end_if;
                 // n <= k - 1 <= 0
                 tn:= trunc(n):
                 tk:= trunc(k):
                 if iszero(k - tk) then
                    // k is a negative integer
                    return(float(0));
                 end_if;
                 // k <> negative integer
                 if iszero(n - tn) then
                    error("singularity");
                 end_if;
                 // k <> negative integer and n <> negative integer
                 // Here, we should use frac(x) instead of
                 // x - trunc(x), because the latter might be
                 // nonzero even if frac(x) is zero ;-(
                 fnk:= frac(n - k);
                 if iszero(fnk) then
                    return(float(0));
                 end_if;
                 // generic return
                 return(gamma(n+1)/gamma(k+1)/gamma(n-k+1));
            end_if;
            //---------
            // now, k > -1
            //---------
            // case b)
            //---------
            if n <= -1 then
               //  n <= -1 < k
               // We need to resolve the singularities if n is a negative
               // integer. Use the identity
               //  binomial(n, k) = sin(PI*(n-k))/sin(PI*n)*binomial(k-n-1, k));
               // (consequence of gamma(x)*gamma(1 - x) = PI/sin(PI*x), x < 0)
               tn:= trunc(n):
               tk:= trunc(k):
               // Here, we should use frac(x) instead of
               // x - trunc(x), because the latter might be
               // nonzero even if frac(x) is zero ;-(
               fnk:= frac(n - k);
               if iszero(n - tn) and not iszero(fnk) then
                  error("singularity");
               end_if;
               if not iszero(n - tn) and iszero(fnk) then
                  return(float(0));
               end_if;
               if iszero(n - tn) and iszero(k - tk) then
                  // n is a negative integer, k is a positive integer
                  // The following is the limit for fixed k and
                  // n tending towards the negative integer:
                  return((-1)^tk*binomial(k-n-1, k));
               end_if;
               // Here, n <= -1 and k > -1 and n and k cannot be both integers.
               // Can either n or n - k be a negative integer?
               // At this point, the following cases are ruled out by the code above:
               // i)   n is a negative integer --> n-k must be an integer
               //                               and k cannot be an integer
               //      This is not possible.
               // ii) n is not a negative integer --> n - k is not an integer
               //      In this case, neither n nor n - k nor k > -1 hits a
               //      singularity in gamma.
               // So, we may use the generic return:
               return(gamma(n+1)/gamma(k+1)/gamma(n-k+1));
               // Altenatively:
               // return(sin(float(PI)*(n-k))/sin(float(PI)*n)*binomial(k-n-1, k));
            end_if;
            //---------
            // now,  k > -1 and n > -1
            //---------
            // case b)
            //---------
            nk:= n - k;
            if nk <= - 1 then
               if iszero(frac(nk)) then
                  // gamma(n - k + 1) is singular, but gamma(n + 1)
                  // and gamma(k + 1) are finite --> the result is 0
                  return(float(0));
               end_if; 
               // generic return
               return(gamma(n+1)/gamma(k+1)/gamma(nk+1));
            end_if;
            // generic return
            return(gamma(n+1)/gamma(k+1)/gamma(n-k+1));
      end_case;
      // n is symbolic
      return(procname(n, k));

    of DOM_COMPLEX do
        if domtype(op(k,1)) = DOM_FLOAT or 
           domtype(op(k,2)) = DOM_FLOAT or
           domtype(op(n,1)) = DOM_FLOAT or 
           domtype(op(n,2)) = DOM_FLOAT then
           // both n and k are (complex) floats, because
           // 'the other guy' was converted above 
           gamma(n+1)/gamma(k+1)/gamma(n-k+1)
        else
           procname(n, k);
        end_if;
        break;
   end_case
end_proc:

binomial := funcenv(binomial):
binomial := slot(binomial, "type", "binomial"):
binomial := slot(binomial, "print", "binomial"):

binomial := slot(binomial, "float", 
                 proc(n, k) begin binomial(float(n), float(k)) end_proc
                    ):

binomial := slot(binomial, "hull", 
                 proc(n, k) begin binomial(hull(n), hull(k)) end_proc
                    ):

binomial := slot(binomial,"expand",
   proc(n,k) 
   begin 
      if testtype(k,Type::NegInt) or 
         testtype(n-k,Type::NegInt) 
      then 0
  //  elif contains(map({k,n-k,n},type),DOM_FLOAT) then
  //       expand(gamma(n+1)/gamma(k+1)/gamma(n-k))
  //  else expand(fact(n)/fact(k)/fact(n-k))         // do not use fact
      else expand(gamma(n+1)/gamma(k+1)/gamma(n-k+1))
      end_if
   end_proc):

binomial := slot(binomial,"simplify",
   proc(e) 
      local n, k;
   begin 
      n:= op(e,1); k:= op(e,2);

      if testtype(k,Type::NegInt) or testtype(n-k,Type::NegInt) then 0
      else e
      end_if
   end_proc):

// series(binomial(f1, f2), x, n, dir)
binomial := slot(binomial,"series",proc(f1, f2, x, n, dir) begin 
   combine(Series::series(gamma(f1+1)/gamma(f2+1)/gamma(f1-f2+1), x, n, dir))
end_proc):

// diff(binomial(f1, f2), x)
binomial::diff := proc(f, x)
  local f1, f2;
begin
  [f1, f2] := [op(f)];
  (diff(f1, x)*psi(f1 + 1) - diff(f2, x)*psi(f2 + 1)
   - diff(f1 - f2, x)*psi(f1 - f2 + 1))*f
end_proc:
