// $Date: 2001/05/29 15:13:49 $  $Author: tonner $  $Revision: 1.25 $ 
// kg, 11.05.93 

/*++
info -- print short information

info(d)
info(f)

d - domain
f - function

info(d) prints a short information about the domain d. d must have a
method 'info' or an text attribute 'info'. Either the method 'd::info'
is executed or the text 'info' is printed. If the domain has the attibute
'interface', which must be a set of interface names, these names are printed
also.

info(f) prints a short information about the function f. f must have
an string attribute "info". The string is printed unquoted.
++*/

info:= proc(dat) begin
    if args(0) = 0 then 
        stdlib::infoLibraries()     
    elif args(0) = 1 then 
        if slot(domtype(dat), "printInfo") = FAIL then
           print(NoNL, "" . expr2text(dat) . " -- of domain type '".
                       expr2text(domtype(dat))."'\n")
        else
           slot(domtype(dat), "printInfo")(dat)
        end_if
    else
        error("wrong no of args") 
    end_if
end_proc:

stdlib::infoLibraries:=
  proc()
  begin
    print(NoNL, "-- Libraries:\n");
    output::tableForm(stdlib::LIBRARIES, ", ", Unquoted);
    null()
end_proc:

DOM_DOMAIN::printInfo:= proc(dm)
    local dom_name, tab, expo;
begin
    dom_name:= (if dm::Name = FAIL then dm::key else dm::Name end_if);
    if dom_name = FAIL then dom_name:="<<unknown>>" end_if;
    if domtype(dom_name) <> DOM_STRING then
       dom_name := expr2text(dom_name);
    end_if;

    if dm::info = FAIL then
      print(NoNL, "".dom_name." -- a domain\n");
      return()
    end_if;

    if domtype(dm::info) = DOM_STRING then
        print(NoNL, dm::info);
        print(NoNL, "\n")
    else
        dm::info();
    end_if;

    tab:= dm::interface;
    if tab = FAIL then
        return()
    elif not testtype(tab, Type::SetOf(DOM_IDENT)) then
        error("in attribute \"interface\"")
    end_if;

    expo:= dm::exported;
    if domtype(expo) <> DOM_SET then expo:= {} end_if;
    if domtype(dm::_exported) = DOM_SET then
      expo:= expo union dm::_exported
    end_if;

    if nops(tab) > 0 then
      tab:= map(tab, f->if not contains(expo, f) then dom_name."::".expr2text(f) else null() end_if);
      if nops(tab) > 0 then
        print(NoNL, "\n-- Interface:\n");
        output::tableForm(tab, ", ", Unquoted)
      end_if
    else
        print(NoNL, "No Interface.\n");
    end_if;

    if nops(expo) <> 0 then
        print(NoNL, "\n-- Exported:\n");
        output::tableForm(expo, ", ", Unquoted);
    end_if;

    null()
end_proc:

DOM_FUNC_ENV::printInfo:=
  proc(f)
    local noInfo, op0, procName;
  begin
    noInfo := TRUE:
    if slot(f, "info") <> FAIL then
      noInfo := FALSE;
      print(NoNL, slot(f, "info")(f));
      print(NoNL, "\n")
    else
      op0 := op(f, 1);
      if domtype(op0) = DOM_EXEC then
        noInfo := FALSE;
        if nops(op0) = 5 then
          print(NoNL, "".expr2text(f)." -- a module function of module '".
                expr2text(op(op0,5))."'\n")
        else
          print(NoNL, "".expr2text(f).
                " -- a function of the system kernel".
                if not contains(Pref::noComplete(), expr2text(f)) then
                  " [try ?".op(op0, 3)." for help]"
                else null() end
                ."\n")
        end:
      end;
      
      if domtype(op0) = DOM_PROC then
        noInfo := FALSE;
        procName := op(op0, 6);
        if procName = NIL then
          procName := "anonymous";
        end:
    if domtype(procName) <> DOM_STRING then
      procName := expr2text(procName);
    end;
        print(NoNL, "".procName." -- a library procedure [try ?".procName." for help]\n")
      end;
    
      if noInfo then
        print(NoNL, "".expr2text(f)." -- a function environment\n")
      end_if:
    end_if;
     
    null()
  end_proc:

DOM_IDENT::printInfo:=
  proc(ID)
    local noInfo;
  begin
    noInfo := TRUE:
    
    if contains(stdlib::OPTIONS, ID) then
      noInfo := FALSE;
      print(NoNL, "" . ID . " -- an option\n")
    elif protected(ID) <> hold(None) then
      noInfo := FALSE;
      print(NoNL, "" . ID . " -- protected identifier\n")
    end:
    
    if properties(ID) <> FAIL then
      noInfo := FALSE;
      print(NoNL, "" . ID . " -- identifier with property '".expr2text(getprop(ID))."'\n")
    end:
    
    if noInfo then
      print(NoNL, "".ID." -- an unprotected identifier without properties\n")
    end_if:
    null()
  end_proc:

DOM_EXPR::printInfo:=
  proc(ID)
  begin
    print(NoNL, "" . expr2text(ID) . " -- an expression of type ".
          expr2text(type(ID))."\n");
    null()
  end_proc:

DOM_PROC::printInfo:=
  proc(ID)
    local procName;
  begin
    procName := op(ID, 6);
    if procName = NIL then
      procName := "anonymous";
    end:
    if domtype(procName) <> DOM_STRING then
      procName := expr2text(procName);
    end;
    print(NoNL, "" . procName . " -- a procedure of domain type 'DOM_PROC'\n");
    null()
  end_proc:

// end of file 
