// $Date: 2002/02/27 12:38:41 $ $Author: bij $ $Revision: 1.133 $ 

// piecewise - the domain of piecewise defined functions 


// an object of type piecewise has one operand:

// 1 - a list of branches


// the value of the piecewise defined function is the first expression
// the condition of which is satisfied
// the expression may be any object; however, the user is responsible
// not to use any operations that do not make sense







// branch - the domain of branches
// each branch has two or more operands
// 1. a condition
// 2...n  an expression (flattening of expression sequences
//                       cannot be prevented, hence n>=3 may happen)





alias(BRANCH=stdlib::branch):

BRANCH:=newDomain("stdlib::branch"):
BRANCH::create_dom:=hold(stdlib::branch):
BRANCH::info:="branch - the domain of branches":

BRANCH::testtype:= (x, T) -> bool(type(x)=T):


BRANCH::new:=
proc(cond)
  local expression;
begin
  if args(0) >= 2 then
    expression:= args(2..args(0));
    cond := piecewise::simplifyCondition(cond);
    case cond
      of TRUE do
        return(new(BRANCH, TRUE, expression))
      of FALSE do
        return(new(BRANCH, FALSE, undefined))
      otherwise
        return(new(BRANCH, cond, expression))
    end_case;
  end_if;  

  if args(0)=0 then
    return(new(BRANCH, FALSE, undefined))
  end_if;
  
  if args(0)<>1 then
    error("Wrong number of arguments")
  end_if;
  
  
  BRANCH::convert(cond);
  if %=FAIL then
    error("Cannot convert to type branch")
  else
    %
  end_if
end_proc:

BRANCH::convert:=
proc(x)
begin
  if args(0)=0 then
    return(new(BRANCH, FALSE, undefined))
  end_if;
  
  case type(x)
    of BRANCH do
      return(x)                       
    of piecewise do
      if nops(x)<>1 then
        return(FAIL)
      else  
        return(extop(x,1))
      end_if
    of DOM_LIST do
      if nops(x)<2 then
        return(FAIL)
      else  
        return(BRANCH::new(op(x)))
      end_if;
    otherwise
      FAIL
  end_case
end_proc:


BRANCH::condition:= br -> extop(br, 1);

BRANCH::expression:= br -> extop(br, 2..extnops(br));

BRANCH::apply:= 
proc(br)
  local errlevel, result;
begin
  
  errlevel:=traperror(
                      (result:=new(BRANCH, extop(br,1),
                                   args(2)(extop(br,2), args(3..args(0)))))
                      );
   

  if errlevel<>0 then
  /*  warning("Leaving out branch ".expr2text(extop(br,1)).
            " which produced an error"); */
    userinfo(1, "Leaving out branch ".expr2text(br).
             " which produced an error");
    return(new(BRANCH, FALSE, undefined))
  end_if;
  result
end_proc:

BRANCH::applyToCondition:=
proc(br)
begin
  new(BRANCH, piecewise::simplifyCondition
      (args(2)(extop(br,1), args(3..args(0)))),
      BRANCH::expression(br))
end_proc:


BRANCH::has:= (br, obj) -> has([extop(br)], obj):

BRANCH::print:=
proc(br)
  local L1, L2, len;
begin
  L1 := expr2text(BRANCH::expression(br));
  L2 := expr2text(BRANCH::condition(br));
  len := length(L1);
  stringlib::format(L1, len)."  if ".
           L2."\n";
end_proc:

piecewise:=newDomain("piecewise"):
piecewise::create_dom:=hold(piecewise):

piecewise::info:= "piecewise -- the domain of piecewise defined functions":

piecewise::interface:={
                       hold(branch),
                       hold(condition),
                       hold(expression),
                       hold(extmap),
                       hold(insert),
                       hold(mapConditions),
                       hold(numberOfBranches),
                       hold(remove),
                       hold(restrict),
                       hold(selectConditions),
                       hold(solveConditions),
                       hold(splitBranch),
                       hold(splitConditions)
                       }:


piecewise::new:=
proc()
  option hold;
  local i, j, result, cond, newbranch, argument, expression,
  beenthere: DOM_TABLE;
  
begin
  if args(0) = 0 then
    return(undefined)
  end_if;

  if args(0)<>1 then
    // convert each argument to branch
    result:=[]; // list of all branches created so far
    beenthere:= table(); // table of all expressions obtained so far
                         // beenthere[object]=i means that object is in the
                         // i-th branch
    for i from 1 to args(0) do
      // errlev:=traperror((argument:= context(args(i))));
      // argument:= context(args(i));

      if type(context(hold(val)(args(i)))) = DOM_LIST then
        cond:= (context(hold(val)(args(i))))[1];
        case (cond:=piecewise::simplifyCondition(context(cond)))
          of TRUE do
            return(context(args(i))[2])
          of FALSE do
            break
          otherwise
            expression:= context(args(i))[2];
            if type((j:= beenthere[expression])) = "_index" then
              // we have never met that object
              newbranch:= new(BRANCH, cond, expression);
              result:=result.[piecewise::flattenBranch(newbranch)];
              beenthere[expression] := nops(result)
            else
              assert(type(j) = DOM_INT);
              // we have already a branch with that object
              // hence we just have to combine the conditions with _or
              result[j]:= BRANCH::applyToCondition(result[j], _or, cond);
              if op(result[j], 1)=TRUE then
                return(BRANCH::expression(result[j]))
              end_if;
            end_if;
        end_case;
      else
        newbranch:=BRANCH(context(args(i)));
        result:=result.[piecewise::flattenBranch(newbranch)]
      end_if;
        
    end_for;
    if nops(result)=0 then
      return(undefined)
    else  
      return(new(piecewise, op(result)))
    end_if;
  end_if;

  // errlev:=traperror((argument:=context(args(1))));
  argument:=context(args(1));

  /*
  if errlev<>0 then
    return(undefined)
  end_if;
  */

  if type(argument)="_exprseq" then
    // force flattening
    return(eval(hold(piecewise::new)(argument)))
  end_if;
  
  piecewise::convert(argument);
  if %=FAIL then
    error("Cannot convert to type piecewise")
  else
    // piecewise::flatten(%)
    %
  end_if
end_proc:

piecewise::convert:=
proc(x)
  local br:BRANCH;
begin
  if args(0)=0 then
    error("Wrong number of arguments")
  end_if;
  
  case type(x)
    of piecewise do
      return(x)
    of BRANCH do
      br:=x;
      break
    otherwise
      br:=BRANCH(x);
  end_case;
  // here br is of type branch
  case BRANCH::condition(br)
    of TRUE do
      return(BRANCH::expression(br))
    of FALSE do
      return(undefined)
    otherwise
      return(new(piecewise, piecewise::flattenBranch(br)))
      //return(new(piecewise, br))
  end_case;
end_proc:

piecewise::convert_to:=
proc(x:piecewise, T:DOM_DOMAIN)

begin
  case T
    of piecewise do
      return(x)
    of DOM_LIST do
      return([extop(x)])
    otherwise
      FAIL
  end_case
end_proc:

      
piecewise::testtype:=
proc(x, T)
  local i;
begin
  if T=piecewise then
    return(TRUE)
  end_if;

  // test whether every object is of the correct type
  for i from 1 to extnops(x) do
    if testtype(piecewise::expression(x, i), T) = FALSE then
      return(FALSE)
    end_if
  end_for;
  TRUE
end_proc:

piecewise::branch:=
proc(pw:piecewise, n:Type::PosInt)
begin
  if n > extnops(pw) then
    FAIL
  else
    [extop(extop(pw, n))]
  end_if
end_proc:

piecewise::numberOfBranches:= extnops:

piecewise::condition:=
proc(pw:piecewise, n:DOM_INT)
// returns the n-th condition 
begin
  BRANCH::condition(extop(pw,n))
end_proc:

piecewise::expression:=
proc(pw:piecewise, n:DOM_INT)
// returns the expression that is valid if the n-th condition holds 
begin
  BRANCH::expression(extop(pw,n))
end_proc:



// flattenBranch
//
// input: a branch [cond, pw]
// 
// output: the input if pw is not piecewise
//         a sequence of branches [cond and c1, xpr1], ...,[cond and cn, xprn]
//                            if pw = piecewise([c1, xpr1],...,[cn, xprn])
// 


piecewise::flattenBranch:=
proc(branch)
  local pw;
begin
  if type((pw := BRANCH::expression(branch))) <> piecewise then
    return(branch)
  end_if;

  assert(type(pw) = piecewise);
  extop(piecewise::mapConditions(pw, _and, BRANCH::condition(branch)))  
end_proc:


piecewise::removePiecewiseFromCondition:=
proc(cond)
  local l, r;
begin
  case type(cond)
    of "_and" do
    of "_or" do
    of "_not" do
      return(map(cond, piecewise::removePiecewiseFromCondition))
    of "_less" do
    of "_leequal" do
    of "_equal" do
    of "_unequal" do
      [l, r]:= [op(cond)];
      if type(l)= piecewise then
        if type(r)=piecewise then
          _or(((dom::condition(l,i) and dom::condition(r, j) and
                op(cond, 0)(dom::expression(l, i), dom::expression(r, j)))
               $i=1..extnops(l)) $j=1..extnops(r))
        else
          _or((dom::condition(l,i) and op(cond, 0)(dom::expression(l, i), r))
              $i=1..extnops(l))
        end_if
      else
        if type(r)=piecewise then
          _or((dom::condition(r, i) and op(cond, 0)(l, dom::expression(r, i)))
              $i=1..extnops(r))
        else
          cond
        end_if
      end_if;
      break
    otherwise
      return(cond)
  end_case
 
end_proc:

// piecewise::simplifyCondition
// a replacement for property::simpex that also uses the solver


piecewise::simplifyCondition:=
proc(cond)
  local
  newcond: Type::Boolean, S, i, inds,
  NoChange: DOM_BOOL,
  collectIn: DOM_PROC,
  extracted: Type::Boolean;

begin

  // local method collectIn
  // tries to use set operations for simplifying
  collectIn:=
  proc(cond)
    local
    operands: DOM_LIST,
    T: DOM_TABLE,
    In, Other, Dummy, j;
  begin
    T:= table();
    case type(cond)
      of "_and" do
        operands:= [op(cond)];
        [In, Other, Dummy]:= split(operands, testtype, "_in");
        assert(Dummy = []);
        for j in In do
          if contains(T, op(j, 1)) then
            T[op(j, 1)]:= T[op(j,1)] intersect op(j, 2)
          else
            T[op(j, 1)]:= op(j, 2)
          end_if
        end_for;
        return(_and(op(eq,1) in op(eq,2) $eq in T) and op(Other))
      of "_or" do
        operands:= [op(cond)];
        [In, Other, Dummy]:= split(operands, testtype, "_in");
        assert(Dummy = []);
        for j in In do
          if contains(T, op(j, 1)) then
            T[op(j, 1)]:= T[op(j,1)] union op(j, 2)
          else
            T[op(j, 1)]:= op(j, 2)
          end_if
        end_for;
        return(_or(op(eq,1) in op(eq,2) $eq in T) or op(Other))
    end_case;
    cond
  end_proc;


  // main program of simplifyCondition
  
  cond:= piecewise::removePiecewiseFromCondition(cond);

//  cond:= property::simpex(cond);
  
  case type(cond)
    of DOM_BOOL do
      return(cond)
    of "_not" do
      cond:= map(cond, piecewise::simplifyCondition);
      if type(cond) = "_not" and type(op(cond, 1)) = "_in" then
        // turn not a in S into a in C_ minus S
        return(op(cond, [1,1]) in C_ minus op(cond, [1, 2]) )
      else
        return(cond)
      end_if
    of "_and" do
      newcond:= map(cond, piecewise::simplifyCondition);
      if type(newcond) <> "_and" then
        return(newcond)
      end_if;
      // substitute constants
      extracted:= TRUE;
      repeat
        NoChange:= TRUE;
        for i from 1 to nops(newcond) do
          S:= op(newcond, i);
          if testtype(S, DOM_IDENT = Type::Constant) then
            extracted:= extracted and S;
            if traperror((newcond:=
                          eval(subs(_and(op(newcond, k) $k=1..i-1,
                                    op(newcond, k) $k=i+1..nops(newcond)),
                               op(S, 1) = op(S, 2))))) <> 0 then
              return(FALSE)
            end_if;
            NoChange:= FALSE;
            break
          end_if
        end_for
      until NoChange or type(newcond) <> "_and" end_repeat;
      collectIn(newcond);
      if length(%) >= 30 then
        return(property::simpex(extracted, hold(Unsimplified)) and
               collectIn(newcond))
      else
        return(property::simpex(extracted, hold(Unsimplified)) and
               property::simpex(collectIn(newcond), hold(Unsimplified)))
      end_if
    of "_or" do
      cond:= map(cond, piecewise::simplifyCondition);
      collectIn(cond);
      if length(%) <= 40 then
        return(property::simpex(%, hold(Unsimplified)))
      else
        return(%)
      end_if;
    of "_xor" do
    of "_implies"do
    of "_equiv" do
      return(map(cond, piecewise::simplifyCondition))
    of "_in" do
      return(property::simpex(cond, hold(Unsimplified)))
    of "_equal" do
      inds:= indets(cond) minus Type::ConstantIdents;
      if nops(inds) = 1 then
        S:=solve(cond, op(inds,1), DontRewriteBySystem);
        if S=C_ then
          return(TRUE)
        end_if;
        if hastype(S, "solve") then
          return(property::simpex(cond, hold(Unsimplified)))
        end_if;
        if type(S)="_minus" and op(S,1)=C_ then
          return(not property::simpex(op(inds,1) in op(S,2), hold(Unsimplified)))
        end_if;
        return(property::simpex(op(inds, 1) in S, hold(Unsimplified)))
      end_if;
      return(property::simpex(cond, hold(Unsimplified)))
    of "_unequal" do
      return(not piecewise::simplifyCondition(op(cond, 1) = op(cond, 2)))
    of "_leequal" do
    of "_less" do
      // only solve linear inequalities in one variable
      inds:= indets(cond) minus Type::ConstantIdents;
      if nops(inds) = 1 and
        map({op(cond)}, testtype, Type::PolyExpr([op(inds, 1)])) = {TRUE} and
        map({op(cond)}, degree, [op(inds, 1)]) minus {0, 1} = {} then
        S:=solve(cond, op(inds,1));
        return(property::simpex(op(inds, 1) in S, hold(Unsimplified)))
      elif length(cond) <= 40 then
        return(property::simpex(cond, hold(Unsimplified)))
      else // do not simplify expressions that are too complicated
           // for simpex
        return(cond)
      end_if
  end_case;

  // default:
  property::simpex(cond, hold(Unsimplified))
end_proc:
  


// cannot map since nops must return a number:
piecewise::nops:= piecewise::extnops:


// op - map the function op to all expressions

piecewise::op:=
proc(pw:piecewise)
begin
  piecewise::extmap(pw, op, args(2..args(0)))
end_proc:


// _index - map the function _index to all expressions
piecewise::_index:=
proc(pw:piecewise)
begin
  piecewise::extmap(pw, _index, args(2..args(0)))
end_proc:



/*++

	extmap - application of a function to all expressions

++*/


piecewise::extmap:=
proc(pw)
begin
  if pw=undefined then
	return(undefined)
  end_if;
  // piecewise::flatten
     piecewise::new(
     BRANCH::apply(extop(pw, i), args(2..args(0))) $i=1..extnops(pw))
end_proc:

/*

   map - map a "map" command to all branches
            but simply map to objects that are not piecewise

*/


piecewise::map:=
proc(pw)
begin
  if type(pw)<>piecewise then
    map(pw, args(2..args(0)))
  else
    piecewise::extmap(pw, map, args(2..args(0)))
  end_if
end_proc:




/*++

	mapConditions - application of a function to all
			conditions

++*/

piecewise::mapConditions:=
proc(pw)
  local i, result;
begin
  result:=null();
  for i from 1 to extnops(pw) do
    BRANCH::applyToCondition(extop(pw,i), args(2..args(0)));
    case BRANCH::condition(%)
      of TRUE do
        return(BRANCH::expression(%))
      of FALSE do
        break;
      otherwise
        result:=result, %
    end_case;
  end_for;
  if nops(result)=0 then
    return(undefined)
  else
    return(new(piecewise, result))
  end_if;
end_proc:


/* piecewise::selectConditions(pw, f)

keep only those conditions for which f(condition)=TRUE

*/

piecewise::selectConditions:=
proc(pw,f)
local argv;
begin
  argv:=args();
  select([extop(pw)], x-> f(op(x,1), op(argv, 3..nops(argv))));
  if nops(%)=0 then
    undefined
  else
    piecewise::new(op(%))
  end_if;
end_proc:



piecewise::selectExpressions:=
proc(pw, f)
  local argv;
begin
  argv:=args();
  select([extop(pw)], x-> f(op(x, 2), op(argv, 3..nops(argv))));
  if nops(%)=0 then
    undefined
  else
    piecewise::new(op(%))
  end_if
end_proc:


/* piecewise::splitConditions(pw, f)

split pw into a list of three piecewise objects depending on whether
f(condition) gives TRUE, FALSE, or UNKNOWN

*/

piecewise::splitConditions:=
proc(pw, f)
local argv;
begin
  argv:=args();
  split([extop(pw)], x-> f(op(x,1), op(argv, 3..nops(argv))));
  map(%, u-> if nops(u)>0 then
               piecewise::new(op(u))
             else
               undefined
             end_if)
end_proc:
  
/*++ 
	zip - argumentwise application of a law of composition 
	for two or more piecewise defined functions
++*/

piecewise::zip:=
proc(pwf, pwg, f)
  local i,j,result,cond,g,conds_so_far;
begin
  if type(pwf)<>piecewise then
    g:= x -> f(pwf, x);
    return(piecewise::extmap(pwg,g))
  elif type(pwg)<>piecewise then
    return(piecewise::extmap(pwf, f, pwg))
  end_if;
  result:=[];
  conds_so_far:=[];
  for i from 1 to extnops(pwf) do
    for j from 1 to extnops(pwg) do
      cond:=piecewise::simplifyCondition(piecewise::condition(pwf,i) and
      piecewise::condition(pwg,j));
      // possibility check 
      case cond
        of TRUE do
          return(f(piecewise::expression(pwf,i),
                          piecewise::expression(pwg,j)))
        of FALSE do
          break
        otherwise
          if contains(conds_so_far, cond)=0 then
            result:=append(result,
                           BRANCH(cond,
                            f(piecewise::expression(pwf,i),
                              piecewise::expression(pwg,j))));
            conds_so_far:=append(conds_so_far, cond)
          end_if
      end_case;
	end_for
  end_for;
  if nops(result)= 0 then
    undefined
  else
    piecewise::new(op(result))
  end_if
end_proc:


piecewise::evaluate:=
proc(pw)
  local i, result;
begin
  result:=null();
  // pw:=piecewise::flatten(pw);
  for i from 1 to extnops(pw) do
    piecewise::simplifyCondition(eval(piecewise::condition(pw,i)));
    case %
      of TRUE do
        return(eval(piecewise::expression(pw, i)))
      of FALSE do
        break
      otherwise
        result:=result, BRANCH(%, eval(piecewise::expression(pw, i)))
    end_case;
  end_for;
  if nops(result)=0 then
    return(undefined)
  else  
    new(piecewise, result)
  end_if
end_proc:


piecewise::subs:=
proc(pw:piecewise)
begin  
  piecewise::new(subs(extop(pw), args(2..args(0))))
end_proc:
                   
piecewise::insert:=
proc(pwf, l, no: Type::PosInt)

begin
  if no > extnops(pwf)+1 then
    error("Illegal position")
  end_if;
  BRANCH::new(l);
  if no = 1 then
    new(piecewise, l, extop(pwf, no..extnops(pwf)))
  else  
    new(piecewise,
        extop(pwf, 1..no-1), l, extop(pwf, no..extnops(pwf)))
  end_if
end_proc:

piecewise::remove:=
proc(pwf, no: Type::PosInt)

begin
if extnops(pwf)<no then
	error("Illegal position")
end_if;
new(piecewise, extop(pwf, i) $i=1..no-1, extop(pwf, i) $i=no+1..extnops(pwf))
end_proc:


// fill cases: produce a nowhere undefined
// piecewise by filling pw with a where it was previously undefined
piecewise::fillCases:=
proc(pw, a)
  local br;
begin
  br:= BRANCH::new([_and(not piecewise::condition(pw, i) $i=1..nops(pw)), a]);
  if BRANCH::condition(br)=FALSE or BRANCH::expression(br)=undefined then
    pw
  else  
    new(piecewise, extop(pw), br)
  end_if
end_proc:


/*

splitBranch(pw, n, cond)

split the nth branch [c, x] into two branches
[c and cond, x], [c and not cond, x]

*/


piecewise::splitBranch:=
proc(pwf: piecewise, no: Type::PosInt, cond): piecewise

begin
  if extnops(pwf) < no then
    error("Not that many branches exist")
  end_if;
  new(piecewise,
      extop(pwf, i) $i=1..no-1,
      [piecewise::condition(pwf, no) and cond, piecewise::expression(pwf, no)],
      [piecewise::condition(pwf, no) and not cond,
       piecewise::expression(pwf, no)], 
      extop(pwf, i) $i=no+1..extnops(pwf) )
end_proc:



/*

disregardPoints: assume every condition of type x=C to be FALSE
and every condition of type x<>C to be TRUE

*/

piecewise::disregardPoints:=
proc(pw)
  local subsCond, result;
begin
  /* subsCond - replace conditions of type _unequal by TRUE
     and conditions of type _equal by FALSE */
  
  subsCond:=
  proc(cond)
    name piecewise::subsCond;
  begin
    case type(cond)
      of "_equal" do
        return(FALSE)
      of "_unequal" do
        return(TRUE)
      of "_or" do 
      of "_not" do
      of "_and" do
        return(map(cond, subsCond))
      otherwise
        cond
    end_case
  end_proc;

  result:=piecewise::mapConditions(pw, subsCond);
  delete subsCond;
  result
end_proc:


// assumeIdentsTRUE(pw, identset) - assume that all conditions in which
// one of the identifiers in identset appears is TRUE


piecewise::assumeIdentsTRUE:=
proc(pw, identset:DOM_SET)
  local subsCond, result;
begin
  subsCond:=
  proc(cond)

  begin
     case type(cond)
      of "_or" do 
      of "_not" do
      of "_and" do
        return(map(cond, subsCond))
      otherwise
        if indets(cond) intersect identset <> {} then
          return(TRUE)
        else
          return(cond)
        end_if
    end_case
  end_proc;

  result:=piecewise::mapConditions(pw, subsCond);  
  delete subsCond;
  result
end_proc:


//  operations on conditions //


// innerOfCondition: restrict condition to interior points 

piecewise::innerOfCondition:=
proc(c)

begin
  case type(c)
    of "_equal" do
      return(FALSE)
    of "_unequal" do
    of "_less" do
      return(c)
    of "_leequal" do
      return(_less(op(c)))
    of "_and" do
    of "_or" do
    of "_not" do
      return(map(c, piecewise::innerOfCondition))
    otherwise
      error("Cannot handle condition of type ".expr2text(type(c)))
  end_case
end_proc:
   
piecewise::borderOfCondition:=
proc(c)
begin
  case type(c)
    of "_equal" do
      return(c)
    of "_unequal" do
    of "_leequal" do
    of "_less" do
      return(_equal(op(c)))
    of "_or" do
      return(map(c, piecewise::borderOfCondition))
    of "_and" do
      return(piecewise::borderOfCondition(op(c,1)) or
             piecewise::borderOfCondition(op(c,2))
             )
    of "_not" do
      return(piecewise::borderOfCondition(op(c,1)))
    otherwise
      error("Cannot handle condition of type ".expr2text(type(c)))
  end_case
end_proc:

piecewise::outerOfCondition:=
proc(c)
begin
  piecewise::innerOfCondition(not c)
end_proc:

piecewise::satisfiesCondition:=
proc(x, value, c)
begin
  is(subs(c, x=value))
end_proc:



//  arithmetical/ set theoretic operations (argumentwise) 

piecewise::_plus:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_plus(args(1), _plus(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(), _plus)
  end_if
end_proc :

piecewise::_mult:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_mult(args(1), _mult(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(),_mult)
  end_if
end_proc :

piecewise::_union:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_union(args(1), _union(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(),_union)
  end_if
end_proc :

piecewise::_intersect:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_intersect(args(1), _intersect(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(),_intersect)
  end_if
end_proc :

piecewise::_negate:=proc() begin piecewise::extmap(args(1), _negate) end_proc :

piecewise::_invert:=proc() begin piecewise::extmap(args(1), _invert) end_proc :

piecewise::_subtract:=proc() begin piecewise::zip(args(),_subtract) end_proc :

piecewise::_divide:=proc() begin piecewise::zip(args(),_divide) end_proc :

piecewise::_power:=proc() begin piecewise::zip(args(),_power) end_proc :

piecewise::_div:=proc() begin piecewise::zip(args(),_div) end_proc :

piecewise::_mod:=proc() begin piecewise::zip(args(),_mod) end_proc :

piecewise::_minus:=proc() begin piecewise::zip(args(),_minus) end_proc:

piecewise::_less:=(x,y) -> piecewise::zip(x,y, _less):

piecewise::_leequal:=(x,y) -> piecewise::zip(x,y,_leequal):



// overload solvelib::Union
/* piecewise::Union(pw, x, xset)
   
   returns the union of all pw(x), x varying over xset

   There are two kinds of branches:
   a) the condition in the branch does not depend on x, i.e. the
      branch is applicable for all or none x \in xset. The Union may be
      computed separately for every such branch

   b) the condition depends on x. We assume that pw is consistent
      (branches agree where they domains of definition overlap).
      Then \bigcup_{x in xset} (S1(x) if cond1(x), S2(x) if cond2(x))
      equals \bigcup{x in {y \in xset; cond1(y)}} S1(x)
       \cup \bigcup{x in {y \in xset; cond2(y)}} S2(x)
*/

piecewise::Union:=
proc(pw:piecewise, x:DOM_IDENT, xset)
  local dep, indep, dummy, iU, dU;
begin

  // simple case
  
  if xset={} then
    return({})
  end_if;
  
  [dep, indep, dummy] := piecewise::splitConditions(pw, has, x);
  assert(dummy = undefined);
  
  // simpler part first: conditions do not depend on the variable
  if indep = undefined then
    iU:= undefined;
  else
    iU:= piecewise::indepUnion(indep, x, xset)
  end_if;

  
  if dep = undefined then
    return(iU)
  elif indep <> undefined then
    dep:= piecewise::mapConditions(dep, _and,
                                   _and (not piecewise::condition(indep, i)
                                         $i=1..nops(indep)))
  end_if;

  assert(type(dep) = piecewise);
  userinfo(20, expr2text(extnops(dep))."conditions depend on variable");
  dU:= piecewise::depUnion(dep, x, xset);
  
  if iU = undefined then
    dU
  else
    iU:= piecewise::fillCases(iU, {}):
    iU union dU
  end_if;
end_proc:

piecewise::indepUnion:= (pw, x, xset) -> piecewise::extmap(pw, solvelib::Union, x, xset):

piecewise::depUnion:=
proc(pw: piecewise, x, xset)
  local mysolve;
begin

  mysolve:=
  proc(cond)
  begin
    /* save x;
    assume(x in xset);
    cond:= eval(cond);
    unassume(x); */
 
    solve(cond, x) intersect xset
  end_proc;
  
  // first step: delete empty sets
  pw:= piecewise::selectExpressions(pw, _unequal, {});
  if pw=undefined then
    return({})
  end_if;
  
  // Todo: be careful with determining x in {y \in xset; condi(y)}
  // solve condi for y only if it is rather simple !?
  
  _union(solvelib::Union
         (piecewise::expression(pw, i), x,
          mysolve(piecewise::condition(pw, i)))
         $i=1..nops(pw))
end_proc:


// special functions

piecewise::sin:= x -> piecewise::extmap(x, sin):
piecewise::cos:= x -> piecewise::extmap(x, cos):
piecewise::tan:= x -> piecewise::extmap(x, tan):
piecewise::exp:= x -> piecewise::extmap(x, exp):
piecewise::ln := x -> piecewise::extmap(x, ln):
piecewise::arcsin:= x-> piecewise::extmap(x, arcsin):
piecewise::arccos:= x -> piecewise::extmap(x, arccos):
piecewise::arctan:= x -> piecewise::extmap(x, arctan):
piecewise::sinh := x -> piecewise::extmap(x, sinh):
piecewise::cosh := x-> piecewise::extmap(x, cosh):
piecewise::tanh := x -> piecewise::extmap(x, tanh):
piecewise::coth := x-> piecewise::extmap(x, coth):
piecewise::arcsinh:= x -> piecewise::extmap(x, arcsinh):
piecewise::arccosh:= x -> piecewise::extmap(x, arccosh):
piecewise::arctanh:= x -> piecewise::extmap(x, arctanh):
piecewise::arccoth:= x -> piecewise::extmap(x, arccoth):
piecewise::sech:= x-> piecewise::extmap(x, sech):
piecewise::csch:= x-> piecewise::extmap(x, csch):
piecewise::sign := x -> piecewise::extmap(x, sign):
piecewise::abs  := x-> piecewise::extmap(x, abs):
piecewise::heaviside:= x -> piecewise::extmap(x, heaviside):

piecewise::Re := x -> piecewise::extmap(x, Re):
piecewise::Im := x -> piecewise::extmap(x, Im):

piecewise::numer := x -> piecewise::extmap(x, numer):
piecewise::denom := x -> piecewise::extmap(x, denom):



/*++    
	restrict - impose an additional restriction
		   on the piecewise function
++*/

piecewise::restrict:=
proc(pw, cond)

begin
  if type(pw) <> piecewise then
    piecewise(BRANCH(cond, pw))
  else
    piecewise::mapConditions(pw, _and, cond)
  end_if
end_proc:

piecewise::_concat:=
proc()
  local argv;
begin
  argv:=select([args()], _unequal, undefined);
  // the everywhere undefined function is neutral
  // element w.r.t. concatenation
  if map({op(argv)}, type)<>{piecewise} then
    error("All arguments must be of type piecewise")
  end_if;
  new(piecewise,op(map(argv, extop))) // flatten !
end_proc:

/*++
	simplify - try to decide the truth of the condition and
	remove unnecessary conditions
++*/ 

piecewise::simplify:=
proc(pw: piecewise)
  local i, j, result, cond, xpr;
  
begin
  result:=[];
  // pw:=piecewise::flatten(pw);
  for i from 1 to extnops(pw) do
    // cond:=is(piecewise::condition(pw, i));
    // if cond=UNKNOWN then
    cond:=piecewise::simplifyCondition(piecewise::condition(pw, i));
    // end_if;
    case cond
      of FALSE do
        break; // do not enter this branch into result 
      of TRUE do
        piecewise::expression(pw,i);
        return(eval(%))
      otherwise
        // test whether the same object already exists 
        xpr:=simplify(piecewise::expression(pw, i));
        xpr:=piecewise::subsEqualities(xpr, cond);
        if xpr=undefined then
          break
        end_if;
        j:=contains(map(result, op, 2), xpr);
        if j>0 then
          // combine old and new condition
          cond:=piecewise::simplifyCondition(result[j][1] or piecewise::condition(pw, i));
          if cond=TRUE then
            return(xpr)
          elif cond=FALSE then
            // should not happen, since result[j][1] was not FALSE
            delete result[j]
          else  
            result[j][1]:=cond;
            result[j][2]:=piecewise::subsEqualities(result[j][2], cond)
          end_if
        else  // cond has still its original value! 
          result:=append(result, [cond, xpr])
        end_if;
    end_case;
  end_for;
  if result=[] then
    undefined
  else
    new(piecewise, op(map(result, BRANCH::new)))
  end_if
end_proc:



  
// piecewise::subsEqualities - simplifies a given expression
// taking into account the given condition

piecewise::subsEqualities:=
proc(xpr, cond)
local i, j, result, Lhs, Rhs, newexpr, newcond;
  
begin
  case type(cond)
    of "_equal" do
      if traperror((result:= eval(subs(xpr, cond))))<>0 then
        return(undefined)
      else
        return(result) 
      end_if;
    // case _less etc.: use assume 
    of "_less" do
    of "_leequal" do
      Lhs:= op(cond, 1);
      Rhs:= op(cond, 2);
      if not has(xpr,Lhs) and not has(xpr, Rhs) then
        return(xpr)
      end_if;
      i:=genident(); j:=genident();
      newexpr:=xpr;
      if testtype(Lhs, Type::Numeric)<>TRUE then
        newexpr:=subs(newexpr, Lhs=i);
        newcond:=subsop(cond, 1=i)
      else
        newcond:=cond
      end_if;
      if testtype(Rhs, Type::Numeric)<>TRUE then
        newexpr:=subs(newexpr, Rhs=j);
        newcond:=subsop(newcond, 2=j)
      end_if;
        // substitute a new variable for the left/right hand side
        // in both the expression and the condition
        // then assume the new condition generated in this way
        // to be true, and simplify the newly generated expression  
      
      /* since assume does not evaluate its arguments */
      eval(subs(hold(assume(dummy)), hold(dummy)=newcond));
      eval(subs(simplify(newexpr), i=Lhs, j=Rhs));
      break
    of "_and" do
      result:=xpr;
      for i in cond do
        result:= piecewise::subsEqualities(result, i)
      end_for;
      result;
      break
    otherwise
      xpr
  end_case
end_proc:



// some methods for expression manipulation, realized by mapping

piecewise::expand:= pw -> piecewise::extmap(pw, expand):

piecewise::normal:= pw -> piecewise::extmap(pw, normal):

piecewise::combine:= pw -> piecewise::extmap(pw, combine, args(2..args(0))):

piecewise::rewrite:= pw -> piecewise::extmap(pw, rewrite, args(2..args(0))):

piecewise::factor:= pw -> piecewise::extmap(pw, factor):

piecewise::partfrac:= pw -> piecewise::extmap(pw, partfrac):

/* solve(pw, x, options) - solves the equation pw=0 for x */

piecewise::solve:=
proc(pw:piecewise, x)
  local otherargs;
begin
  otherargs:=args(3..args(0));
  _union(map(extop(pw),
             /* for every branch, fin the set of x that both satisfy the
                given condition and solve the given expression */
             proc(branch)
               name solveBranch;
             begin
               userinfo(30, "Solving case ".
                        expr2text(BRANCH::condition(branch)));
               solve(BRANCH::condition(branch), x) intersect
               solve(BRANCH::expression(branch), x, otherargs)
             end_proc
             ))
end_proc:


/* piecewise::solveConditions

   solve all conditions for the given variable x and reconvert
   the sets of solutions into expressions
*/

piecewise::solveConditions:=
proc(pw:piecewise, x:DOM_IDENT)
begin
  piecewise::mapConditions(pw, cond -> (x in solve(cond,x)))
end_proc:

/*
	discont - determine the discontinuities (the points for which
	          the function is defined on both sides of the point,
		  but the limits from the left/right are not the same
		  or do not agree with the function value at the point
*/

piecewise::discont:=
proc(pw:piecewise, x, F=null())
  local discontlist:DOM_SET,
  i:DOM_INT,
  j:DOM_INT,
  cond, disc;

begin
  if args(0)<2 or args(0)>3 then
    error("Two arguments expected")
  end_if;

  if type(x)="_equal" then
    if type(op(x, 2))<> "_range" then
      error("Right hand side of second argument must be range")
    else  
      return(piecewise::discont(pw, op(x, 1), F)
             intersect Dom::Interval(op(x, 2)))
    end_if
  elif type(x) <> DOM_IDENT then
    error("Second argument must be identifier or equation")
  end_if;
  
  discontlist:={};
  for i from 1 to extnops(pw) do
    cond:=piecewise::condition(pw, i);
    disc:= solve(piecewise::borderOfCondition(cond, x),x);
    userinfo(10, "Found border points ".expr2text(disc));
    discontlist:= discontlist union
                  select(discont(piecewise::expression(pw, i), x, F),
                         u -> piecewise::satisfiesCondition( x, u, cond)=TRUE );
    for j from 1 to extnops(pw) do
      if j<>i then
        userinfo(10, "Checking whether border points are inner points of ".
                 output::ordinal(j)." condition");
        disc:=select(disc, u -> not piecewise::satisfiesCondition
                     (x, u, piecewise::innerOfCondition
                      (piecewise::condition(pw,j))) )
      end_if;
    end_for;
    discontlist:= discontlist union disc;
  end_for;
  userinfo(5, "Discontinuities are ".expr2text(discontlist));
  discontlist
end_proc:

piecewise::indefint :=
proc(pw:piecewise, x:DOM_IDENT)

begin
piecewise::extmap(pw, int, x)
end_proc:

piecewise::defint :=
proc(pw:piecewise, eq: "_equal")
  local x, a,b, i: DOM_INT, S, J, cond, res,
  newpw: piecewise, // the rest which cannot be integrated
  covered,         // the subset of [a,b] already covered;
  integral,        // int_{ x \in covered} pw(x) dx
  defint_b_gr_a: DOM_PROC; // local procedure
  
begin
 

  // local method
  // performs the operation assuming in addition that a <= b
  
  defint_b_gr_a:=
  proc(a, b)
    begin
  
      J:= Dom::Interval([a,b]);
      // we just have to compute intOverSet(pw, x, [a, b])
      // since a <= b

      
      // and the subset already integrated is empty
      // the integral is zero

      covered:={};
      integral:=0;
      newpw:=[];
  
      for i from 1 to extnops(pw) do
        S:=solve((cond:=piecewise::condition(pw,i)), x) intersect
            (J minus covered);
        userinfo(20, "Condition ".expr2text(cond)." satisfied over ".
                 expr2text(S));
        

        res:=intlib::intOverSet(piecewise::expression(pw, i), x, S);
        
        if type(res)<> "intOverSet" then
          userinfo(20, "Integral over ".expr2text(S)." is ".
                   expr2text(res));
          covered:=covered union S;
          integral:=integral+ res;
        else
          newpw:=append(newpw, extop(pw, i))
        end_if;
        
        if type(J minus covered)=DOM_SET then
          return(integral)
        end_if;
      end_for;
    if extnops(newpw)>0 then
      newpw:=new(piecewise, extop(newpw));
      integral + hold(intlib::intOverSet)(newpw, x, J minus covered)
    else
      // parts of [a,b] were not covered because func is not defined there
      integral
    end_if;

  end_proc;


  // end of local method


  x:=op(eq,1);
  if type(x) <> DOM_IDENT then
    error("First argument must be identifier")
  end_if;
  if type(op(eq,2)) <> "_range" then
    error("Right hand side of second argument must be range")
  end_if;
  a:=op(eq,[2,1]);
  b:=op(eq,[2,2]);

  if a=b then
    return(0)
  end_if;

  return(piecewise([a <= b, defint_b_gr_a(a, b)],
                   [b <= a, -defint_b_gr_a(b, a)]))
  
  
end_proc:

piecewise::int :=
proc(pw:piecewise, x)
begin
  if type(x)="_equal" then
    piecewise::defint(pw, x)
  elif type(x)=DOM_IDENT then
    piecewise::indefint(pw, x)
  end_if
end_proc:

piecewise::intOverSet :=
proc(xpr, x, pw:piecewise)

begin
  // the conditions of pw must *not* depend on x
  piecewise::extmap(pw, S -> intlib::intOverSet(xpr, x, S))
end_proc:


piecewise::laplace:=
proc(pw)
begin
  piecewise::extmap(pw, transform::laplace, args(2..args(0)))
end_proc:

piecewise::invlaplace:=
proc(pw)
begin
  piecewise::extmap(pw, transform::invlaplace, args(2..args(0)))
end_proc:
  

piecewise::diff :=
proc(pw:piecewise, x)
begin
case args(0)
  of 1 do return(pw)
  of 2 do 
	return(piecewise::extmap(piecewise::mapConditions
                         (pw, piecewise::innerOfCondition), diff, x))
  otherwise
	piecewise::diff(piecewise::diff(pw, x), args(3..args(0)) )
end_case
end_proc:

piecewise::contains:=
proc(pw:piecewise, x)
begin
  if args(0)<>2 then
    error("Wrong number of arguments")
  end_if;
  piecewise::extmap(pw, contains, x)
end_proc:

piecewise::has:=
proc(pw:piecewise, x)
begin
has([extop(pw)], x)
end_proc:

// overload _in::expand

piecewise::_in:=
proc(pw:piecewise, S)
begin
  if args(0)<>2 then
    error("Wrong number of arguments")
  end_if;
  _or((piecewise::condition(pw,i) and piecewise::expression(pw,i) in S )
      $i=1..extnops(pw))
end_proc:


piecewise::simpexIn:= (expression, pw) -> expand(expression in pw):

  
piecewise::set2expr:=
proc(pw:piecewise, x)
begin
  if args(0)<>2 then
    error("Wrong number of arguments")
  end_if;
  _or((piecewise::condition(pw,i) and x in piecewise::expression(pw,i))
      $i=1..extnops(pw))
end_proc:

// since the result of set2prop should be a valid input
// for assume, mapping does not suffice

piecewise::set2prop:=
proc(pw:piecewise, opt = null())
begin
  [solvelib::set2prop(piecewise::expression(pw, i), opt) $i=1..extnops(pw)];
  if has(%, FAIL) then
    FAIL
  else
    _or(op(%))
  end_if
end_proc:

// overload solvelib::getElement
// return FAIL if some conditions cannot be removed

piecewise::getElement:=
proc(pw:piecewise)
begin
  solvelib::getElement(_intersect(piecewise::expression(pw, i) $i=1..extnops(pw)))
end_proc:

piecewise::isFinite:=
proc(pw:piecewise):DOM_BOOL
begin
  [solvelib::isFinite(piecewise::expression(pw, i)) $i=1..extnops(pw)];
  // is the result independent of the conditions ?
  if nops(%)=1 then
    op(%,1)
  else
    UNKNOWN
  end_if
end_proc:


// overload fp::unapply

piecewise::unapply:=
proc(pw /* ,var1, ...,vark */)
  local listform, Args;
  
begin
  listform:=map(extop(pw), u->[op(u,1), op(u,2)]);
  if args(0) = 1 then
    Args:= [ op(indets(pw)) ]
  else
    Args:= [ args(2..args(0)) ]
  end_if;

  if Args = [] then
    subsop(proc() option arrow; begin end, 4=hold(piecewise)(listform),
           Unsimplified)
  else
    subsop(proc() option arrow; begin end, 1=op(Args),
           4=hold(piecewise)
           (subs(listform, [ Args[i]=DOM_VAR(0,1+i) $ i=1..nops(Args) ],
                  Unsimplified)))
  end_if
end_proc:

piecewise::PWif := funcenv(hold(slot)(piecewise, "PWif"),
                           builtin(1100, 100, " if ", "piecewise::PWif")):
piecewise::print := 
proc(pw)
  local i, nargs;
begin
  if PRETTYPRINT=FALSE then
    hold(piecewise)([piecewise::condition(pw,i),
                     piecewise::expression(pw, i)] $i=1..extnops(pw))
  else
    nargs := extnops(pw);
    hold(piecewise)(subsop(piecewise::PWif
                           (hold(dummy),
                            piecewise::condition(pw, i)),
                           1 = piecewise::expression(pw, i),
                           Unsimplified)
                    $i = 1..nargs)
  end_if
end_proc:


piecewise::TeX :=
proc(pw)
begin
  "\\left\\{\n\\begin{array}{ll}\n".
  _concat((generate::TeX(piecewise::expression(pw,i))." & \\mbox{if}\\, ".
          generate::TeX(piecewise::condition(pw, i))."\\\\\n") $i=1..extnops(pw)).
  "\\end{array}\n\\right."
end_proc:

// Typesetting output of piecewise
piecewise::Content := 
proc(Out, pw)
  local i;
begin
    Out::Cpiecewise(
		 (Out(piecewise::expression(pw, i)), Out(piecewise::condition(pw, i)))
                         $ i = 1..extnops(pw)
                 )
end_proc:
