// Dirk Mueller, olorien, 28/10/96, v1.0 

alias( IntTrans = transform::IntTrans ):
alias( invlaplace = transform::invlaplace ):

invlaplace:=newDomain("transform::invlaplace"):
invlaplace::create_dom:=hold(transform::invlaplace):
invlaplace::type:="transform::invlaplace":

invlaplace::has:=proc() local T, v;
begin
  T:=args(1): v:=args(2):
  bool( has(extop(T,1), v) or
        has(extop(T,2), v) or
        has(extop(T,3), v) or
        v=hold(invlaplace)
       )
end_proc:

invlaplace::numer:= id:
invlaplace::denom:= 1:
invlaplace::normal:= id:


invlaplace::new:=
proc(_f, var1, var2) 
  local _f2, back;
begin
   if args(0) = 0 then error("no arguments given") end_if;

   // allow overloading 
   if _f::dom::invlaplace<>FAIL then
     back := _f::dom::invlaplace(args());
     if back <> FAIL then
       return(back)
     else
       return(new(invlaplace, args()))
     end_if
   end_if;

  if testargs() then
    if args(0)<>3 then error("wrong number of arguments") end_if;
 
    if type(_f) = "_equal" then
        if not testtype(_f,Type::Arithmetical=Type::Arithmetical) then
            error("invalid 1st argument")
        end_if
    elif not testtype(_f,Type::Arithmetical) then
        error("1st argument must be of 'Type::Arithmetical'")
    end_if;

    if domtype(var1)<>DOM_IDENT then
      error("second argument must be of type DOM_IDENT");
    end_if;
  end_if;

  back:=invlaplace::compute(_f, var1, var2);

  if back=FAIL and testtype(_f,Type::RatExpr(var1)) then
      back:=invlaplace::compute(partfrac(_f, var1), var1, var2);
  end_if;

  if back=FAIL then
    _f2:=expand(_f);
    if (not (_f=_f2)) then
      back:=invlaplace::compute(_f2, var1, var2);
    end_if;
  end_if;

  if back=FAIL then
    return(new(invlaplace, _f, var1, var2)) 
  end_if;

  return(back)

end_proc:


invlaplace::print:=proc(T)
begin
  hold(invlaplace)(extop(T,1), extop(T,2), extop(T,3))
end_proc:


invlaplace::laplace:=
proc(_f: transform::laplace, var1: DOM_IDENT, var2)
  local t0;
begin
  if extop(_f, 3) = var1 then
    return(eval(subs(extop(_f,1), extop(_f,2)=var2)))
  end_if;
     // Let h(t + t0) = heaviside(t + t0):
     //   laplace( h(t + t0) * invlaplace(f(s), s, t + t0), t, ss)
     // = laplace( invlaplace(exp(s*t0)*f(s), s, t), t, ss)
     // =  exp(ss*t0)*f(ss)
  t0 := extop(_f, 3) - var1;
     // If t0 >= 0, heaviside(t + t0) = 1 for all t >= 0,
     // hence laplace(h(t+t0)*g(t), t, s) = laplace(g(t), t, s)
  if not has(t0, var1) and is(t0 >= 0) = TRUE then
    return(exp(var2*t0)*eval(subs(extop(_f,1), extop(_f,2)=var2)));
  end_if;
end_proc:


//this part computes the laplacian transform
invlaplace::compute:=proc(_g, var1, var2) 
local _f, back, typ,
      exps, others, dummy, processed, a, b, aa, bb, term, pow, i, r;
begin
  _f:=_g;
  if not has(_f, var1) then
    return(_f*dirac(var2));
  elif _f=var1 then
    return(dirac(var2, 1));
  end_if;

  typ:=type(_f);

	// ----------------------
	// Stephan Jan. 3. 2002
	// IL(s^k, s, x) = dirac(x, k)
	if typ = "_power" and op(_f,1) = var1 then
		pow := op(_f,2);
		if testtype(pow, DOM_INT) and pow > 1 then;
		    return(dirac(var2, pow));
		end_if;
	end_if;

  if typ="_equal" then
    return(_equal(invlaplace(op(_f,1), var1, var2), invlaplace(op(_f,2), var1, var2)))
  end_if;

  // -------------------------------------------------------------
  // the following implements
  //     invlaplace( exp(-a*s)*g(s), s, t) = 
  //         heaviside(t - a)* invlaplace(g(s), s, t - a)
  // for a >= 0
  // -------------------------------------------------------------
  if has(_g, exp) and
     (typ = "_mult" or typ = "_power" or typ = "exp")
  then
     if typ = "exp" then
        // _g = exp(a*s + b) * 1
        exps := [_g];
        others := 1;
     end_if;
     if typ = "_power" then
        if type(op(_g, 1)) = "exp" 
        then // _g = exp(a*s + b)^n * 1
             exps := [_g];
             others := 1;
        else exps:= [1];
             others:= _g;
        end_if;
     end_if;
     if typ = "_mult" then
        [exps, others, dummy] := split([op(_g)], 
           proc(x)
           begin
             if type(x) = "exp" or
               (type(x) = "_power" and type(op(x, 1)) = "exp") then
                return(TRUE):
             end_if;
             FALSE;
           end_proc);
        if dummy <> [] then
           // this should not happen
           error("failed to split off exponential factors in ".
                 expr2text(_g))
        end_if;
        others:= _mult(op(others));
     end_if;

     // all factors exp(a1*s + b1)^n1*exp(a2*s + b2)^n1* ...
     // were collected in 'exps', all other factors are
     // collected in 'others' = _g / exp(a1*s + b1)^n1 / exp(a2*s + b2)^n2 / ...
     aa:= 0;  // aa = n1*a1 + n2*a2 + ...
     bb:= 0;  // bb = n1*b1 + n2*b2 + ...
     for i from 1 to nops(exps) do
         processed:= FALSE;
         term:= op(exps, i):
         if not has(term, exp) then next end_if;
         pow := 1:
         // case: term = exp(a*s + b)^n
         if type(term) = "_power" then
            pow:= op(term, 2);
            term:= op(term, 1): // = a*s + b
         end_if;
         if type(term) = "exp" then
              a := diff(pow*op(term, 1), var1);
              if not has(a, var1) then
                 b:= expand(pow*op(term, 1) - a*var1);
                 if not has(b, var1) 
                 then // op(term, 1) = a*var1 + b
                      aa := aa + a;
                      bb := bb + b;
                      processed:= TRUE;
                 end_if;
               end_if;
         end_if;
         if not processed then
            // term is not of the form a*s + b
            // -> collect in others
            others:= others*term^pow;
         end_if;
     end_for;
     if others <> _g then
        r := exp(bb)*transform::invlaplace(others, var1, var2);
        if has(r, FAIL) then return(FAIL); end_if;
  
        if is(aa <= 0) = TRUE then
          //special case:  return dirac(x + a) rather than dirac(x + a)*heaviside(x + a)
          if hastype(r, "dirac") then
             return(eval(subs(r, var2 = var2 + aa)));
          end_if;
  
          return(heaviside(var2 + aa) * eval(subs(r, var2 = var2 + aa)));
        end_if;
     end_if;
  end_if;
  // nothing was changed, if no factors of the form exp(a*x + b)^n
  // were found. Proceed as usual.
  // -------------------------------------------------------------

  if typ="_mult" or typ="_plus" then
    back:=IntTrans::linear(new(invlaplace, _f, var1, var2));
    if not (back=FAIL) then
      return(back);
    end_if;
  end_if;

  if typ="function" and type(op(_f,0))="D" then
    _f:=IntTrans::D2diff(_f)
  end_if;

  back:=IntTrans::builtin(new(invlaplace, _f, var1, var2));
  if not (back=FAIL) then
    return(back);
  end_if;

 return(FAIL);
end_proc:

unalias( invlaplace ):
unalias( IntTrans ):

// end of file 
